/* 
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 * 
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_attr.c Linux driver configuration interface module.
 */

#include "bfad_drv.h"
#include "bfad_im.h"
#include "bfad_trcmod.h"
#include "bfad_attr.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)
struct Scsi_Host *
bfad_os_starget_to_shost(struct scsi_target *starget)
{
	return dev_to_shost(starget->dev.parent);
}
#endif

/*
* TBD: Need to find out the kernel version starting to support
 * FC transport.
 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 10)
/**
 * @brief
 * FC transport template entry, get SCSI host port type.
 */
static void
bfad_im_get_host_port_type(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_pport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_type) {
	case BFA_PPORT_TYPE_NPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_NPORT;
		break;
	case BFA_PPORT_TYPE_NLPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_NLPORT;
		break;
	case BFA_PPORT_TYPE_P2P:
		fc_host_port_type(shost) = FC_PORTTYPE_PTP;
		break;
	case BFA_PPORT_TYPE_LPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_LPORT;
		break;
	default:
		fc_host_port_type(shost) = FC_PORTTYPE_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host port state.
 */
static void
bfad_im_get_host_port_state(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_pport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_state) {
	case BFA_PPORT_ST_LINKDOWN:
		fc_host_port_state(shost) = FC_PORTSTATE_LINKDOWN;
		break;
	case BFA_PPORT_ST_LINKUP:
		fc_host_port_state(shost) = FC_PORTSTATE_ONLINE;
		break;
	case BFA_PPORT_ST_UNINIT:
	case BFA_PPORT_ST_ENABLING_QWAIT:
	case BFA_PPORT_ST_ENABLING:
	case BFA_PPORT_ST_DISABLING_QWAIT:
	case BFA_PPORT_ST_DISABLING:
	case BFA_PPORT_ST_DISABLED:
	case BFA_PPORT_ST_STOPPED:
	case BFA_PPORT_ST_IOCDOWN:
	default:
		fc_host_port_state(shost) = FC_PORTSTATE_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host active fc4s.
 */
static void
bfad_im_get_host_active_fc4s(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;

	memset(fc_host_active_fc4s(shost), 0,
	       sizeof(fc_host_active_fc4s(shost)));

	if (port->
	    supported_fc4s & (BFA_PORT_ROLE_FCP_IM | BFA_PORT_ROLE_FCP_TM))
		fc_host_active_fc4s(shost)[2] = 1;

	if (port->supported_fc4s & BFA_PORT_ROLE_FCP_IPFC)
		fc_host_active_fc4s(shost)[3] = 0x20;

	fc_host_active_fc4s(shost)[7] = 1;
}

/**
 * @brief
 * FC transport template entry, get SCSI host link speed.
 */
static void
bfad_im_get_host_speed(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_pport_get_attr(&bfad->bfa, &attr);
	switch (attr.speed) {
	case BFA_PPORT_SPEED_8GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_8GBIT;
		break;
	case BFA_PPORT_SPEED_4GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_4GBIT;
		break;
	case BFA_PPORT_SPEED_2GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_2GBIT;
		break;
	case BFA_PPORT_SPEED_1GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_1GBIT;
		break;
	default:
		fc_host_speed(shost) = FC_PORTSPEED_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host port type.
 */
static void
bfad_im_get_host_fabric_name(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	wwn_t           fabric_nwwn = 0;

	/*
	 * TBD: Set to fabric nwwn if in switch enviroment. Set to local port
	 * nwwn if private loop or P2P.
	 */
	fabric_nwwn = bfa_fcs_port_get_fabric_name(port->fcs_port);

	fc_host_fabric_name(shost) = bfa_os_htonll(fabric_nwwn);

}

/**
 * @brief
 * FC transport template entry, get BFAD statistics.
 */
static struct fc_host_statistics *
bfad_im_get_stats(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	struct bfad_hal_comp fcomp;
	struct fc_host_statistics *hstats;
	bfa_status_t    rc;
	unsigned long   flags;

	hstats = &bfad->link_stats;
	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	memset(hstats, 0, sizeof(struct fc_host_statistics));
	rc = bfa_pport_get_stats(&bfad->bfa,
				     (bfa_pport_stats_t *) hstats,
				     bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (rc != BFA_STATUS_OK)
		return NULL;

	wait_for_completion(&fcomp.comp);

	return hstats;
}

/**
 * @brief
 * FC transport template entry, reset BFAD statistics.
 */
static void
bfad_im_reset_stats(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;
	bfa_status_t    rc;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	rc = bfa_pport_clear_stats(&bfad->bfa, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (rc != BFA_STATUS_OK)
		return;

	wait_for_completion(&fcomp.comp);

	return;
}

/**
 * @brief
 * FC transport template entry, get rport loss timeout.
 */
static void
bfad_im_get_rport_loss_tmo(struct fc_rport *rport)
{
	struct bfad_itnim_data_s *itnim_data = rport->dd_data;
	struct bfad_itnim_s   *itnim = itnim_data->itnim;
	struct bfad_s         *bfad = itnim->im->bfad;

	rport->dev_loss_tmo = bfad->cfg_data.rport_del_timeout;
}

/**
 * @brief
 * FC transport template entry, set rport loss timeout.
 */
static void
bfad_im_set_rport_loss_tmo(struct fc_rport *rport, uint32_t timeout)
{
	if (timeout > 0)
		rport->dev_loss_tmo = timeout;
	else {
		struct bfad_itnim_data_s *itnim_data = rport->dd_data;
		struct bfad_itnim_s   *itnim = itnim_data->itnim;
		struct bfad_s         *bfad = itnim->im->bfad;
		rport->dev_loss_tmo = bfad->cfg_data.rport_del_timeout;
	}

}
#endif

struct fc_function_template bfad_im_fc_function_template = {

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)
	/*
	 * Target dynamic attributes
	 */
	.get_starget_port_id = bfad_im_get_starget_port_id,
	.show_starget_port_id = 1,
	.get_starget_node_name = bfad_im_get_starget_node_name,
	.show_starget_node_name = 1,
	.get_starget_port_name = bfad_im_get_starget_port_name,
	.show_starget_port_name = 1,

	/*
	 * Host dynamic attribute
	 */
	.get_host_port_id = bfad_im_get_host_port_id,
	.show_host_port_id = 1,

	.issue_fc_host_lip = bfad_im_issue_fc_host_lip,
#endif

	/*
	 * TBD: Need to find out the kernel version starting to support
	 * FC transport
	 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 10)
	/*
	 * Host fixed attributes
	 */
	.show_host_node_name = 1,
	.show_host_port_name = 1,
	.show_host_supported_classes = 1,
	.show_host_supported_fc4s = 1,
	.show_host_supported_speeds = 1,
	.show_host_maxframe_size = 1,

	/*
	 * More host dynamic attributes
	 */
	.show_host_port_type = 1,
	.get_host_port_type = bfad_im_get_host_port_type,
	.show_host_port_state = 1,
	.get_host_port_state = bfad_im_get_host_port_state,
	.show_host_active_fc4s = 1,
	.get_host_active_fc4s = bfad_im_get_host_active_fc4s,
	.show_host_speed = 1,
	.get_host_speed = bfad_im_get_host_speed,
	.show_host_fabric_name = 1,
	.get_host_fabric_name = bfad_im_get_host_fabric_name,

	/*
	 * This is host fixed attribute in lk 2.6.16 but changed to dynamic
	 * in lk 2.6.?
	 */
	.show_host_symbolic_name = 1,

	/*
	 * Statistics
	 */
	.get_fc_host_stats = bfad_im_get_stats,
	.reset_fc_host_stats = bfad_im_reset_stats,

	/*
	 * Allocation length for host specific data
	 */
	.dd_fcrport_size = sizeof(struct bfad_itnim_data_s *),

	/*
	 * Remote port fixed attributes
	 */
	.show_rport_maxframe_size = 1,
	.show_rport_supported_classes = 1,
	.show_rport_dev_loss_tmo = 1,
	.get_rport_dev_loss_tmo = bfad_im_get_rport_loss_tmo,
	.set_rport_dev_loss_tmo = bfad_im_set_rport_loss_tmo,
#endif
};

/**
 * @}
 * @dg Scsi_Host_attrs SCSI host attributes
 * @{
 */
static ssize_t
bfad_im_serial_num_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.serial_num);
}

static ssize_t
bfad_im_model_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.model);
}

static ssize_t
bfad_im_model_desc_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.model_descr);
}

static ssize_t
bfad_im_node_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	uint64_t        nwwn;

	nwwn = bfa_fcs_port_get_nwwn(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", bfa_os_htonll(nwwn));
}

static ssize_t
bfad_im_symbolic_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);

	return snprintf(buf, PAGE_SIZE, "Brocade %s FV%s DV%s\n",
			ioc_attr.adapter_attr.model,
			ioc_attr.adapter_attr.fw_ver, BFAD_DRIVER_VERSION);
}

static ssize_t
bfad_im_hw_version_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.hw_ver);
}

static ssize_t
bfad_im_drv_version_show(struct class_device *cdev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", BFAD_DRIVER_VERSION);
}

static ssize_t
bfad_im_optionrom_version_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.optrom_ver);
}

static ssize_t
bfad_im_fw_version_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.fw_ver);
}

static ssize_t
bfad_im_num_of_ports_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%d\n", ioc_attr.adapter_attr.nports);
}

static ssize_t
bfad_im_drv_name_show(struct class_device *cdev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", BFAD_DRIVER_NAME);
}

static ssize_t
bfad_im_num_of_discovered_ports_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	struct bfad_s         *bfad = im_port->bfad;
	uint32_t        nrports = 2048;
	wwn_t          *rports = NULL;
	unsigned long   flags;

	
	rports = kzalloc(sizeof(wwn_t) * nrports , GFP_ATOMIC);
	if (rports == NULL)
		return snprintf(buf, PAGE_SIZE, "Failed\n");
	
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcs_port_get_rports(port->fcs_port, rports, &nrports);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	kfree(rports);

	return snprintf(buf, PAGE_SIZE, "%d\n", nrports);
}

static int
bfad_im_parse_wwn(const char *buf, uint8_t *wwn)
{
	unsigned int i, j;
	memset(wwn, 0, 8);

	/* Validate and store the new name */
	for (i=0, j=0; i < 16; i++) {
		if ((*buf >= '0') && (*buf <= '9'))
			j = ((j << 4) | (*buf++ -'0'));
		else if ((*buf >= 'a') && (*buf <= 'f'))
			j = ((j << 4) | ((*buf++ -'a') + 10));
		else if ((*buf >= 'A') && (*buf <= 'F'))
			j = ((j << 4) | ((*buf++ -'A') + 10));
		else
			return -EINVAL;
		if (i % 2) {
			wwn[i/2] = j & 0xff;
			j = 0;
		}
	}

	return 0;
}

static ssize_t
bfad_im_vport_create(struct class_device *cdev, const char *buf, size_t count)
{
        struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	bfa_port_cfg_t port_cfg;
        uint8_t wwn[8];
        int status;

	memset(&port_cfg, 0, sizeof(port_cfg));
        status = bfad_im_parse_wwn(&buf[0], wwn);
        if (status)
                return status;
	memcpy(&port_cfg.pwwn, wwn, sizeof(wwn));

        status = bfad_im_parse_wwn(&buf[17], wwn);
        if (status)
                return status;
	memcpy(&port_cfg.nwwn, wwn, sizeof(wwn));

	port_cfg.roles = BFA_PORT_ROLE_FCP_IM;
	status = bfad_vport_create(bfad, 0, &port_cfg);
        if (status != BFA_STATUS_OK)
                return -EIO;

        return count;
}

static ssize_t
bfad_im_vport_delete(struct class_device *cdev, const char *buf, size_t count)
{
        struct Scsi_Host  *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
        uint8_t wwn[8];
        int status;
	unsigned long flags;
	bfa_fcs_vport_t *fcs_vport;
	struct bfad_vport_s *bfad_vport;
	struct completion fcomp;

        status = bfad_im_parse_wwn(&buf[0], wwn);
        if (status)
                return status;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, 0, *(wwn_t *)wwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return -EINVAL;
	}

	bfad_vport = fcs_vport->vport_drv;
	bfad_vport->drv_port.flags |= BFAD_PORT_DELETE;
	bfad_vport->comp_del = &fcomp;
	init_completion(bfad_vport->comp_del);

	status = bfa_fcs_vport_delete(fcs_vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(bfad_vport->comp_del);

	if (status != BFA_STATUS_OK) {
		return -EIO;
	}

	bfad_os_scsi_host_free(bfad, bfad_vport->drv_port.im_port);
	kfree(bfad_vport);
	return count;
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10))
static ssize_t
bfad_im_port_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	uint64_t        pwwn;

	pwwn = bfa_fcs_port_get_pwwn(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", bfa_os_htonll(pwwn));
}

static ssize_t
bfad_im_port_id_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	uint32_t        fc_id;

	fc_id = bfa_fcs_port_get_fcid(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%x\n", bfa_os_hton3b(fc_id));
}

static ssize_t
bfad_im_port_type_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;
	char            porttype[16];

	bfa_pport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_type) {
	case BFA_PPORT_TYPE_NPORT:
		strcpy(porttype, "NPORT");
		break;
	case BFA_PPORT_TYPE_NLPORT:
		strcpy(porttype, "NLPORT");
		break;
	case BFA_PPORT_TYPE_P2P:
		strcpy(porttype, "P2P");
		break;
	case BFA_PPORT_TYPE_LPORT:
		strcpy(porttype, "LPORT");
		break;
	default:
		strcpy(porttype, "UNKNOWN");
		break;
	}

	return snprintf(buf, PAGE_SIZE, "%s\n", porttype);
}

static ssize_t
bfad_im_port_state_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;
	char            portstate[16];

	bfa_pport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_state) {
	case BFA_PPORT_ST_LINKDOWN:
		strcpy(portstate, "Link Down");
		break;
	case BFA_PPORT_ST_LINKUP:
		strcpy(portstate, "Link Up");
		break;
	case BFA_PPORT_ST_UNINIT:
	case BFA_PPORT_ST_ENABLING_QWAIT:
	case BFA_PPORT_ST_ENABLING:
	case BFA_PPORT_ST_DISABLING_QWAIT:
	case BFA_PPORT_ST_DISABLING:
	case BFA_PPORT_ST_DISABLED:
	case BFA_PPORT_ST_STOPPED:
	case BFA_PPORT_ST_IOCDOWN:
	default:
		strcpy(portstate, "Unknown");
		break;
	}

	return snprintf(buf, PAGE_SIZE, "%s\n", portstate);
}

static ssize_t
bfad_im_supported_classes_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;
	uint32_t        classes;

	bfa_pport_get_attr(&bfad->bfa, &attr);
	classes = attr.cos_supported;
	return snprintf(buf, PAGE_SIZE, "%d\n", classes);
}

static ssize_t
bfad_im_supported_fc4s_show(struct class_device *cdev, char *buf)
{
	u8              fc4s[32];
	int             i, len = 0, total = 0;

	memset(fc4s, 0, sizeof(fc4s));
	if (supported_fc4s & (BFA_PORT_ROLE_FCP_IM | BFA_PORT_ROLE_FCP_TM))
		/*
		 * For FCP type 0x08
		 */
		fc4s[2] = 1;
	if (supported_fc4s | BFA_PORT_ROLE_FCP_IPFC)
		/*
		 * For LLC/SNAP type 0x05
		 */
		fc4s[3] = 0x20;
	/*
	 * For fibre channel services type 0x20
	 */
	fc4s[7] = 1;

	for (i = 0; i < 32; i++) {
		len = sprintf(buf + total, "0x%02x", fc4s[i]);
		total += len;
		if (i != 31) {
			len = sprintf(buf + total, "%s", " ");
			total += len;
		}
	}

	return total;
}

static ssize_t
bfad_im_active_fc4s_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	u8              fc4s[32];
	int             i, len = 0, total = 0;

	memset(fc4s, 0, sizeof(fc4s));
	if (port->
	    supported_fc4s & (BFA_PORT_ROLE_FCP_IM | BFA_PORT_ROLE_FCP_TM))
		/*
		 * For FCP type 0x08
		 */
		fc4s[2] = 1;
	if (port->supported_fc4s | BFA_PORT_ROLE_FCP_IPFC)
		/*
		 * For LLC/SNAP type 0x05
		 */
		fc4s[3] = 0x20;
	/*
	 * For fibre channel services type 0x20
	 */
	fc4s[7] = 1;

	for (i = 0; i < 32; i++) {
		len = sprintf(buf + total, "0x%02x", fc4s[i]);
		total += len;
		if (i != 31) {
			len = sprintf(buf + total, "%s", " ");
			total += len;
		}
	}

	return total;
}

static ssize_t
bfad_im_supported_speeds_show(struct class_device *cdev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n",
			"1 Gbit, 2 Gbit, 4 Gbit," " 8 Gbit");
}

static ssize_t
bfad_im_speed_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;
	int             speed;

	bfa_pport_get_attr(&bfad->bfa, &attr);
	speed = attr.speed;
	return snprintf(buf, PAGE_SIZE, "%d Gbit\n", speed);
}

static ssize_t
bfad_im_maxframe_size_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;
	uint32_t        maxfrsize;

	bfa_pport_get_attr(&bfad->bfa, &attr);
	maxfrsize = attr.pport_cfg.maxfrsize;
	return snprintf(buf, PAGE_SIZE, "%d\n", maxfrsize);
}

static ssize_t
bfad_im_fabric_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port = (struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	wwn_t           fabric_nwwn = 0;

	/*
	 * TBD: Set to fabric nwwn if in switch enviroment. Set to local port
	 * nwwn if private loop or P2P.
	 */
	fabric_nwwn = bfa_fcs_port_get_fabric_name(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", bfa_os_htonll(fabric_nwwn));
}
#endif

static CLASS_DEVICE_ATTR(serial_number, S_IRUGO,
				bfad_im_serial_num_show, NULL);
static CLASS_DEVICE_ATTR(model, S_IRUGO, bfad_im_model_show, NULL);
static CLASS_DEVICE_ATTR(model_description, S_IRUGO,
				bfad_im_model_desc_show, NULL);
static CLASS_DEVICE_ATTR(node_name, S_IRUGO, bfad_im_node_name_show, NULL);
static CLASS_DEVICE_ATTR(symbolic_name, S_IRUGO,
				bfad_im_symbolic_name_show, NULL);
static CLASS_DEVICE_ATTR(hardware_version, S_IRUGO,
				bfad_im_hw_version_show, NULL);
static CLASS_DEVICE_ATTR(driver_version, S_IRUGO,
				bfad_im_drv_version_show, NULL);
static CLASS_DEVICE_ATTR(option_rom_version, S_IRUGO,
				bfad_im_optionrom_version_show, NULL);
static CLASS_DEVICE_ATTR(firmware_version, S_IRUGO,
				bfad_im_fw_version_show, NULL);
static CLASS_DEVICE_ATTR(number_of_ports, S_IRUGO,
				bfad_im_num_of_ports_show, NULL);
static CLASS_DEVICE_ATTR(driver_name, S_IRUGO, bfad_im_drv_name_show, NULL);
static CLASS_DEVICE_ATTR(number_of_discovered_ports, S_IRUGO,
				bfad_im_num_of_discovered_ports_show, NULL);
static CLASS_DEVICE_ATTR(vport_create, S_IWUSR, NULL, bfad_im_vport_create);
static CLASS_DEVICE_ATTR(vport_delete, S_IWUSR, NULL, bfad_im_vport_delete);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10))
static CLASS_DEVICE_ATTR(port_name, S_IRUGO, bfad_im_port_name_show, NULL);
static CLASS_DEVICE_ATTR(port_id, S_IRUGO, bfad_im_port_id_show, NULL);
static CLASS_DEVICE_ATTR(port_type, S_IRUGO, bfad_im_port_type_show, NULL);
static CLASS_DEVICE_ATTR(port_state, S_IRUGO, bfad_im_port_state_show, NULL);
static CLASS_DEVICE_ATTR(supported_classes, S_IRUGO,
				bfad_im_supported_classes_show, NULL);
static CLASS_DEVICE_ATTR(supported_fc4s, S_IRUGO,
				bfad_im_supported_fc4s_show, NULL);
static CLASS_DEVICE_ATTR(active_fc4s, S_IRUGO, bfad_im_active_fc4s_show, NULL);
static CLASS_DEVICE_ATTR(supported_speeds, S_IRUGO,
				bfad_im_supported_speeds_show, NULL);
static CLASS_DEVICE_ATTR(speed, S_IRUGO, bfad_im_speed_show, NULL);
static CLASS_DEVICE_ATTR(maxframe_size, S_IRUGO,
				bfad_im_maxframe_size_show, NULL);
static CLASS_DEVICE_ATTR(fabric_name, S_IRUGO, bfad_im_fabric_name_show, NULL);
#endif

struct class_device_attribute *bfad_im_host_attrs[] = {
	&class_device_attr_serial_number,
	&class_device_attr_model,
	&class_device_attr_model_description,
	&class_device_attr_node_name,
	&class_device_attr_symbolic_name,
	&class_device_attr_hardware_version,
	&class_device_attr_driver_version,
	&class_device_attr_option_rom_version,
	&class_device_attr_firmware_version,
	&class_device_attr_number_of_ports,
	&class_device_attr_driver_name,
	&class_device_attr_number_of_discovered_ports,
        &class_device_attr_vport_create,
        &class_device_attr_vport_delete,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10))
	&class_device_attr_port_name,
	&class_device_attr_port_id,
	&class_device_attr_port_type,
	&class_device_attr_port_state,
	&class_device_attr_supported_classes,
	&class_device_attr_supported_fc4s,
	&class_device_attr_active_fc4s,
	&class_device_attr_supported_speeds,
	&class_device_attr_speed,
	&class_device_attr_maxframe_size,
	&class_device_attr_fabric_name,
#endif
	NULL,
};

struct class_device_attribute *bfad_im_vport_attrs[] = {
	&class_device_attr_serial_number,
	&class_device_attr_model,
	&class_device_attr_model_description,
	&class_device_attr_node_name,
	&class_device_attr_symbolic_name,
	&class_device_attr_hardware_version,
	&class_device_attr_driver_version,
	&class_device_attr_option_rom_version,
	&class_device_attr_firmware_version,
	&class_device_attr_number_of_ports,
	&class_device_attr_driver_name,
	&class_device_attr_number_of_discovered_ports,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10))
	&class_device_attr_port_name,
	&class_device_attr_port_id,
	&class_device_attr_port_type,
	&class_device_attr_port_state,
	&class_device_attr_supported_classes,
	&class_device_attr_supported_fc4s,
	&class_device_attr_active_fc4s,
	&class_device_attr_supported_speeds,
	&class_device_attr_speed,
	&class_device_attr_maxframe_size,
	&class_device_attr_fabric_name,
#endif
	NULL,
};

#else /* linux version > 2.6.26 */
struct Scsi_Host *
bfad_os_starget_to_shost(struct scsi_target *starget)
{
	return dev_to_shost(starget->dev.parent);
}

/**
 * @brief
 * FC transport template entry, get SCSI host port type.
 */
static void
bfad_im_get_host_port_type(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_pport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_type) {
	case BFA_PPORT_TYPE_NPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_NPORT;
		break;
	case BFA_PPORT_TYPE_NLPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_NLPORT;
		break;
	case BFA_PPORT_TYPE_P2P:
		fc_host_port_type(shost) = FC_PORTTYPE_PTP;
		break;
	case BFA_PPORT_TYPE_LPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_LPORT;
		break;
	default:
		fc_host_port_type(shost) = FC_PORTTYPE_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host port state.
 */
static void
bfad_im_get_host_port_state(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_pport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_state) {
	case BFA_PPORT_ST_LINKDOWN:
		fc_host_port_state(shost) = FC_PORTSTATE_LINKDOWN;
		break;
	case BFA_PPORT_ST_LINKUP:
		fc_host_port_state(shost) = FC_PORTSTATE_ONLINE;
		break;
	case BFA_PPORT_ST_UNINIT:
	case BFA_PPORT_ST_ENABLING_QWAIT:
	case BFA_PPORT_ST_ENABLING:
	case BFA_PPORT_ST_DISABLING_QWAIT:
	case BFA_PPORT_ST_DISABLING:
	case BFA_PPORT_ST_DISABLED:
	case BFA_PPORT_ST_STOPPED:
	case BFA_PPORT_ST_IOCDOWN:
	default:
		fc_host_port_state(shost) = FC_PORTSTATE_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host active fc4s.
 */
static void
bfad_im_get_host_active_fc4s(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;

	memset(fc_host_active_fc4s(shost), 0,
	       sizeof(fc_host_active_fc4s(shost)));

	if (port->
	    supported_fc4s & (BFA_PORT_ROLE_FCP_IM | BFA_PORT_ROLE_FCP_TM))
		fc_host_active_fc4s(shost)[2] = 1;

	if (port->supported_fc4s & BFA_PORT_ROLE_FCP_IPFC)
		fc_host_active_fc4s(shost)[3] = 0x20;

	fc_host_active_fc4s(shost)[7] = 1;
}

/**
 * @brief
 * FC transport template entry, get SCSI host link speed.
 */
static void
bfad_im_get_host_speed(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_pport_get_attr(&bfad->bfa, &attr);
	switch (attr.speed) {
	case BFA_PPORT_SPEED_8GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_8GBIT;
		break;
	case BFA_PPORT_SPEED_4GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_4GBIT;
		break;
	case BFA_PPORT_SPEED_2GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_2GBIT;
		break;
	case BFA_PPORT_SPEED_1GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_1GBIT;
		break;
	default:
		fc_host_speed(shost) = FC_PORTSPEED_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host port type.
 */
static void
bfad_im_get_host_fabric_name(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	wwn_t           fabric_nwwn = 0;

	fabric_nwwn = bfa_fcs_port_get_fabric_name(port->fcs_port);

	fc_host_fabric_name(shost) = bfa_os_htonll(fabric_nwwn);

}

/**
 * @brief
 * FC transport template entry, get BFAD statistics.
 */
static struct fc_host_statistics *
bfad_im_get_stats(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	struct bfad_hal_comp fcomp;
	struct fc_host_statistics *hstats;
	bfa_status_t    rc;
	unsigned long   flags;

	hstats = &bfad->link_stats;
	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	memset(hstats, 0, sizeof(struct fc_host_statistics));
	rc = bfa_pport_get_stats(&bfad->bfa,
				     (bfa_pport_stats_t *) hstats,
				     bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (rc != BFA_STATUS_OK)
		return NULL;

	wait_for_completion(&fcomp.comp);

	return hstats;
}

/**
 * @brief
 * FC transport template entry, reset BFAD statistics.
 */
static void
bfad_im_reset_stats(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;
	bfa_status_t    rc;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	rc = bfa_pport_clear_stats(&bfad->bfa, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (rc != BFA_STATUS_OK)
		return;

	wait_for_completion(&fcomp.comp);

	return;
}

/**
 * @brief
 * FC transport template entry, get rport loss timeout.
 */
static void
bfad_im_get_rport_loss_tmo(struct fc_rport *rport)
{
	struct bfad_itnim_data_s *itnim_data = rport->dd_data;
	struct bfad_itnim_s   *itnim = itnim_data->itnim;
	struct bfad_s         *bfad = itnim->im->bfad;

	rport->dev_loss_tmo = bfad->cfg_data.rport_del_timeout;
}

/**
 * @brief
 * FC transport template entry, set rport loss timeout.
 */
static void
bfad_im_set_rport_loss_tmo(struct fc_rport *rport, uint32_t timeout)
{
	if (timeout > 0)
		rport->dev_loss_tmo = timeout;
	else {
		struct bfad_itnim_data_s *itnim_data = rport->dd_data;
		struct bfad_itnim_s   *itnim = itnim_data->itnim;
		struct bfad_s         *bfad = itnim->im->bfad;
		rport->dev_loss_tmo = bfad->cfg_data.rport_del_timeout;
	}

}

struct fc_function_template bfad_im_fc_function_template = {

	/*
	 * Target dynamic attributes
	 */
	.get_starget_port_id = bfad_im_get_starget_port_id,
	.show_starget_port_id = 1,
	.get_starget_node_name = bfad_im_get_starget_node_name,
	.show_starget_node_name = 1,
	.get_starget_port_name = bfad_im_get_starget_port_name,
	.show_starget_port_name = 1,

	/*
	 * Host dynamic attribute
	 */
	.get_host_port_id = bfad_im_get_host_port_id,
	.show_host_port_id = 1,

	.issue_fc_host_lip = bfad_im_issue_fc_host_lip,

	/*
	 * Host fixed attributes
	 */
	.show_host_node_name = 1,
	.show_host_port_name = 1,
	.show_host_supported_classes = 1,
	.show_host_supported_fc4s = 1,
	.show_host_supported_speeds = 1,
	.show_host_maxframe_size = 1,

	/*
	 * More host dynamic attributes
	 */
	.show_host_port_type = 1,
	.get_host_port_type = bfad_im_get_host_port_type,
	.show_host_port_state = 1,
	.get_host_port_state = bfad_im_get_host_port_state,
	.show_host_active_fc4s = 1,
	.get_host_active_fc4s = bfad_im_get_host_active_fc4s,
	.show_host_speed = 1,
	.get_host_speed = bfad_im_get_host_speed,
	.show_host_fabric_name = 1,
	.get_host_fabric_name = bfad_im_get_host_fabric_name,

	.show_host_symbolic_name = 1,

	/*
	 * Statistics
	 */
	.get_fc_host_stats = bfad_im_get_stats,
	.reset_fc_host_stats = bfad_im_reset_stats,

	/*
	 * Allocation length for host specific data
	 */
	.dd_fcrport_size = sizeof(struct bfad_itnim_data_s *),

	/*
	 * Remote port fixed attributes
	 */
	.show_rport_maxframe_size = 1,
	.show_rport_supported_classes = 1,
	.show_rport_dev_loss_tmo = 1,
	.get_rport_dev_loss_tmo = bfad_im_get_rport_loss_tmo,
	.set_rport_dev_loss_tmo = bfad_im_set_rport_loss_tmo,
};

/**
 * @}
 * @dg Scsi_Host_attrs SCSI host attributes
 * @{
 */
static ssize_t
bfad_im_serial_num_show(struct device *dev, struct device_attribute *attr,
			 char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.serial_num);
}

static ssize_t
bfad_im_model_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.model);
}

static ssize_t
bfad_im_model_desc_show(struct device *dev, struct device_attribute *attr,
				 char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.model_descr);
}

static ssize_t
bfad_im_node_name_show(struct device *dev, struct device_attribute *attr,
				 char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	uint64_t        nwwn;

	nwwn = bfa_fcs_port_get_nwwn(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", bfa_os_htonll(nwwn));
}

static ssize_t
bfad_im_symbolic_name_show(struct device *dev, struct device_attribute *attr,
				 char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);

	return snprintf(buf, PAGE_SIZE, "Brocade %s FV%s DV%s\n",
			ioc_attr.adapter_attr.model,
			ioc_attr.adapter_attr.fw_ver, BFAD_DRIVER_VERSION);
}

static ssize_t
bfad_im_hw_version_show(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.hw_ver);
}

static ssize_t
bfad_im_drv_version_show(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", BFAD_DRIVER_VERSION);
}

static ssize_t
bfad_im_optionrom_version_show(struct device *dev,
			 struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.optrom_ver);
}

static ssize_t
bfad_im_fw_version_show(struct device *dev, struct device_attribute *attr,
				 char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.fw_ver);
}

static ssize_t
bfad_im_num_of_ports_show(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_ioc_get_attr(&bfad->bfa, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%d\n", ioc_attr.adapter_attr.nports);
}

static ssize_t
bfad_im_drv_name_show(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", BFAD_DRIVER_NAME);
}

static ssize_t
bfad_im_num_of_discovered_ports_show(struct device *dev,
			struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	struct bfad_s         *bfad = im_port->bfad;
	int        nrports = 2048;
	wwn_t          *rports = NULL;
	unsigned long   flags;

	rports = kzalloc(sizeof(wwn_t) * nrports , GFP_ATOMIC);
	if (rports == NULL)
		return snprintf(buf, PAGE_SIZE, "Failed\n");

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcs_port_get_rports(port->fcs_port, rports, &nrports);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	kfree(rports);

	return snprintf(buf, PAGE_SIZE, "%d\n", nrports);
}

static int
bfad_im_parse_wwn(const char *buf, uint8_t *wwn)
{
	unsigned int i, j;
	memset(wwn, 0, 8);

	/* Validate and store the new name */
	for (i=0, j=0; i < 16; i++) {
		if ((*buf >= '0') && (*buf <= '9'))
			j = ((j << 4) | (*buf++ -'0'));
		else if ((*buf >= 'a') && (*buf <= 'f'))
			j = ((j << 4) | ((*buf++ -'a') + 10));
		else if ((*buf >= 'A') && (*buf <= 'F'))
			j = ((j << 4) | ((*buf++ -'A') + 10));
		else
			return -EINVAL;
		if (i % 2) {
			wwn[i/2] = j & 0xff;
			j = 0;
		}
	}

	return 0;
}

static ssize_t
bfad_im_vport_create(struct device *dev, struct device_attribute *attr,
			const char *buf, size_t count)
{
        struct Scsi_Host *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	bfa_port_cfg_t port_cfg;
        uint8_t wwn[8];
        int status;

	memset(&port_cfg, 0, sizeof(port_cfg));
        status = bfad_im_parse_wwn(&buf[0], wwn);
        if (status)
                return status;
	memcpy(&port_cfg.pwwn, wwn, sizeof(wwn));

        status = bfad_im_parse_wwn(&buf[17], wwn);
        if (status)
                return status;
	memcpy(&port_cfg.nwwn, wwn, sizeof(wwn));

	port_cfg.roles = BFA_PORT_ROLE_FCP_IM;
	status = bfad_vport_create(bfad, 0, &port_cfg);
        if (status != BFA_STATUS_OK)
                return -EIO;

        return strlen(buf);
}

static ssize_t
bfad_im_vport_delete(struct device *dev, struct device_attribute *attr,
			const char *buf, size_t count)
{
        struct Scsi_Host  *shost = class_to_shost(dev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
        uint8_t wwn[8];
        int status;
	unsigned long flags;
	bfa_fcs_vport_t *fcs_vport;
	struct bfad_vport_s *bfad_vport;
	struct completion fcomp;

        status = bfad_im_parse_wwn(&buf[0], wwn);
        if (status)
                return status;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, 0, *(wwn_t *)wwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return -EINVAL;
	}

	bfad_vport = fcs_vport->vport_drv;
	bfad_vport->drv_port.flags |= BFAD_PORT_DELETE;
	bfad_vport->comp_del = &fcomp;
	init_completion(bfad_vport->comp_del);

	status = bfa_fcs_vport_delete(fcs_vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(bfad_vport->comp_del);

	if (status != BFA_STATUS_OK) {
		return -EIO;
	}

	bfad_os_scsi_host_free(bfad, bfad_vport->drv_port.im_port);
	kfree(bfad_vport);
        return strlen(buf);
}

static          DEVICE_ATTR(serial_number, S_IRUGO,
				bfad_im_serial_num_show, NULL);
static          DEVICE_ATTR(model, S_IRUGO, bfad_im_model_show, NULL);
static          DEVICE_ATTR(model_description, S_IRUGO,
				bfad_im_model_desc_show, NULL);
static          DEVICE_ATTR(node_name, S_IRUGO, bfad_im_node_name_show, NULL);
static          DEVICE_ATTR(symbolic_name, S_IRUGO,
				bfad_im_symbolic_name_show, NULL);
static          DEVICE_ATTR(hardware_version, S_IRUGO,
				bfad_im_hw_version_show, NULL);
static          DEVICE_ATTR(driver_version, S_IRUGO,
				bfad_im_drv_version_show, NULL);
static          DEVICE_ATTR(option_rom_version, S_IRUGO,
				bfad_im_optionrom_version_show, NULL);
static          DEVICE_ATTR(firmware_version, S_IRUGO,
				bfad_im_fw_version_show, NULL);
static          DEVICE_ATTR(number_of_ports, S_IRUGO,
				bfad_im_num_of_ports_show, NULL);
static          DEVICE_ATTR(driver_name, S_IRUGO, bfad_im_drv_name_show, NULL);
static          DEVICE_ATTR(number_of_discovered_ports, S_IRUGO,
				bfad_im_num_of_discovered_ports_show, NULL);
static		DEVICE_ATTR(vport_create, S_IWUSR, NULL, bfad_im_vport_create);
static		DEVICE_ATTR(vport_delete, S_IWUSR, NULL, bfad_im_vport_delete);

struct device_attribute *bfad_im_host_attrs[] = {
	&dev_attr_serial_number,
	&dev_attr_model,
	&dev_attr_model_description,
	&dev_attr_node_name,
	&dev_attr_symbolic_name,
	&dev_attr_hardware_version,
	&dev_attr_driver_version,
	&dev_attr_option_rom_version,
	&dev_attr_firmware_version,
	&dev_attr_number_of_ports,
	&dev_attr_driver_name,
	&dev_attr_number_of_discovered_ports,
        &dev_attr_vport_create,
        &dev_attr_vport_delete,
	NULL,
};

struct device_attribute *bfad_im_vport_attrs[] = {
	&dev_attr_serial_number,
	&dev_attr_model,
	&dev_attr_model_description,
	&dev_attr_node_name,
	&dev_attr_symbolic_name,
	&dev_attr_hardware_version,
	&dev_attr_driver_version,
	&dev_attr_option_rom_version,
	&dev_attr_firmware_version,
	&dev_attr_number_of_ports,
	&dev_attr_driver_name,
	&dev_attr_number_of_discovered_ports,
	NULL,
};

#endif
/**
 * @}
 */
