/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @brief
 * Contains base driver definitions.
 */

/**
 * @file bfa_drv.h Linux driver data structures.
 */

#ifndef __BFAD_DRV_H__
#define __BFAD_DRV_H__

#include "bfa_os_inc.h"

#include <bfa.h>
#include <bfa_svc.h>
#include <fcs/bfa_fcs.h>
#include <defs/bfa_defs_pci.h>
#include <defs/bfa_defs_port.h>
#include <defs/bfa_defs_rport.h>
#include <fcs/bfa_fcs_rport.h>
#include <defs/bfa_defs_vport.h>
#include <fcs/bfa_fcs_vport.h>

#include <cs/bfa_plog.h>
#include "aen/bfa_aen.h"
#include <log/bfa_log_linux.h>

#define BFAD_DRIVER_NAME        "bfa"
#ifdef BFA_DRIVER_VERSION
#define BFAD_DRIVER_VERSION    BFA_DRIVER_VERSION
#else
#define BFAD_DRIVER_VERSION    "1.1.0.1"
#endif

/*
 * TBD: PCIE AER is only availabe for the recently release kernels, need to
 * find out the exact kernel version started to support this
 */
#if defined(CONFIG_PCIEPORTBUS) && defined(CONFIG_PCIEAER)
#define BFAD_ENABLE_PCIE_AER(x) pci_enable_pcie_error_reporting(x)
#else
#define BFAD_ENABLE_PCIE_AER(x) {}
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11))
#define pm_message_t u32
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18))
#define BFAD_IRQ_FLAGS SA_SHIRQ
#else
#define BFAD_IRQ_FLAGS IRQF_SHARED
#endif

#ifndef FC_PORTSPEED_8GBIT
#define FC_PORTSPEED_8GBIT 0x10
#endif

/*
 * BFAD flags
 */
#define BFAD_MSIX_ON				0x00000001
#define BFAD_HAL_INIT_DONE			0x00000002
#define BFAD_DRV_INIT_DONE			0x00000004
#define BFAD_CFG_PPORT_DONE			0x00000008
#define BFAD_HAL_START_DONE			0x00000010
#define BFAD_PORT_ONLINE			0x00000020
#define BFAD_RPORT_ONLINE			0x00000040

#define BFAD_PORT_DELETE			0x00000001

/*
 * BFAD related definition
 */
#define SCSI_SCAN_DELAY		HZ
#define BFAD_STOP_TIMEOUT	30
#define BFAD_SUSPEND_TIMEOUT	BFAD_STOP_TIMEOUT

/*
 * BFAD configuration parameter default values
 */
#define BFAD_LUN_QUEUE_DEPTH 		32
#define BFAD_IO_MAX_SGE 		SG_ALL

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
#define bfad_isr_t irq_handler_t
#else
typedef irqreturn_t(*bfad_isr_t) (int, void *, struct pt_regs *);
#endif

#ifdef CONFIG_PCI_MSI
#define MAX_MSIX_ENTRY 24

struct bfad_msix {
	bfad_isr_t      handler;
	struct msix_entry msix;
};
#endif

enum bfad_port_pvb_type {
	BFAD_PORT_PHYS_BASE = 0,
	BFAD_PORT_PHYS_VPORT = 1,
	BFAD_PORT_VF_BASE = 2,
	BFAD_PORT_VF_VPORT = 3,
};

/*
 * PORT data structure
 */
struct bfad_port_s {
	struct list_head list_entry;
	struct bfad_s         *bfad;
	bfa_fcs_port_t *fcs_port;
	uint32_t        roles;
	int32_t         flags;
	uint32_t        supported_fc4s;
	uint8_t		ipfc_flags;
	enum bfad_port_pvb_type pvb_type;
	struct bfad_im_port_s *im_port;	/* IM specific data */
	struct bfad_tm_port_s *tm_port;	/* TM specific data */
	struct bfad_ipfc_port_s *ipfc_port;	/* IPFC specific data */
};

/*
 * VPORT data structure
 */
struct bfad_vport_s {
	struct bfad_port_s     drv_port;
	bfa_fcs_vport_t fcs_vport;
	struct completion *comp_del;
};

/*
 * VF data structure
 */
struct bfad_vf_s {
	bfa_fcs_vf_t    fcs_vf;
	struct bfad_port_s    base_port;	/* base port for vf */
	struct bfad_s   *bfad;
};

struct bfad_cfg_param_s {
	uint32_t        rport_del_timeout;
	uint32_t        ioc_queue_depth;
	uint32_t        lun_queue_depth;
	uint32_t        io_max_sge;
	uint32_t        binding_method;
};

/*
 * BFAD (PCI function) data structure
 */
struct bfad_s {
	struct list_head list_entry;
	bfa_t       bfa;
	bfa_fcs_t       bfa_fcs;
	struct pci_dev *pcidev;
	const char           *pci_name;
	bfa_pcidev_t hal_pcidev;
	bfa_ioc_pci_attr_t pci_attr;
	unsigned long   pci_bar0_map;
	void __iomem   *pci_bar0_kva;
	struct completion comp;
	struct completion suspend;
	struct completion disable_comp;
	bfa_boolean_t   disable_active;
	struct bfad_port_s     pport;	/* physical port of the BFAD */
	bfa_meminfo_t meminfo;
	bfa_ioc_cfg_t   ioc_cfg;
	uint32_t        inst_no;	/* BFAD instance number */
	uint32_t        bfad_flags;
	spinlock_t      bfad_lock;
	struct bfad_cfg_param_s cfg_data;
#ifdef CONFIG_PCI_MSI
	struct bfad_msix msix_tab[MAX_MSIX_ENTRY];
	int             nvec;
#endif

#if CONFIG_PM && (LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 9))
	uint32_t        pci_state[64];	/* save PCI config space */
#endif
	char            adapter_name[BFA_ADAPTER_SYM_NAME_LEN];
	char            port_name[BFA_ADAPTER_SYM_NAME_LEN];
	struct timer_list hal_tmo;
	unsigned long   hs_start;
	struct bfad_im_s *im;		/* IM specific data */
	struct bfad_tm_s *tm;		/* TM specific data */
	struct bfad_ipfc_s *ipfc;	/* IPFC specific data */
	bfa_log_mod_t   log_data;
	bfa_trc_mod_t  *trcmod;
	bfa_log_mod_t  *logmod;
	bfa_aen_t      *aen;
	bfa_aen_t       aen_buf;
	bfa_plog_t      plog_buf;
	int             ref_count;
	bfa_boolean_t	ipfc_enabled;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 10))
	struct fc_host_statistics link_stats;
#endif

	struct kobject *bfa_kobj;
	struct kobject *ioc_kobj;
	struct kobject *pport_kobj;
	struct kobject *lport_kobj;
};

/*
 * RPORT data structure
 */
struct bfad_rport_s {
	bfa_fcs_rport_t fcs_rport;
};

struct bfad_buf_info {
	void           *virt;
	dma_addr_t      phys;
	uint32_t        size;
};

struct bfad_fcxp {
	struct bfad_port_s    *port;
	bfa_rport_t *bfa_rport;
	bfa_status_t    req_status;
	uint16_t        tag;
	uint16_t        rsp_len;
	uint16_t        rsp_maxlen;
	uint8_t         use_ireqbuf;
	uint8_t         use_irspbuf;
	uint32_t        num_req_sgles;
	uint32_t        num_rsp_sgles;
	fchs_t          fchs;
	void           *reqbuf_info;
	void           *rspbuf_info;
	bfa_sge_t  *req_sge;
	bfa_sge_t  *rsp_sge;
	fcxp_send_cb_t  send_cbfn;
	void           *send_cbarg;
	void           *bfa_fcxp;
	struct completion comp;
};

struct bfad_hal_comp {
	bfa_status_t    status;
	struct completion comp;
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 14)
#ifndef kzalloc
#define kzalloc bfad_kzalloc
static inline void *
bfad_kzalloc(size_t size, int flags)
{
	void           *ptr = kmalloc(size, flags);
	if (ptr)
		memset(ptr, 0, size);
	return ptr;
}
#endif
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
#define BFAD_WORK_HANDLER(name) void name(struct work_struct *work)
#define BFAD_INIT_WORK(x, work, func) INIT_WORK(&(x)->work, func)
#else
#define BFAD_WORK_HANDLER(name) void name(void *work)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0))
#define BFAD_INIT_WORK(x, work, func) 			\
	INIT_WORK(&(x)->work, func, &(x)->work)
#else
#define BFAD_INIT_WORK(x, work, func) 			\
	INIT_TQUEUE(&(x)->work, func, &(x)->work)
#endif
#endif

#define list_remove_head(list, entry, type, member)       	\
do {                                                    	\
	entry = NULL;                                           \
	if (!list_empty(list)) {                                \
		entry = list_entry((list)->next, type, member); 	\
		list_del_init(&entry->member);                  	\
	}								\
} while (0)

#define list_get_first(list, type, member)				\
((list_empty(list)) ? NULL :						\
	list_entry((list)->next, type, member))

bfa_status_t    bfad_vport_create(struct bfad_s *bfad, uint16_t vf_id,
				  bfa_port_cfg_t *port_cfg);
bfa_status_t    bfad_vf_create(struct bfad_s *bfad, uint16_t vf_id,
			       bfa_port_cfg_t *port_cfg);
bfa_status_t    bfad_cfg_pport(struct bfad_s *bfad, bfa_port_role_t role);
bfa_status_t    bfad_fc4_probe(struct bfad_s *bfad);
bfa_status_t    bfad_drv_init(struct bfad_s *bfad);
struct bfad_s         *bfad_find_bfad_by_inst_no(int inst_no);
void            bfad_drv_start(struct bfad_s *bfad);
void            bfad_uncfg_pport(struct bfad_s *bfad);
void            bfad_drv_stop(struct bfad_s *bfad);
void            bfad_remove_intr(struct bfad_s *bfad);
void            bfad_fc4_probe_undo(struct bfad_s *bfad);
void            bfad_hal_mem_release(struct bfad_s *bfad);
void            bfad_hcb_comp(void *arg, bfa_status_t status);

int             bfad_os_ioctl_init(void);
int             bfad_os_ioctl_exit(void);
int             bfad_setup_intr(struct bfad_s *bfad);
void            bfad_remove_intr(struct bfad_s *bfad);
int             bfad_os_pci_register_driver(struct pci_driver *drv);
void            bfad_os_pci_unregister_driver(struct pci_driver *drv);
void            bfad_os_device_sysfs_create(struct bfad_s *);
void            bfad_os_device_sysfs_remove(struct bfad_s *);
void            bfad_os_pci_set_mwi(struct pci_dev *pdev);
void            bfad_os_idr_init(struct idr *idr);
void            bfad_os_idr_destroy(struct idr *idr);
void 		*bfad_os_dma_alloc(struct bfad_s *bfad, bfa_mem_elem_t
				*meminfo_elem, dma_addr_t *phys_addr);
void 		bfad_os_dma_free(struct bfad_s *bfad, bfa_mem_elem_t
					*meminfo_elem);
void 		bfad_os_idr_remove(struct idr *idp, int id);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 9))
int 		bfad_os_idr_pre_get(struct idr *idp, int gfp_mask);
#else
int 		bfad_os_idr_pre_get(struct idr *idp, gfp_t gfp_mask);
#endif
void 		bfad_os_iounmap(struct pci_dev *pdev, struct bfad_s *bfad);

void		bfad_update_hal_cfg(bfa_ioc_cfg_t *hal_cfg);
bfa_status_t	bfad_hal_mem_alloc(struct bfad_s *bfad);
void		bfad_bfa_tmo(unsigned long data);
void		bfad_init_timer(struct bfad_s *bfad);
int		bfad_pci_init(struct pci_dev *pdev, struct bfad_s *bfad);
void		bfad_pci_uninit(struct pci_dev *pdev, struct bfad_s *bfad);
void		bfad_fcs_port_cfg(struct bfad_s *bfad);
void		bfad_drv_uninit(struct bfad_s *bfad);
void		bfad_drv_log_level_set(struct bfad_s *bfad);
bfa_status_t	bfad_fc4_module_init(void);
void		bfad_fc4_module_exit(void);
int 		bfad_ipfc_probe(struct bfad_s *bfad);
int 		bfad_ipfc_probe_undo(struct bfad_s *bfad);
void 		bfad_ipfc_probe_post(struct bfad_s *bfad);
int 		bfad_ipfc_module_init(void);
void 		bfad_ipfc_module_exit(void);
int             bfad_ipfc_port_online(struct bfad_s *bfad,
                struct bfad_port_s *port);
int             bfad_ipfc_port_offline(struct bfad_s *bfad,
                struct bfad_port_s *port);
int             bfad_ipfc_port_new(struct bfad_s *bfad,
                struct bfad_port_s *port, int port_type);
int             bfad_ipfc_port_delete(struct bfad_s *bfad,
                struct bfad_port_s *port);




void bfad_pci_remove(struct pci_dev *pdev);
int bfad_pci_probe(struct pci_dev *pdev, const struct pci_device_id *pid);
void bfad_os_rport_online_wait(struct bfad_s *bfad);

extern struct idr bfad_im_port_index;
extern struct pci_driver *bfad_pci_driver_p;

extern struct pci_device_id bfad_id_table[];
extern struct semaphore bfad_sem;
extern int bfad_scan_done;
extern struct list_head bfad_list;
extern char	*os_name;
extern char	*os_patch;
extern char	*host_name;
extern int	num_rports;
extern int	num_ios;
extern int	num_tms;
extern int	num_fcxps;
extern int	num_ufbufs;
extern int	reqq_size;
extern int	rspq_size;
extern int	num_sgpgs;
extern int      rport_del_timeout;
extern int      msix_disable;
extern int      bfa_lun_queue_depth;
extern int      bfa_io_max_sge;
extern int      log_level;
extern int      ioc_auto_recover;
extern int      ipfc_enable;
extern int      ipfc_mtu;

extern int      supported_fc4s;

#endif /* __BFAD_DRV_H__ */
