/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfad_im_2.6.c Linux driver IM module.
 * common code for Linux ver 2.6.*
 */

#include "bfad_drv.h"
#include "bfad_im.h"
#include "bfad_trcmod.h"

BFA_TRC_FILE(LDRV, IM_2_6);

BFAD_WORK_HANDLER(bfad_im_itnim_work_handler);

struct Scsi_Host *
bfad_os_scsi_host_alloc(struct bfad_im_port_s *im_port, struct bfad_s *bfad)
{
	struct scsi_host_template *sht; 

	if (im_port->port->pvb_type == BFAD_PORT_PHYS_BASE)
		sht = &bfad_im_scsi_host_template;
	else
		sht = &bfad_im_vport_template;

	sht->sg_tablesize = bfad->cfg_data.io_max_sge;

	return scsi_host_alloc(sht, sizeof(unsigned long));
}

void
bfad_os_scsi_host_free(struct bfad_s *bfad, struct bfad_im_port_s *im_port)
{
	flush_workqueue(bfad->im->drv_workq);
	bfad_im_scsi_host_free(im_port->bfad, im_port);
	bfad_im_port_clean(im_port);
	kfree(im_port);
}

int
bfad_os_scsi_register(struct Scsi_Host *shost, struct bfad_im_port_s *im_port,
		      void *key)
{
	return 0;
}

int
bfad_os_fc_attach(struct Scsi_Host *shost, struct fc_function_template *fct)
{
	return 0;
}

int
bfad_os_reset_tgt(struct scsi_cmnd *cmnd)
{
	return 0;
}

void
bfad_os_destroy_workq(struct bfad_im_s *im)
{
	if (im && im->drv_workq) {
		destroy_workqueue(im->drv_workq);
		im->drv_workq = NULL;
	}
}

bfa_status_t
bfad_os_thread_workq(struct bfad_s *bfad)
{
	struct bfad_im_s      *im = bfad->im;

	bfa_trc(bfad, 0);
	snprintf(im->drv_workq_name, KOBJ_NAME_LEN, "bfad_wq_%d",
		 bfad->inst_no);
	im->drv_workq = create_singlethread_workqueue(im->drv_workq_name);
	if (!im->drv_workq)
		return BFA_STATUS_FAILED;

	return BFA_STATUS_OK;
}

struct bfad_itnim_data_s *
bfad_os_itnim_data(struct bfad_im_port_s *im_port, struct scsi_cmnd *cmnd)
{
	if (!cmnd || !cmnd->device)
		return NULL;
	return cmnd->device->hostdata;
}

/**
 * @brief
 * Scsi_Host template entry.
 *
 * Description:
 * OS entry point to adjust the queue_depths on a per-device basis.
 * Called once per device during the bus scan.
 * Return non-zero if fails.
 */
static int
bfad_im_slave_configure(struct scsi_device *sdev)
{
	if (sdev->tagged_supported)
		scsi_activate_tcq(sdev, bfa_lun_queue_depth);
	else
		scsi_deactivate_tcq(sdev, bfa_lun_queue_depth);

	return 0;
}

struct scsi_host_template bfad_im_scsi_host_template = {
	.module = THIS_MODULE,
	.name = BFAD_DRIVER_NAME,
	.info = bfad_im_info,
	.queuecommand = bfad_im_queuecommand,
	.eh_abort_handler = bfad_im_abort_handler,
	.eh_device_reset_handler = bfad_im_reset_lun_handler,
	.eh_bus_reset_handler = bfad_im_reset_bus_handler,

	.slave_alloc = bfad_im_slave_alloc,
	.slave_configure = bfad_im_slave_configure,
	.slave_destroy = bfad_im_slave_destroy,

	.this_id = -1,
	.sg_tablesize = BFAD_IO_MAX_SGE,
	.cmd_per_lun = 3,
	.use_clustering = ENABLE_CLUSTERING,
	.shost_attrs = bfad_im_host_attrs,
	.max_sectors = 0xFFFF,
};

struct scsi_host_template bfad_im_vport_template = {
	.module = THIS_MODULE,
	.name = BFAD_DRIVER_NAME,
	.info = bfad_im_info,
	.queuecommand = bfad_im_queuecommand,
	.eh_abort_handler = bfad_im_abort_handler,
	.eh_device_reset_handler = bfad_im_reset_lun_handler,
	.eh_bus_reset_handler = bfad_im_reset_bus_handler,

	.slave_alloc = bfad_im_slave_alloc,
	.slave_configure = bfad_im_slave_configure,
	.slave_destroy = bfad_im_slave_destroy,

	.this_id = -1,
	.sg_tablesize = BFAD_IO_MAX_SGE,
	.cmd_per_lun = 3,
	.use_clustering = ENABLE_CLUSTERING,
	.shost_attrs = bfad_im_vport_attrs,
	.max_sectors = 0xFFFF,
};

void
bfad_im_probe_post(struct bfad_im_s *im)
{
	flush_workqueue(im->drv_workq);
}

bfa_status_t
bfad_im_module_init(void)
{
	bfad_im_scsi_transport_template =
		fc_attach_transport(&bfad_im_fc_function_template);
	if (!bfad_im_scsi_transport_template)
		return BFA_STATUS_ENOMEM;

	return BFA_STATUS_OK;
}

void
bfad_im_module_exit(void)
{
	if (bfad_im_scsi_transport_template)
		fc_release_transport(bfad_im_scsi_transport_template);
}

void
bfad_os_transp_templ(struct Scsi_Host *sh, struct scsi_transport_template *stt)
{
	sh->transportt = stt;
}

int
bfad_os_im_dma_map(struct bfad_s *bfad, struct scsi_cmnd *cmnd)
{
	int             sg_cnt, rc = 0;

	if (scsi_sg_count(cmnd)) {
		sg_cnt = bfad_os_dma_map_sg((struct device *)&bfad->pcidev->dev,
					(struct scatterlist *)scsi_sglist(cmnd),
					scsi_sg_count(cmnd),
					cmnd->sc_data_direction);
		if (sg_cnt == 0 || sg_cnt > bfad->cfg_data.io_max_sge) {
			/*
			 * log info and return
			 */
			printk(KERN_WARNING "dma_map_sg failure\n");
			rc = 1;
		}

	} else if (scsi_bufflen(cmnd)) {
		cmnd->SCp.dma_handle =
			bfad_os_dma_map_single((struct device *)&bfad->pcidev->
					       dev, scsi_sglist(cmnd),
					       scsi_bufflen(cmnd),
					       cmnd->sc_data_direction);
	}

	return rc;
}

void
bfad_os_im_dma_unmap(struct bfad_s *bfad, struct scsi_cmnd *cmnd)
{
	if (scsi_sg_count(cmnd))
		bfad_os_dma_unmap_sg((struct device *)&bfad->pcidev->dev,
				     scsi_sglist(cmnd), scsi_sg_count(cmnd),
				     cmnd->sc_data_direction);
	else if (scsi_bufflen(cmnd))
		bfad_os_dma_unmap_single((struct device *)&bfad->pcidev->dev,
					 cmnd->SCp.dma_handle,
					 scsi_bufflen(cmnd),
					 cmnd->sc_data_direction);
}

void
bfad_os_itnim_alloc(struct bfad_itnim_s *itnim_drv)
{
	BFAD_INIT_WORK(itnim_drv, itnim_work, bfad_im_itnim_work_handler);
}

void
bfad_os_itnim_process(struct bfad_itnim_s *itnim_drv)
{
	struct bfad_im_s      *im = itnim_drv->im;

	if (itnim_drv->queue_work)
		bfad_os_queue_work(im->drv_workq, &itnim_drv->itnim_work);
}

void
bfad_os_ramp_up_qdepth(struct bfad_itnim_s *itnim, struct scsi_device *sdev)
{
	struct scsi_device *tmp_sdev;

	if (((jiffies - itnim->last_ramp_up_time) >
		BFA_QUEUE_FULL_RAMP_UP_TIME * HZ) &&
		((jiffies - itnim->last_queue_full_time) >
		BFA_QUEUE_FULL_RAMP_UP_TIME * HZ)) {
		shost_for_each_device(tmp_sdev, sdev->host) {
			if (bfa_lun_queue_depth > tmp_sdev->queue_depth) {
				if (tmp_sdev->id != sdev->id)
					continue;
				if (tmp_sdev->ordered_tags)
					scsi_adjust_queue_depth(tmp_sdev,
						MSG_ORDERED_TAG,
						tmp_sdev->queue_depth + 1);
				else
					scsi_adjust_queue_depth(tmp_sdev,
						MSG_SIMPLE_TAG,
						tmp_sdev->queue_depth + 1);

				itnim->last_ramp_up_time = jiffies;
			}
		}
	}
}

void
bfad_os_handle_qfull(struct bfad_itnim_s *itnim, struct scsi_device *sdev)
{
	struct scsi_device *tmp_sdev;

	itnim->last_queue_full_time = jiffies;

	shost_for_each_device(tmp_sdev, sdev->host) {
		if (tmp_sdev->id != sdev->id)
			continue;
		scsi_track_queue_full(tmp_sdev, tmp_sdev->queue_depth - 1);
	}
}

void
bfad_os_scsi_state_changed(struct Scsi_Host *shost, bfad_itnim_t *itnim)
{
	return;
}
