/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @brief ms_sm_info Port/Fabric MS State Machine
 *
 */
#include <bfa.h>
#include <bfa_svc.h>
#include <fcs/bfa_fcs_port.h>
#include <fcs/bfa_fcs_port.h>
#include "fcs_port.h"
#include "fcs_rport.h"
#include "port_priv.h"
#include "fcs_trcmod.h"
#include "fcs_fcxp.h"

BFA_TRC_FILE(FCS, MS);

#define BFA_FCS_MS_CMD_MAX_RETRIES  2
/*
 * forward declarations
 */
static void     bfa_fcs_port_ms_send_plogi(void *ms_cbarg,
					   bfa_fcxp_t *fcxp_alloced);
static void     bfa_fcs_port_ms_timeout(void *arg);
static void     bfa_fcs_port_ms_plogi_response(void *fcsarg,
					       bfa_fcxp_t *fcxp,
					       void *cbarg,
					       bfa_status_t req_status,
					       uint32_t rsp_len,
					       uint32_t resid_len,
					       fchs_t *rsp_fchs);

static void 	bfa_fcs_port_ms_send_gmal(void *ms_cbarg, 
                                          bfa_fcxp_t *fcxp_alloced);
static void     bfa_fcs_port_ms_gmal_response(void *fcsarg, 
					       bfa_fcxp_t *fcxp,
					       void *cbarg,
					       bfa_status_t req_status,
					       uint32_t rsp_len,
					       uint32_t resid_len,
					       fchs_t *rsp_fchs);
/**
 * @dg fcs_ms_sm FCS MS state machine
 * @{
 */

/**
 *  MS State Machine events
 */
typedef enum {
	MSSM_EVENT_PORT_ONLINE = 1,
	MSSM_EVENT_PORT_OFFLINE = 2,
	MSSM_EVENT_PLOGI_SENT = 3,
	MSSM_EVENT_RSP_OK = 4,
	MSSM_EVENT_RSP_ERROR = 5,
	MSSM_EVENT_TIMEOUT = 6,
	MSSM_EVENT_GMAL_SENT = 7,
} port_ms_event_t;

static void     bfa_fcs_port_ms_sm_offline(bfa_fcs_port_ms_t *ms,
					   port_ms_event_t event);
static void     bfa_fcs_port_ms_sm_plogi_sending(bfa_fcs_port_ms_t *ms,
						 port_ms_event_t event);
static void     bfa_fcs_port_ms_sm_plogi(bfa_fcs_port_ms_t *ms,
					 port_ms_event_t event);
static void     bfa_fcs_port_ms_sm_plogi_retry(bfa_fcs_port_ms_t *ms,
					       port_ms_event_t event);
static void     bfa_fcs_port_ms_sm_gmal_sending(bfa_fcs_port_ms_t *ms,
						 port_ms_event_t event);
static void     bfa_fcs_port_ms_sm_gmal(bfa_fcs_port_ms_t *ms,
					 port_ms_event_t event);
static void     bfa_fcs_port_ms_sm_gmal_retry(bfa_fcs_port_ms_t *ms,
					       port_ms_event_t event);
static void     bfa_fcs_port_ms_sm_online(bfa_fcs_port_ms_t *ms,
					  port_ms_event_t event);
/**
 * @brief
 * 		Start in offline state - awaiting NS to send start.
 */
static void
bfa_fcs_port_ms_sm_offline(bfa_fcs_port_ms_t *ms, port_ms_event_t event)
{
	bfa_trc(ms->port->fcs, ms->port->port_cfg.pwwn);
	bfa_trc(ms->port->fcs, event);

	switch (event) {
	case MSSM_EVENT_PORT_ONLINE:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_plogi_sending);
		bfa_fcs_port_ms_send_plogi(ms, NULL);
		break;

	case MSSM_EVENT_PORT_OFFLINE:
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_ms_sm_plogi_sending(bfa_fcs_port_ms_t *ms, port_ms_event_t event)
{
	bfa_trc(ms->port->fcs, ms->port->port_cfg.pwwn);
	bfa_trc(ms->port->fcs, event);

	switch (event) {
	case MSSM_EVENT_PLOGI_SENT:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_plogi);
		break;

	case MSSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_offline);
		bfa_fcxp_walloc_cancel(BFA_FCS_GET_HAL_FROM_PORT(ms->port),
					   &ms->fcxp_wqe);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_ms_sm_plogi(bfa_fcs_port_ms_t *ms, port_ms_event_t event)
{
	bfa_trc(ms->port->fcs, ms->port->port_cfg.pwwn);
	bfa_trc(ms->port->fcs, event);

	switch (event) {
	case MSSM_EVENT_RSP_ERROR:
		/*
		 * Start timer for a delayed retry
		 */
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_plogi_retry);
		bfa_timer_start(BFA_FCS_GET_HAL_FROM_PORT(ms->port),
				    &ms->timer, bfa_fcs_port_ms_timeout, ms,
				    BFA_FCS_RETRY_TIMEOUT);
		break;

	case MSSM_EVENT_RSP_OK:
		/*
		 * since plogi is done, now invoke MS related sub-modules
		 */
		bfa_fcs_port_fdmi_online(ms);

		/**
		 * if this is a Vport, go to online state.
		 */
		if (ms->port->vport) {
			bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_online);
			break;
		}

		/* 
         * For a base port we need to get the 
		 * switch's IP address.
         */
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_gmal_sending);
		bfa_fcs_port_ms_send_gmal(ms, NULL);
		break;

	case MSSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_offline);
		bfa_fcxp_discard(ms->fcxp);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_ms_sm_plogi_retry(bfa_fcs_port_ms_t *ms, port_ms_event_t event)
{
	bfa_trc(ms->port->fcs, ms->port->port_cfg.pwwn);
	bfa_trc(ms->port->fcs, event);

	switch (event) {
	case MSSM_EVENT_TIMEOUT:
		/*
		 * Retry Timer Expired. Re-send
		 */
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_plogi_sending);
		bfa_fcs_port_ms_send_plogi(ms, NULL);
		break;

	case MSSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_offline);
		bfa_timer_stop(&ms->timer);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_ms_sm_online(bfa_fcs_port_ms_t *ms, port_ms_event_t event)
{
	bfa_trc(ms->port->fcs, ms->port->port_cfg.pwwn);
	bfa_trc(ms->port->fcs, event);

	switch (event) {
	case MSSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_offline);
		/*
		 * now invoke MS related sub-modules
		 */
		bfa_fcs_port_fdmi_offline(ms);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_ms_sm_gmal_sending(bfa_fcs_port_ms_t *ms, port_ms_event_t event)
{
	bfa_trc(ms->port->fcs, ms->port->port_cfg.pwwn);
	bfa_trc(ms->port->fcs, event);

	switch (event) {
	case MSSM_EVENT_GMAL_SENT:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_gmal);
		break;

	case MSSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_offline);
		bfa_fcxp_walloc_cancel(BFA_FCS_GET_HAL_FROM_PORT(ms->port),
					   &ms->fcxp_wqe);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_ms_sm_gmal(bfa_fcs_port_ms_t *ms, port_ms_event_t event)
{
	bfa_trc(ms->port->fcs, ms->port->port_cfg.pwwn);
	bfa_trc(ms->port->fcs, event);

	switch (event) {
	case MSSM_EVENT_RSP_ERROR:
		/*
		 * Start timer for a delayed retry
		 */
		if (ms->retry_cnt++ < BFA_FCS_MS_CMD_MAX_RETRIES) { 
			bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_gmal_retry);
			bfa_timer_start(BFA_FCS_GET_HAL_FROM_PORT(ms->port),
						&ms->timer, bfa_fcs_port_ms_timeout, ms,
						BFA_FCS_RETRY_TIMEOUT);
		} else {
			bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_online);
			ms->retry_cnt = 0;
		}
		break;

	case MSSM_EVENT_RSP_OK:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_online);
		break;

	case MSSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_offline);
		bfa_fcxp_discard(ms->fcxp);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_ms_sm_gmal_retry(bfa_fcs_port_ms_t *ms, port_ms_event_t event)
{
	bfa_trc(ms->port->fcs, ms->port->port_cfg.pwwn);
	bfa_trc(ms->port->fcs, event);

	switch (event) {
	case MSSM_EVENT_TIMEOUT:
		/*
		 * Retry Timer Expired. Re-send
		 */
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_gmal_sending);
		bfa_fcs_port_ms_send_gmal(ms, NULL);
		break;

	case MSSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_offline);
		bfa_timer_stop(&ms->timer);
		break;

	default:
		bfa_assert(0);
	}
}
/**
 * @dg ms_pvt MS local functions
 * @{
 */

static void
bfa_fcs_port_ms_send_gmal(void *ms_cbarg, bfa_fcxp_t *fcxp_alloced)
{
	bfa_fcs_port_ms_t *ms = ms_cbarg;
	bfa_fcs_port_t *port = ms->port;
	fchs_t          fchs;
	int             len;
	bfa_fcxp_t *fcxp;

	bfa_trc(port->fcs, port->pid);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_fcxp_alloc_wait(port->fcs->bfa, &ms->fcxp_wqe,
					bfa_fcs_port_ms_send_gmal, ms);
		return;
	}
	ms->fcxp = fcxp;

	len = fc_gmal_req_build(&fchs, bfa_fcxp_get_reqbuf(fcxp),
			     bfa_fcs_port_get_fcid(port), 
				 bfa_fcs_port_get_fabric_name(port));

	bfa_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs,
			  bfa_fcs_port_ms_gmal_response, (void *)ms,
			  FC_MAX_PDUSZ, FC_RA_TOV);

	bfa_sm_send_event(ms, MSSM_EVENT_GMAL_SENT);
}

static void
bfa_fcs_port_ms_gmal_response(void *fcsarg, bfa_fcxp_t *fcxp, void *cbarg,
			       bfa_status_t req_status, uint32_t rsp_len,
			       uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_port_ms_t *ms = (bfa_fcs_port_ms_t *) cbarg;
	bfa_fcs_port_t *port = ms->port;
	ct_hdr_t       *cthdr = NULL;
	fcgs_gmal_resp_t *gmal_resp;
	fcgs_gmal_entry_t *gmal_entry;
	uint32_t		num_entries;
	

	bfa_trc(port->fcs, req_status);
	bfa_trc(port->fcs, port->port_cfg.pwwn);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(port->fcs, req_status);
		bfa_sm_send_event(ms, MSSM_EVENT_RSP_ERROR);
		return;
	}

	cthdr = (ct_hdr_t *) BFA_FCXP_RSP_PLD(fcxp);
	cthdr->cmd_rsp_code = bfa_os_ntohs(cthdr->cmd_rsp_code);

	if (cthdr->cmd_rsp_code == CT_RSP_ACCEPT) {
		gmal_resp = (fcgs_gmal_resp_t *)(cthdr +1);
		
		num_entries = bfa_os_ntohl(gmal_resp->ms_len);
		if (num_entries == 0) {
			bfa_sm_send_event(ms, MSSM_EVENT_RSP_ERROR);
			return;
		}
		/*
		* The response could contain multiple Entries.
		* Entries for SNMP interface, etc.
		* We look for the entry with a telnet prefix.
		* First "telnet://" entry refers to IP addr
		*/

		gmal_entry = (fcgs_gmal_entry_t *)gmal_resp->ms_ma;
		while (num_entries > 0) {
			if (strncmp(gmal_entry->prefix,
						CT_GMAL_RESP_PREFIX_TELNET,
						sizeof (gmal_entry->prefix)) == 0) {

				/* if the IP address is terminating with a '/', remove it.
                 * Byte 0 consists of the length of the string.
                 */
				if (gmal_resp->ms_ma[gmal_resp->ms_ma[0]] == '/')
					gmal_resp->ms_ma[gmal_resp->ms_ma[0]] = 0;

					/* copy IP Address to fabric */
					strncpy(bfa_fcs_port_get_fabric_ipaddr(port),
						gmal_entry->ip_addr,
						BFA_FCS_FABRIC_IPADDR_SZ);
					break;
			} else {
				--num_entries;
				++gmal_entry;
			}
		}

		bfa_sm_send_event(ms, MSSM_EVENT_RSP_OK);
		return;
	}

	bfa_trc(port->fcs, cthdr->reason_code);
	bfa_trc(port->fcs, cthdr->exp_code);
	bfa_sm_send_event(ms, MSSM_EVENT_RSP_ERROR);
}
/**
 * @dg ms_pvt MS local functions
 * @{
 */

static void
bfa_fcs_port_ms_send_plogi(void *ms_cbarg, bfa_fcxp_t *fcxp_alloced)
{
	bfa_fcs_port_ms_t *ms = ms_cbarg;
	bfa_fcs_port_t *port = ms->port;
	fchs_t          fchs;
	int             len;
	bfa_fcxp_t *fcxp;

	bfa_trc(port->fcs, port->pid);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		port->stats.ms_plogi_alloc_wait++;
		bfa_fcxp_alloc_wait(port->fcs->bfa, &ms->fcxp_wqe,
					bfa_fcs_port_ms_send_plogi, ms);
		return;
	}
	ms->fcxp = fcxp;

	len = fc_plogi_build(&fchs, bfa_fcxp_get_reqbuf(fcxp),
			     bfa_os_hton3b(FC_MGMT_SERVER),
			     bfa_fcs_port_get_fcid(port), 0,
			     port->port_cfg.pwwn, port->port_cfg.nwwn,
				 bfa_pport_get_maxfrsize(port->fcs->bfa));

	bfa_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs,
			  bfa_fcs_port_ms_plogi_response, (void *)ms,
			  FC_MAX_PDUSZ, FC_RA_TOV);

	port->stats.ms_plogi_sent++;
	bfa_sm_send_event(ms, MSSM_EVENT_PLOGI_SENT);
}

static void
bfa_fcs_port_ms_plogi_response(void *fcsarg, bfa_fcxp_t *fcxp, void *cbarg,
			       bfa_status_t req_status, uint32_t rsp_len,
			       uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_port_ms_t *ms = (bfa_fcs_port_ms_t *) cbarg;
	bfa_fcs_port_t *port = ms->port;
	fc_els_cmd_t   *els_cmd;
	fc_ls_rjt_t    *ls_rjt;

	bfa_trc(port->fcs, req_status);
	bfa_trc(port->fcs, port->port_cfg.pwwn);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		port->stats.ms_plogi_rsp_err++;
		bfa_trc(port->fcs, req_status);
		bfa_sm_send_event(ms, MSSM_EVENT_RSP_ERROR);
		return;
	}

	els_cmd = (fc_els_cmd_t *) BFA_FCXP_RSP_PLD(fcxp);

	switch (els_cmd->els_code) {

	case FC_ELS_ACC:
		if (rsp_len < sizeof(fc_plogi_t)) {
			bfa_trc(port->fcs, rsp_len);
			port->stats.ms_plogi_acc_err++;
			bfa_sm_send_event(ms, MSSM_EVENT_RSP_ERROR);
			break;
		}
		port->stats.ms_plogi_accepts++;
		bfa_sm_send_event(ms, MSSM_EVENT_RSP_OK);
		break;

	case FC_ELS_LS_RJT:
		ls_rjt = (fc_ls_rjt_t *) BFA_FCXP_RSP_PLD(fcxp);

		bfa_trc(port->fcs, ls_rjt->reason_code);
		bfa_trc(port->fcs, ls_rjt->reason_code_expl);

		port->stats.ms_rejects++;
		bfa_sm_send_event(ms, MSSM_EVENT_RSP_ERROR);
		break;

	default:
		port->stats.ms_plogi_unknown_rsp++;
		bfa_trc(port->fcs, els_cmd->els_code);
		bfa_sm_send_event(ms, MSSM_EVENT_RSP_ERROR);
	}
}

static void
bfa_fcs_port_ms_timeout(void *arg)
{
	bfa_fcs_port_ms_t *ms = (bfa_fcs_port_ms_t *) arg;

	ms->port->stats.ms_timeouts++;
	bfa_sm_send_event(ms, MSSM_EVENT_TIMEOUT);
}

/**
* @brief : Public Functions
 */
void
bfa_fcs_port_ms_init(bfa_fcs_port_t *port)
{
	bfa_fcs_port_ms_t *ms = BFA_FCS_GET_MS_FROM_PORT(port);

	ms->port = port;
	bfa_sm_set_state(ms, bfa_fcs_port_ms_sm_offline);

	/*
	 * Invoke init routines of sub modules.
	 */
	bfa_fcs_port_fdmi_init(ms);
}

void
bfa_fcs_port_ms_offline(bfa_fcs_port_t *port)
{
	bfa_fcs_port_ms_t *ms = BFA_FCS_GET_MS_FROM_PORT(port);

	ms->port = port;
	bfa_sm_send_event(ms, MSSM_EVENT_PORT_OFFLINE);
}

void
bfa_fcs_port_ms_online(bfa_fcs_port_t *port)
{
	bfa_fcs_port_ms_t *ms = BFA_FCS_GET_MS_FROM_PORT(port);

	ms->port = port;
	bfa_sm_send_event(ms, MSSM_EVENT_PORT_ONLINE);
}
