/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <bfa.h>
#include <defs/bfa_defs_pci.h>
#include <cs/bfa_debug.h>
#include <bfa_iocfc.h>

#define DEF_CFG_NUM_CQS             4
#define DEF_CFG_NUM_IOIM_REQS       (512)
#define DEF_CFG_NUM_TSKIM_REQS      128
#define DEF_CFG_NUM_FCXP_REQS       64
#define DEF_CFG_NUM_UF_BUFS         64
#define DEF_CFG_NUM_RPORTS          1024
#define DEF_CFG_NUM_ITNIMS          (DEF_CFG_NUM_RPORTS)
#define DEF_CFG_NUM_ITNTMS          256

#define DEF_CFG_NUM_SGPGS           512
#define DEF_CFG_NUM_REQQ_ELEMS      256
#define DEF_CFG_NUM_RSPQ_ELEMS      64
#define DEF_CFG_NUM_SBOOT_TGTS      16
#define DEF_CFG_NUM_SBOOT_LUNS      16

/**
 * @todo
 * Move this initialization to function. Error prone since fields
 * can be inserted, rearranged etc ...
 */
static bfa_ioc_cfg_t   hal_cfg_default = {
	{                               /*!< default firmware configuration  */
		DEF_CFG_NUM_CQS,        /*!< number of req/rsp queue pairs   */
		0,			/*!< reserved                        */
		DEF_CFG_NUM_ITNIMS,     /*!< i-t nexus (fcp sessions)        */
		DEF_CFG_NUM_IOIM_REQS,  /*!< max number of IO requests       */
		DEF_CFG_NUM_TSKIM_REQS, /*!< max number task mgmt commands   */
		DEF_CFG_NUM_FCXP_REQS,  /*!< max unassisted FC exchanges     */
		DEF_CFG_NUM_UF_BUFS,    /*!< max number of unsol recv frames */
		DEF_CFG_NUM_RPORTS,     /*!< max remote ports                */
		DEF_CFG_NUM_ITNTMS,     /*!< scsi target sessions            */
	},
	DEF_CFG_NUM_REQQ_ELEMS,         /*! request queue num elements       */
	DEF_CFG_NUM_RSPQ_ELEMS,         /*! comp queue number of elements    */
	DEF_CFG_NUM_SGPGS,              /*! default number of SG pages       */
	DEF_CFG_NUM_SBOOT_TGTS,         /*! bios only: max targets           */
	DEF_CFG_NUM_SBOOT_LUNS,         /*! bios only: max luns              */
	0,	                        /*! min cfg flag                     */
	0,                              /*! pad                              */
	BFA_FCPIM_PATH_TOV_DEF,     /*!< offline TOV in non-mpio mode    */
	BFA_FALSE,                      /*! mpio_mode is OFF by default      */
	BFA_FALSE,                      /*! delay failed IO completions	*/
};

/**
 * @ig hal_api
 * @{
 */

/**
 * @brief
 * Return the list of PCI vendor/device id lists supported by this
 * HAL instance.
 */
void
bfa_get_pciids(bfa_pciid_t **pciids, int *npciids)
{
	static bfa_pciid_t __pciids[] = {
		{BFA_PCI_VENDOR_ID_BROCADE, BFA_PCI_DEVICE_ID_FC_8G2P},
		{BFA_PCI_VENDOR_ID_BROCADE, BFA_PCI_DEVICE_ID_FC_8G1P},
	};

	*npciids = sizeof(__pciids) / sizeof(__pciids[0]);
	*pciids = __pciids;
}

/**
 * @brief
 * Use this function query the default bfa_ioc_cfg_t value (compiled into
 * BFA layer). The OS driver can then turn back and overwrite entries that
 * have been configured by the user.
 *
 * @param[in] cfg - pointer to bfa_ioc_cfg_t
 *
 * @return
 *	void
 *
 * Special Considerations:
 * 	@note
 */
void
bfa_cfg_get_default(bfa_ioc_cfg_t *cfg)
{
	*cfg = hal_cfg_default;
}

void
bfa_cfg_get_min(bfa_ioc_cfg_t *cfg)
{
	*cfg = hal_cfg_default;
	cfg->fwcfg.num_itnims     = BFA_ITNIM_MIN;
	cfg->fwcfg.num_ioim_reqs  = BFA_IOIM_MIN;
	cfg->fwcfg.num_tskim_reqs = BFA_TSKIM_MIN;
	cfg->fwcfg.num_fcxp_reqs  = BFA_FCXP_MIN;
	cfg->fwcfg.num_uf_bufs    = BFA_UF_MIN;
	cfg->fwcfg.num_rports     = BFA_RPORT_MIN;
	cfg->num_sgpgs            = BFA_SGPG_MIN;
	cfg->num_reqq_elems       = BFA_REQQ_NELEMS_MIN;
	cfg->num_rspq_elems       = BFA_RSPQ_NELEMS_MIN;
	cfg->min_cfg              = 1;
	cfg->fcxp_sbuf			  = BFA_TRUE;
}

/**
 * @brief
 * Use this function query the memory requirement of the BFA library.
 * This function needs to be called before bfa_attach() to get the
 * memory required of the BFA layer for a given driver configuration.
 *
 * This call will fail, if the cap is out of range compared to pre-defined
 * values within the BFA library
 *
 * @param[in] cfg - 	pointer to bfa_ioc_cfg_t. Driver layer should indicate
 * 			its configuration in this structure.
 *			The default values for bfa_ioc_cfg_t can be fetched
 *			using bfa_cfg_get_default() API.
 *
 * 			If cap's boundary check fails, the library will use
 *			the default bfa_cap_t values (and log a warning msg).
 *
 * @param[out] meminfo - pointer to bfa_meminfo_t. This content
 * 			indicates the memory type (see bfa_mem_type_t) and
 *			amount of memory required.
 *
 *			Driver should allocate the memory, populate the
 *			starting address for each block and provide the same
 *			structure as input parameter to bfa_attach() call.
 *
 * @return void
 *
 * Special Considerations: @note
 */
void
bfa_cfg_get_meminfo(bfa_ioc_cfg_t *cfg, bfa_meminfo_t *meminfo)
{
	int             i;
	uint32_t        km_len = 0, dm_len = 0;

	bfa_assert((cfg != NULL) && (meminfo != NULL));

	bfa_os_memset((void *)meminfo, 0, sizeof(bfa_meminfo_t));
	meminfo->meminfo[BFA_MEM_TYPE_KVA - 1].mem_type =
		BFA_MEM_TYPE_KVA;
	meminfo->meminfo[BFA_MEM_TYPE_DMA - 1].mem_type =
		BFA_MEM_TYPE_DMA;

	bfa_iocfc_meminfo(cfg, &km_len, &dm_len);

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->meminfo(cfg, &km_len, &dm_len);

	meminfo->meminfo[BFA_MEM_TYPE_KVA - 1].mem_len = km_len;
	meminfo->meminfo[BFA_MEM_TYPE_DMA - 1].mem_len = dm_len;
}

/**
 * @brief
 * Use this function to do attach the driver instance with the BFA HAL
 * library. This function will not trigger any HW initialization
 * process (which will be done in bfa_init() call)
 *
 * This call will fail, if the cap is out of range compared to
 * pre-defined values within the BFA library
 *
 * @param[out]	bfa	Pointer to bfa_t.
 * @param[in]	bfad 	Opaque handle back to the driver's IOC structure
 * @param[in]	cfg	Pointer to bfa_ioc_cfg_t. Should be same structure
 * 			that was used in bfa_cfg_get_meminfo().
 * @param[in] 	meminfo Pointer to bfa_meminfo_t. The driver should
 * 			use the bfa_cfg_get_meminfo() call to
 * 			find the memory blocks required, allocate the
 * 			required memory and provide the starting addresses.
 * @param[in] 	pcidev	pointer to bfa_pcidev_t
 *
 * @return
 * void
 *
 * Special Considerations:
 *
 * @note
 *
 */
void
bfa_attach(bfa_t *bfa, void *bfad, bfa_ioc_cfg_t *cfg,
	       bfa_meminfo_t *meminfo, bfa_pcidev_t *pcidev)
{
	int             i;
	bfa_mem_elem_t *melem;

	bfa->pcidev = pcidev;
	bfa->fcs = BFA_FALSE;

	bfa_assert((cfg != NULL) && (meminfo != NULL));

	/**
	 * initialize all memory pointers for iterative allocation
	 */
	for (i = 0; i < BFA_MEM_TYPE_MAX; i++) {
		melem = meminfo->meminfo + i;
		melem->kva_curp = melem->kva;
		melem->dma_curp = melem->dma;
	}

	bfa_iocfc_attach(bfa, bfad, cfg, meminfo, pcidev);

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->attach(bfa, bfad, cfg, meminfo, pcidev);
}

/**
 * @brief
 * Use this function to delete a HAL IOC. IOC should be stopped (by
 * calling bfa_stop()) before this function call.
 *
 * @param[in] bfa - pointer to bfa_t.
 *
 * @return
 * void
 *
 * Special Considerations:
 *
 * @note
 */
void
bfa_detach(bfa_t *bfa)
{
	int	i;

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->detach(bfa);

	bfa_iocfc_detach(bfa);
}

/**
 * @brief Initialize tracing in BFA HAL module.
 */
void
bfa_init_trc(bfa_t *bfa, bfa_trc_mod_t *trcmod)
{
	bfa->trcmod = trcmod;
}

/**
 * @brief Initialize logging in BFA HAL module.
 */
void
bfa_init_log(bfa_t *bfa, bfa_log_mod_t *logmod)
{
	bfa->logm = logmod;
}

/**
 * @brief Initialize AEN in BFA HAL module.
 */
void
bfa_init_aen(bfa_t *bfa, bfa_aen_t *aen)
{
	bfa->aen = aen;
}

void
bfa_init_plog(bfa_t *bfa, bfa_plog_t *plog)
{
	bfa->plog = plog;
	bfa_plog_init(plog);
}

/**
 * @brief
 * Initialize IOC.
 *
 * This function will return immediately, when the IOC initialization is
 * completed, the bfa_cb_init() will be called.
 *
 * @param[in]	bfa	HAL instance
 *
 * @return void
 *
 * Special Considerations:
 *
 * @note
 * When this function returns, the driver should register the interrupt service
 * routine(s) and enable the device interrupts. If this is not done,
 * bfa_cb_init() will never get called
 */
void
bfa_init(bfa_t *bfa)
{
	bfa_iocfc_init(bfa);
}

/**
 * @brief
 * Use this function initiate the IOC configuration setup. This function
 * will return immediately.
 *
 * @param[in]	bfa	HAL instance
 *
 * @return None
 */
void
bfa_start(bfa_t *bfa)
{
	bfa_iocfc_start(bfa);
}

/**
 * @brief
 * Use this function quiese the IOC. This function will return immediately,
 * when the IOC is actually stopped, the bfa_cb_stop() will be called.
 *
 * @param[in] 	bfa - pointer to bfa_t.
 *
 * @return None
 *
 * Special Considerations:
 * bfa_cb_stop() could be called before or after bfa_stop() returns.
 *
 * @note
 * In case of any failure, we could handle it automatically by doing a
 * reset and then succeed the bfa_stop() call.
 */
void
bfa_stop(bfa_t *bfa)
{
	bfa_iocfc_stop(bfa);
}

void
bfa_comp_deq(bfa_t *bfa, bfa_q_t *comp_q)
{
	bfa_q_init(comp_q);
	bfa_q_mv(&bfa->comp_q, comp_q);
}

void
bfa_comp_process(bfa_t *bfa, bfa_q_t *comp_q)
{
	bfa_q_t        *qe;
	bfa_q_t        *qen;
	bfa_cb_qe_t   *hcb_qe;

	bfa_q_iter_safe(comp_q, qe, qen) {
		hcb_qe = (bfa_cb_qe_t *) qe;
		hcb_qe->cbfn(hcb_qe->cbarg, BFA_TRUE);
	}
}

void
bfa_comp_free(bfa_t *bfa, bfa_q_t *comp_q)
{
	bfa_q_t        *qe;
	bfa_cb_qe_t   *hcb_qe;

	while (!bfa_q_is_empty(comp_q)) {
		bfa_q_deq(comp_q, &qe);
		hcb_qe = (bfa_cb_qe_t *) qe;
		hcb_qe->cbfn(hcb_qe->cbarg, BFA_FALSE);
	}
}

void
bfa_attach_fcs(bfa_t *bfa)
{
	bfa->fcs = BFA_TRUE;
}

/**
 * @}
 */
