/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <bfa.h>
#include <bfi/bfi_cbreg.h>

BFA_TRC_FILE(HAL, DEBUG);

/**
 * @ig hal_debug_api
 * @{
 */

#define BFA_DBG_FWTRC_ENTS	(BFI_IOC_TRC_ENTS)
#define BFA_DBG_FWTRC_LEN	(BFA_DBG_FWTRC_ENTS * sizeof(bfa_trc_t) + \
					(sizeof(bfa_trc_mod_t) - 	\
					BFA_TRC_MAX * sizeof(bfa_trc_t)))
#define BFA_DBG_FWTRC_OFF(_fn)	(BFI_IOC_TRC_OFF + \
					BFA_DBG_FWTRC_LEN * (_fn))

/**
 * Memory required for saving firmware trace in case of crash
 */
void
bfa_debug_fwsave_sz(bfa_ioc_cfg_t *cfg, uint32_t *km_len,
			bfa_boolean_t auto_recover)
{
	if (auto_recover)
		*km_len += BFA_DBG_FWTRC_LEN;
}

/**
 * Initialize memory for saving firmware trace.
 */
void
bfa_debug_mem_claim(bfa_t *bfa, bfa_ioc_cfg_t *cfg,
			bfa_meminfo_t *meminfo, bfa_boolean_t auto_recover)
{
	struct bfa_ioc_s *ioc = &bfa->ioc;

	if (auto_recover) {
		ioc->dbg_fwsave     = bfa_meminfo_kva(meminfo);
		ioc->dbg_fwsave_len = BFA_DBG_FWTRC_LEN;
		bfa_meminfo_kva(meminfo) += BFA_DBG_FWTRC_LEN;
	} else {
		ioc->dbg_fwsave     = NULL;
		ioc->dbg_fwsave_len = 0;
	}
}

/**
 * Retrieve firmware trace information on IOC failure.
 */
bfa_status_t
bfa_debug_fwsave(bfa_t *bfa, void *trcdata, int *trclen)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	int		tlen;

	if (ioc->dbg_fwsave_len == 0)
		return BFA_STATUS_ENOFSAVE;

	tlen = *trclen;
	if (tlen > ioc->dbg_fwsave_len)
		tlen = ioc->dbg_fwsave_len;

	bfa_os_memcpy(trcdata, ioc->dbg_fwsave, tlen);
	*trclen = tlen;
	return BFA_STATUS_OK;
}

void
bfa_debug_ioc_save(bfa_t *bfa)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	int		tlen;

	if (ioc->dbg_fwsave_len) {
		tlen = ioc->dbg_fwsave_len;
		bfa_debug_fwtrc(bfa, ioc->dbg_fwsave, &tlen);
	}
}

/**
 * @brief
 * 		Fetch firmware trace data.
 *
 * @param[in]		bfa			HAL instance
 * @param[out]		trcdata		Firmware trace buffer
 * @param[in,out]	trclen		Firmware trace buffer len
 *
 * @retval BFA_STATUS_OK			Firmware trace is fetched.
 * @retval BFA_STATUS_INPROGRESS	Firmware trace fetch is in progress.
 */
bfa_status_t
bfa_debug_fwtrc(bfa_t *bfa, void *trcdata, int *trclen)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	uint32_t        pgnum;
	uint32_t        loff = BFA_DBG_FWTRC_OFF(bfa->pcidev->pci_func);
	int             i, tlen;
	uint32_t       *tbuf = trcdata, r32;

	bfa_trc(bfa, *trclen);

	pgnum = PSS_SMEM_PGNUM(loff);
	loff = PSS_SMEM_PGOFF(loff);
	bfa_reg_write(ioc->bfa_regs.host_page_num_fn, pgnum);

	tlen = *trclen;
	if (tlen > BFA_DBG_FWTRC_LEN)
		tlen = BFA_DBG_FWTRC_LEN;
	tlen /= sizeof(uint32_t);

	bfa_trc(bfa, tlen);

	for (i = 0; i < tlen; i++) {
		r32 = bfa_mem_read(ioc->bfa_regs.smem_page_start, loff);
		tbuf[i] = bfa_os_ntohl(r32);
		loff += sizeof(uint32_t);

		/**
		 * handle page offset wrap around
		 */
		loff = PSS_SMEM_PGOFF(loff);
		if (loff == 0) {
			pgnum++;
			bfa_reg_write(ioc->bfa_regs.host_page_num_fn,
					  pgnum);
		}
	}
	bfa_reg_write(ioc->bfa_regs.host_page_num_fn, PSS_SMEM_PGNUM(0));
	bfa_trc(bfa, pgnum);

	*trclen = tlen * sizeof(uint32_t);
	return BFA_STATUS_OK;
}

/**
 * @}
 */
