/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <bfa.h>
#include <defs/bfa_defs_diag.h>
#include <cs/bfa_debug.h>
#include <bfi/bfi_boot.h>
#include <bfi/bfi_cbreg.h>
#include <bfi/bfi_diag.h>
#include <aen/bfa_aen.h>
#include <aen/bfal_aen_port.h>

BFA_TRC_FILE(HAL, DIAG);
BFA_MODULE(diag);

#define	BFA_DIAG_MEMTEST_TOV	50000	/* !< memtest timeout in msec */
#define BFA_DIAG_FWPING_TOV       	1000	/* msec */

static void     diag_fwping_complete(bfa_t *bfa, bfa_status_t status);

/**
 * @dg hal_diag_pvt HAL DIAG private functions
 * @{
 */

static void
bfa_diag_set_busy_status(bfa_diag_mod_t *diag)
{
	if (diag->block || diag->lb.lock)
		bfa_pport_busy(diag->bfa, BFA_TRUE);
	else
		bfa_pport_busy(diag->bfa, BFA_FALSE);
}

static void
bfa_cb_diag(void *cbarg, bfa_boolean_t complete)
{
	bfa_diag_mod_t *diag = cbarg;

	if (complete) {
		diag->cbfn(diag->cbarg, diag->status);
	} else {
		diag->block = 0;
		bfa_diag_set_busy_status(diag);
	}
}

static void
bfa_cb_diag_fwping(void *cbarg, bfa_boolean_t complete)
{
	bfa_diag_mod_t *diag = cbarg;

	if (complete)
		diag->fwping.cbfn(diag->fwping.cbarg, diag->fwping.status);
	else
		diag->fwping.lock = 0;
}

static void
bfa_cb_diag_tsensor(void *cbarg, bfa_boolean_t complete)
{
	bfa_diag_mod_t *diag = cbarg;

	if (complete)
		diag->tsensor.cbfn(diag->tsensor.cbarg, diag->tsensor.status);
	else
		diag->tsensor.lock = 0;
}

static void
bfa_cb_diag_sfpshow(void *cbarg, bfa_boolean_t complete)
{
	bfa_diag_mod_t *diag = cbarg;

	if (complete)
		diag->sfpshow.cbfn(diag->sfpshow.cbarg, diag->sfpshow.status);
	else
		diag->sfpshow.lock = 0;
}

static void
bfa_cb_diag_lb(void *cbarg, bfa_boolean_t complete)
{
	bfa_diag_mod_t *diag = cbarg;

	if (complete) {
		diag->lb.cbfn(diag->lb.cbarg, diag->lb.status);
	} else {
		diag->lb.lock = 0;
		bfa_diag_set_busy_status(diag);
	}
}

/**
 * @brief Update virtual and physical memory for Diag.
 *
 * @param[in] cfg - ioc config
 * @param[in] ndm_len - non-dma memory length
 * @param[in] dm_len - dma memory length
 */
static void
bfa_diag_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *ndm_len, uint32_t *dm_len)
{
	/* min driver doesn't need diag */
	if (cfg->min_cfg)
		return;

	*dm_len += BFI_DIAG_DMA_BUF_SZ;
}

/**
 * @brief Attach virtual and physical memory for Diag.
 *
 * @param[in] bfa - bfa structure
 * @param[in] bfad - driver structure
 * @param[in] cfg - ioc config
 * @param[in] meminfo - bfa memory info
 * @param[in] pcidev - pci device
 */
static void
bfa_diag_attach(bfa_t *bfa, void *bfad, bfa_ioc_cfg_t *cfg,
		    bfa_meminfo_t *meminfo, bfa_pcidev_t *pcidev)
{
	bfa_diag_mod_t *mod = BFA_DIAG_MOD(bfa);
	uint8_t        *dm_kva = NULL;
	uint64_t        dm_pa;

	mod->bfa = bfa;

	/* min driver doesn't need diag */
	if (cfg->min_cfg) {
		mod->dbuf_kva = NULL;
		mod->dbuf_pa = 0;
		return;
	}

	dm_kva = bfa_meminfo_dma_virt(meminfo);
	dm_pa = bfa_meminfo_dma_phys(meminfo);

	mod->dbuf_kva = dm_kva;
	mod->dbuf_pa = dm_pa;
	dm_kva += BFI_DIAG_DMA_BUF_SZ;
	dm_pa += BFI_DIAG_DMA_BUF_SZ;
	bfa_os_memset(mod->dbuf_kva, 0, BFI_DIAG_DMA_BUF_SZ);

	bfa_meminfo_dma_virt(meminfo) = dm_kva;
	bfa_meminfo_dma_phys(meminfo) = dm_pa;

	mod->block = 0;
	mod->cbfn = NULL;
	mod->cbarg = NULL;
	mod->result = NULL;
}

static void
bfa_diag_initdone(bfa_t *bfa)
{
}

static void
bfa_diag_detach(bfa_t *bfa)
{
}

static void
bfa_diag_start(bfa_t *bfa)
{
}

static void
bfa_diag_stop(bfa_t *bfa)
{
}

static void
bfa_diag_iocdisable(bfa_t *bfa)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);

	bfa_trc(bfa, diag->block);
	bfa_trc(bfa, diag->fwping.lock);
	bfa_trc(bfa, diag->sfpshow.lock);
	if (!diag->fwping.lock && !diag->sfpshow.lock &&
		!diag->lb.lock && !diag->tsensor.lock && !diag->block)
		return;

	if (diag->fwping.lock) {
		if (diag->fwping.timer_active) {
			bfa_timer_stop(&diag->fwping.timer);
			diag->fwping.timer_active = 0;
		}

		diag->fwping.status = BFA_STATUS_IOC_FAILURE;
		bfa_cb_queue(bfa, &diag->fwping.hcb_qe, bfa_cb_diag_fwping, diag);
	} else if (diag->sfpshow.lock) {
		diag->sfpshow.status = BFA_STATUS_IOC_FAILURE;
		bfa_cb_queue(bfa, &diag->sfpshow.hcb_qe, bfa_cb_diag_sfpshow, diag);
	} else if (diag->lb.lock) {
		diag->lb.status = BFA_STATUS_IOC_FAILURE;
		bfa_cb_queue(bfa, &diag->lb.hcb_qe, bfa_cb_diag_lb, diag);
	} else if (diag->tsensor.lock) {
		diag->tsensor.status = BFA_STATUS_IOC_FAILURE;
		bfa_cb_queue(bfa, &diag->tsensor.hcb_qe, bfa_cb_diag_tsensor, diag);
	} else if (diag->block) {
		if (diag->timer_active) {
			bfa_timer_stop(&diag->timer);
			diag->timer_active = 0;
		}

		diag->status = BFA_STATUS_IOC_FAILURE;
		bfa_cb_queue(bfa, &diag->hcb_qe, bfa_cb_diag, diag);
	}
}

/**
 * @brief Set diag request scatter gather element.
 *
 * @param[in] mod - bfa diag module
 * @param[in] sge - scatter gather element
 * @param[in] len - scatter gather len
 */
static void
bfa_diag_sge(bfa_diag_mod_t *mod, bfi_sge_t *sge, uint32_t len)
{
	bfi_addr_t      sga_zero = { {0} };

	sge->sg_len = len;
	sge->flags = BFI_SGE_DATA_LAST;
	bfa_dma_addr_set(sge->sga, mod->dbuf_pa);
	bfa_sge_to_be(sge);

	sge++;
	sge->sg_len = len;
	sge->flags = BFI_SGE_PGDLEN;
	sge->sga = sga_zero;
	bfa_sge_to_be(sge);
}

/*
 * Register access
 */
static bfa_status_t
diag_reg_offset_check(uint32_t offset, uint32_t len)
{
	uint8_t         area;

	/*
	 * check [16:15]
	 */
	area = (offset >> 15) & 0x3;
	if (area == 0) {
		/*
		 * PCIe core register
		 */
		if ((offset + len) > 0x7ffc)	/* 8k dwords or 32KB */
			return BFA_STATUS_EINVAL;
	} else if (area == 0x1) {
		/*
		 * CB 32 KB memory page
		 */
		if ((offset + len) > 0xfffc)	/* 8k dwords or 32KB */
			return BFA_STATUS_EINVAL;
	} else {
		/*
		 * CB register space 64KB
		 */
		if ((offset + len) > 0x1fffc)
			return BFA_STATUS_EINVAL;
	}
	return BFA_STATUS_OK;
}

static void
bfa_diag_memtest_done(void *cbarg)
{
	bfa_t      *bfa = cbarg;
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	bfa_diag_memtest_result_t *res = diag->result;
	uint32_t        loff = BFI_BOOT_MEMTEST_RES_ADDR;
	uint32_t        pgnum, pgoff, i;

	pgnum = PSS_SMEM_PGNUM(loff);
	pgoff = PSS_SMEM_PGOFF(loff);

	bfa_reg_write(ioc->bfa_regs.host_page_num_fn, pgnum);

	for (i = 0; i < (sizeof(bfa_diag_memtest_result_t) / sizeof(uint32_t));
	     i++) {
		/*
		 * read test result from smem
		 */
		*((uint32_t *) res + i) =
			bfa_mem_read(ioc->bfa_regs.smem_page_start, loff);
		loff += sizeof(uint32_t);
	}

	res->status = bfa_os_swap32(res->status);
	bfa_trc(bfa, res->status);

	if (res->status == BFI_BOOT_MEMTEST_RES_SIG) {
		diag->status = BFA_STATUS_OK;
	} else {
		diag->status = BFA_STATUS_MEMTEST_FAILED;

		res->addr = bfa_os_swap32(res->addr);
		res->exp = bfa_os_swap32(res->exp);
		res->act = bfa_os_swap32(res->act);
		res->err_status = bfa_os_swap32(res->err_status);
		res->err_addr = bfa_os_swap32(res->err_addr);

		bfa_trc(bfa, res->addr);
		bfa_trc(bfa, res->exp);
		bfa_trc(bfa, res->act);
		bfa_trc(bfa, res->err_status);
		bfa_trc(bfa, res->err_addr);
		bfa_trc(bfa, res->algo);
	}

	diag->timer_active = 0;
	bfa_cb_queue(bfa, &diag->hcb_qe, bfa_cb_diag, diag);
}

/*
 * Firmware ping
 */
static void
diag_fwping_timeout(void *cbarg)
{
	bfa_t      *bfa = cbarg;
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	uint32_t        vcidx = diag->fwping.vcidx;

	bfa_trc(bfa, vcidx);

	diag->fwping.result->status[vcidx] = BFA_STATUS_ETIMER;
	diag->fwping.timer_active = 0;

	diag_fwping_complete(bfa, BFA_STATUS_ETIMER);
}


static void
diag_fwping_send(bfa_t *bfa, uint8_t dma_test)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	bfi_diag_fwping_req_t *fwping_req;

	bfa_trc(bfa, diag->fwping.vcidx);

	if (dma_test) {
		uint32_t	i;

		bfa_trc(bfa, diag->dbuf_pa);
		
		/* fill DMA area with pattern */
		for (i=0; i<(BFI_DIAG_DMA_BUF_SZ >> 2); i++) {
			*((uint32_t *)diag->dbuf_kva + i) = diag->fwping.data;
		}

		fwping_req = bfa_reqq_next(bfa, BFA_REQQ_DIAG);

		/* Setup SG list */
		bfa_diag_sge(diag, fwping_req->sge, BFI_DIAG_DMA_BUF_SZ);
		/* Set up dma count */
		fwping_req->count = bfa_os_htonl(diag->fwping.count);
	} else {
		fwping_req = bfa_reqq_next(bfa, diag->fwping.vcidx);
		fwping_req->count = 0;
	}

	/* build host command */
	bfi_h2i_set(fwping_req->mh, BFI_MC_DIAG, BFI_DIAG_H2I_FWPING,
			bfa_ioc_lpuid(bfa));

	fwping_req->data = diag->fwping.data;
	fwping_req->qtag = (dma_test) ? BFA_REQQ_DIAG : diag->fwping.vcidx;
	fwping_req->dma_test = dma_test;

	/* ring door bell */
	bfa_reqq_produce(bfa, fwping_req->qtag);

	bfa_trc(bfa, bfa_reqq_pi(bfa, fwping_req->qtag));
	bfa_trc(bfa, bfa_reqq_ci(bfa, fwping_req->qtag));
}

static void
diag_fwping_complete(bfa_t *bfa, bfa_status_t status)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);

	bfa_trc(bfa, diag->fwping.vcidx);
	bfa_trc(bfa, status);

	diag->fwping.status = status;
	bfa_cb_queue(bfa, &diag->fwping.hcb_qe, bfa_cb_diag_fwping, diag);
}

static void
diag_fwping_nextvc(bfa_t *bfa)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	uint32_t i;
	/* Get next vc */
	diag->fwping.vcidx++;
	bfa_trc(bfa, diag->fwping.vcidx);
	if (diag->fwping.vcidx < BFI_IOC_MAX_CQS) {
		diag_fwping_send(bfa, 0);
	} else {
		/* Stop timer when we comp all vc
		 */
		if (diag->fwping.timer_active) {
			bfa_timer_stop(&diag->fwping.timer);
			diag->fwping.timer_active = 0;
		}
		// Check if all vc is good
		for (i=0; i<BFI_IOC_MAX_CQS; i++) {
			if (diag->fwping.result->status[i] != BFA_STATUS_OK) {
				diag_fwping_complete(bfa, BFA_STATUS_DATACORRUPTED);
				return;
			}
		}
		/* if all vc are good, do DMA test */
		diag_fwping_send(bfa, 1);
	}
}

static void
diag_fwping_comp(bfa_t *bfa, bfi_diag_fwping_rsp_t *diag_rsp)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	uint32_t        rsp_data = diag_rsp->data;
	uint8_t         rsp_qtag = diag_rsp->qtag;
	uint8_t         rsp_dma_status = diag_rsp->dma_status;

	bfa_trc(bfa, rsp_qtag);
	bfa_trc(bfa, rsp_data);
	bfa_trc(bfa, rsp_dma_status);

	if (diag->fwping.vcidx < BFI_IOC_MAX_CQS) {
		/* Check timer
		 */
		if (!diag->fwping.timer_active) {
			bfa_trc(bfa, diag->fwping.timer_active);
			return;
		}

		/* Check vc pattern
		 */
		if (diag->fwping.data == rsp_data) {
			diag->fwping.result->status[rsp_qtag] = BFA_STATUS_OK;
		} else {
			diag->fwping.result->status[rsp_qtag] =	BFA_STATUS_DATACORRUPTED;
		}
		diag->fwping.result->data[rsp_qtag] = rsp_data;

		/* send next vc */
		diag_fwping_nextvc(bfa);
	} else {
		if (rsp_dma_status == BFA_STATUS_OK) {
			uint32_t i, pat;
			pat = (diag->fwping.count & 0x1) ? ~(diag->fwping.data) : diag->fwping.data;
			/* Check vc pattern
			 */
			if (diag->fwping.data != rsp_data) {
				diag->fwping.result->dmastatus = BFA_STATUS_DATACORRUPTED;
				diag_fwping_complete(bfa, BFA_STATUS_DATACORRUPTED);
				return;
			}
			/* Check dma pattern
			 */
			for (i=0; i<(BFI_DIAG_DMA_BUF_SZ >> 2); i++) {
				if (*((uint32_t *)diag->dbuf_kva + i) != pat) {
					bfa_trc(bfa, i);
					bfa_trc(bfa, *((uint32_t *)diag->dbuf_kva + i));
					diag->fwping.result->dmastatus = BFA_STATUS_DATACORRUPTED;
					diag_fwping_complete(bfa, BFA_STATUS_DATACORRUPTED);
					return;
				}
			}
			diag->fwping.result->dmastatus = BFA_STATUS_OK;
			diag_fwping_complete(bfa, BFA_STATUS_OK);
		} else {
			diag_fwping_complete(bfa, BFA_STATUS_HDMA_FAILED);
		}
	}
}

/*
 * Loopback
 */

static void
diag_loopback_send(bfa_t *bfa, bfa_diag_loopback_t *loopback)
{
	/* build message */
	bfi_diag_lb_req_t *lb_req;

	lb_req = bfa_reqq_next(bfa, BFA_REQQ_DIAG);

	/* build host command */
	bfi_h2i_set(lb_req->mh, BFI_MC_DIAG, BFI_DIAG_H2I_LOOPBACK,
			bfa_ioc_lpuid(bfa));

	lb_req->lb_mode = loopback->lb_mode;
	lb_req->speed = loopback->speed;
	lb_req->loopcnt = loopback->loopcnt;
	lb_req->pattern = loopback->pattern;

	/* ring door bell */
	bfa_reqq_produce(bfa, BFA_REQQ_DIAG);

	bfa_trc(bfa, loopback->lb_mode);
	bfa_trc(bfa, loopback->speed);
	bfa_trc(bfa, loopback->loopcnt);
	bfa_trc(bfa, loopback->pattern);
}

static void
diag_loopback_comp(bfa_t *bfa, bfi_diag_lb_rsp_t *rsp)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	bfa_diag_loopback_result_t *res = diag->lb.result;

	res->numtxmfrm	= bfa_os_ntohl(rsp->res.numtxmfrm);
	res->numosffrm	= bfa_os_ntohl(rsp->res.numosffrm);
	res->numrcvfrm	= bfa_os_ntohl(rsp->res.numrcvfrm);
	res->badfrminf	= bfa_os_ntohl(rsp->res.badfrminf);
	res->badfrmnum	= bfa_os_ntohl(rsp->res.badfrmnum);
	res->status	= rsp->res.status;
	diag->lb.status = rsp->res.status;
	bfa_trc(bfa, diag->lb.status);
	bfa_cb_queue(bfa, &diag->lb.hcb_qe, bfa_cb_diag_lb, diag);
}

/*
 * Temperature Sensor
 */

static void
diag_tempsensor_send(bfa_t *bfa)
{
	/* build message */
	bfi_diag_ts_req_t *ts_req;

	ts_req = bfa_reqq_next(bfa, BFA_REQQ_DIAG);

	/* build host command */
	bfi_h2i_set(ts_req->mh, BFI_MC_DIAG, BFI_DIAG_H2I_TEMPSENSOR,
			bfa_ioc_lpuid(bfa));

	/* ring door bell */
	bfa_reqq_produce(bfa, BFA_REQQ_DIAG);

	bfa_trc(bfa, bfa_reqq_pi(bfa, BFA_REQQ_DIAG));
	bfa_trc(bfa, bfa_reqq_ci(bfa, BFA_REQQ_DIAG));
}

static void
diag_tempsensor_comp(bfa_t *bfa, bfi_diag_ts_rsp_t *rsp)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	uint16_t        ts_value = bfa_os_ntohs(rsp->temp);

	*(diag->tsensor.temp) = ts_value;

	bfa_trc(bfa, rsp->temp);
	bfa_trc(bfa, ts_value);
	bfa_trc(bfa, *(diag->tsensor.temp));

	diag->tsensor.status = BFA_STATUS_OK;
	bfa_cb_queue(bfa, &diag->tsensor.hcb_qe, bfa_cb_diag_tsensor, diag);
}

static void
diag_sfpshow_send(bfa_t *bfa)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	bfi_diag_sfp_req_t *sfp_req;

	sfp_req = bfa_reqq_next(bfa, BFA_REQQ_DIAG);
	memset(sfp_req, 0, sizeof(*sfp_req));

	/* build host command */
	bfi_h2i_set(sfp_req->mh, BFI_MC_DIAG, BFI_DIAG_H2I_SFPSHOW,
			bfa_ioc_lpuid(bfa));
	sfp_req->memtype = BFI_DIAG_SFP_MEM_ALL;
	sfp_req->static_data = diag->sfpshow.static_data;

	/* Setup SG list */
	bfa_diag_sge(diag, sfp_req->req_sge, sizeof(sfp_mem_t));

	/* ring door bell */
	bfa_reqq_produce(bfa, BFA_REQQ_DIAG);
}

static void
diag_sfpshow_comp(bfa_t *bfa, bfi_diag_sfp_rsp_t *rsp)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);

	if (!diag->sfpshow.lock) {
		/*
	 	 * receiving response after ioc failure
	 	 */
		bfa_trc(bfa, diag->sfpshow.lock);
		return;
	}

	if (rsp->status == BFA_STATUS_OK) {
		uint32_t        size;
		uint8_t        *des;

		if (rsp->memtype == BFI_DIAG_SFP_MEM_ALL) {
			size = sizeof(sfp_mem_t);
			des = (uint8_t *) &(diag->sfpshow.sfpmem->srlid_base);
		} else {
			size = 0;
			des = NULL;
		}
		memcpy(des, diag->dbuf_kva, size);
	}

	/**
	 * Queue completion callback.
	 */
	diag->sfpshow.status = rsp->status;
	bfa_trc(bfa, rsp->status);
	bfa_cb_queue(bfa, &diag->sfpshow.hcb_qe, bfa_cb_diag_sfpshow, diag);
}

/*
 *      LED Test command
 */
static void
diag_ledtest_send(bfa_t *bfa, bfa_diag_ledtest_t *ledtest)
{
	/* build message */
	bfi_diag_ledtest_req_t *ledtest_req;

	ledtest_req = bfa_reqq_next(bfa, BFA_REQQ_DIAG);

	/* build host command */
	bfi_h2i_set(ledtest_req->mh, BFI_MC_DIAG, BFI_DIAG_H2I_LEDTEST,
			bfa_ioc_lpuid(bfa));

	/*
	 * convert the freq from N blinks per 10 sec to
	 * crossbow ontime value. We do it here because division is need
	 */
	if (ledtest->freq)
		ledtest->freq = 500 / ledtest->freq;

	if (ledtest->freq == 0)
		ledtest->freq = 1;

	/* mcpy(&ledtest_req->req, ledtest, sizeof(bfa_diag_ledtest_t)); */
	ledtest_req->cmd = (uint8_t) ledtest->cmd;
	ledtest_req->color = (uint8_t) ledtest->color;
	ledtest_req->portid = bfa->pcidev->pci_func;
	ledtest_req->led = ledtest->led;
	ledtest_req->freq = bfa_os_htons(ledtest->freq);

	bfa_reqq_produce(bfa, BFA_REQQ_DIAG);
}

/*
 * Port beaconing
 */
static void
diag_port_beacon_send(bfa_t *bfa, bfa_boolean_t beacon, uint32_t sec)
{
	bfi_diag_portbeacon_req_t *req;

	req = bfa_reqq_next(bfa, BFA_REQQ_DIAG);

	/* build host command */
	bfi_h2i_set(req->mh, BFI_MC_DIAG, BFI_DIAG_H2I_PORTBEACON,
			bfa_ioc_lpuid(bfa));

	req->beacon = bfa_os_htonl(beacon);
	req->period = bfa_os_htonl(sec);
	/* ring door bell */
	bfa_reqq_produce(bfa, BFA_REQQ_DIAG);
}

/*
 * SFP's State Change Notification post to AEN 
 */
static void
diag_sfp_scn_aen_post(bfa_t *bfa, bfi_diag_sfp_event_t *rsp)
{
	bfa_aen_data_t	aen_data;
	bfa_log_mod_t	*logmod = bfa->logm;
	wwn_t			pwwn = rsp->pwwn;
	char			pwwn_buf[BFA_STRING_32];
	char			*pwwn_ptr;
        char            	*pomlvl_str[BFA_PORT_AEN_SFP_POM_MAX] = 
					{"Green", "Amber", "Red"};
	char			*pomlvl_ptr;

	bfa_trc(bfa, rsp->pomlvl << 16 | rsp->sfpid << 8 | rsp->event);
	bfa_trc(bfa, rsp->pwwn);

	pwwn_ptr = wwn2str(pwwn_buf, sizeof(pwwn_buf), pwwn);
	aen_data.port.phy_port_num = rsp->sfpid;
	aen_data.port.pwwn = pwwn;
	if (rsp->event == BFA_DIAG_SFP_EVENT_INSERTED) {
		// Insertion event
		bfa_log(logmod, BFAL_AEN_PORT_SFP_INSERT, rsp->sfpid, pwwn_ptr);
		bfa_aen_post(bfa->aen, BFA_AEN_CAT_PORT, BFA_PORT_AEN_SFP_INSERT,
					 &aen_data);
	} else if (rsp->event == BFA_DIAG_SFP_EVENT_REMOVED) {
		bfa_log(logmod, BFAL_AEN_PORT_SFP_REMOVE, rsp->sfpid, pwwn_ptr);
		bfa_aen_post(bfa->aen, BFA_AEN_CAT_PORT, BFA_PORT_AEN_SFP_REMOVE,
					 &aen_data);
	} else if (rsp->event == BFA_DIAG_SFP_EVENT_POM) {
		 bfa_assert((rsp->pomlvl >= BFA_PORT_AEN_SFP_POM_GREEN) && 
			    (rsp->pomlvl <= BFA_PORT_AEN_SFP_POM_RED));
		aen_data.port.level = rsp->pomlvl;
		pomlvl_ptr = pomlvl_str[rsp->pomlvl - 1];
		bfa_log(logmod, BFAL_AEN_PORT_SFP_POM, 
			pomlvl_ptr, rsp->sfpid, pwwn_ptr);
		bfa_aen_post(bfa->aen, BFA_AEN_CAT_PORT, BFA_PORT_AEN_SFP_POM,
					 &aen_data);
	}
}

static void
diag_portbeacon_comp(bfa_t *bfa, bfi_diag_portbeacon_rsp_t * msg)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);

	bfa_trc(bfa, diag->beacon);
	diag->beacon = BFA_FALSE;
	/* no bfa_cb_queue is needed because driver is not waiting */

	/* set port attributes */
	bfa_pport_beacon(bfa, diag->beacon, diag->link_e2e_beacon);
}

/**
 * @}
 */

/**
 * @dg hal_diag_public
 * @{
 */
void
diag_intr(bfa_t *bfa, bfi_msg_t *msg)
{

	switch (msg->mhdr.msg_id) {
	case BFI_DIAG_I2H_PORTBEACON:
		diag_portbeacon_comp(bfa, (bfi_diag_portbeacon_rsp_t *) msg);
		break;

	case BFI_DIAG_I2H_FWPING:
		diag_fwping_comp(bfa, (bfi_diag_fwping_rsp_t *) msg);
		break;

	case BFI_DIAG_I2H_LOOPBACK:
		diag_loopback_comp(bfa, (bfi_diag_lb_rsp_t *) msg);
		break;

	case BFI_DIAG_I2H_TEMPSENSOR:
		diag_tempsensor_comp(bfa, (bfi_diag_ts_rsp_t *) msg);
		break;

	case BFI_DIAG_I2H_SFPSHOW:
		diag_sfpshow_comp(bfa, (bfi_diag_sfp_rsp_t *) msg);
		break;

	case BFI_DIAG_I2H_SFP_SCN:
		diag_sfp_scn_aen_post(bfa, (bfi_diag_sfp_event_t *) msg);
		break;

	default:
		bfa_trc(bfa, msg->mhdr.msg_id);
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @ig hal_diag_api
 * @{
 */

/*
 * Register access
 */

/**
 * @brief
 *   DIAG register read
 */
bfa_status_t
bfa_diag_reg_read(bfa_t *bfa, uint32_t offset, uint32_t len,
		      uint32_t *buf)
{
	bfa_status_t    status;
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	bfa_os_addr_t   rb;
	bfa_os_addr_t   addr;
	uint32_t        i;

	rb = ioc->pci_info.pci_bar_kva;

	offset &= 0x1fffc;	/* offset only 17 bit and word align */

	/*
	 * offset and len sanity check
	 */
	status = diag_reg_offset_check(offset, len);
	if (status)
		return status;

	addr = rb + offset;
	for (i = 0; i < len; i++) {
		*buf = bfa_reg_read(addr);
		buf++;
		addr += sizeof(uint32_t);
	}
	return BFA_STATUS_OK;
}

/**
 * @brief
 *   DIAG register write
 */
bfa_status_t
bfa_diag_reg_write(bfa_t *bfa, uint32_t offset, uint32_t len,
		       uint32_t value)
{
	bfa_status_t    status;
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	uint32_t       *addr;
	uint32_t        i;

	offset &= 0x1fffc;	/* offset only 17 bit and word align */

	/*
	 * offset and len sanity check
	 */
	status = diag_reg_offset_check(offset, len);
	if (status)
		return status;

	addr = (uint32_t *) ((uint8_t *) ioc->pci_info.pci_bar_kva + offset);
	for (i = 0; i < len; i++) {
		bfa_reg_write(addr, value);
		addr++;
	}
	return BFA_STATUS_OK;
}

/**
 * @brief
 * Gen RAM Test
 *
 *   @param[in] *bfa		- bfa data struct
 *   @param[in] *memtest	- mem test params input from upper layer,
 *   @param[in] pattern		- mem test pattern
 *   @param[in] *result		- mem test result
 *   @param[in] cbfn		- mem test callback functioin
 *   @param[in] cbarg		- callback functioin arg
 *
 *   @param[out]
 */
bfa_status_t
bfa_diag_memtest(bfa_t *bfa, bfa_diag_memtest_t *memtest,
		     uint32_t pattern, bfa_diag_memtest_result_t *result,
		     bfa_cb_diag_t cbfn, void *cbarg)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	uint32_t        pss_ctl;

	bfa_trc(bfa, pattern);

	if (!bfa_ioc_adapter_is_disabled(bfa))
		return BFA_STATUS_ADAPTER_ENABLED;

	/*
	 * check to see if there is another destructive diag cmd running
	 */
	if (diag->block) {
		bfa_trc(bfa, diag->block);
		return (BFA_STATUS_DEVBUSY);
	} else {
		diag->block = 1;
		bfa_diag_set_busy_status(diag);
	}

	diag->result = result;
	diag->cbfn = cbfn;
	diag->cbarg = cbarg;

	/*
	 * put both LPUs into reset
	 */
	pss_ctl = bfa_reg_read(bfa->ioc.bfa_regs.pss_ctl_reg);
	pss_ctl |= (__PSS_LPU0_RESET | __PSS_LPU1_RESET);

	bfa_reg_write(bfa->ioc.bfa_regs.pss_ctl_reg, pss_ctl);

	/*
	 * download memtest code and take LPU0 out of reset
	 */
	bfa_ioc_boot(bfa, BFI_BOOT_TYPE_MEMTEST, pattern);

	bfa_timer_start(bfa, &diag->timer, bfa_diag_memtest_done, bfa,
			    BFA_DIAG_MEMTEST_TOV);
	diag->timer_active = 1;
	return (BFA_STATUS_OK);
}

/**
 * @brief
 * DIAG firmware ping command
 *
 *   @param[in] *bfa		- bfa data struct
 *   @param[in] cnt		- dma loop count for testing PCIE
 *   @param[in] data		- data pattern to pass in fw
 *   @param[in] *result		- pt to bfa_diag_fwping_result_t data struct
 *   @param[in] cbfn		- callback function
 *   @param[in] *cbarg		- callback functioin arg
 *
 *   @param[out]
 */
bfa_status_t
bfa_diag_fwping(bfa_t *bfa, uint32_t cnt, uint32_t data,
		    bfa_diag_fwping_result_t *result, bfa_cb_diag_t cbfn,
		    void *cbarg)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	uint32_t        i;

	bfa_trc(bfa, cnt);
	bfa_trc(bfa, data);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	/*
	 * check to see if there is another destructive diag cmd running
	 * fwping and sfpshow share the same dma area
	 */
	if (diag->block || diag->fwping.lock || diag->sfpshow.lock) {
		bfa_trc(bfa, diag->block);
		bfa_trc(bfa, diag->fwping.lock);
		bfa_trc(bfa, diag->sfpshow.lock);
		return (BFA_STATUS_DEVBUSY);
	}

	/* Initialization */
	diag->fwping.lock = 1;
	diag->fwping.cbfn = cbfn;
	diag->fwping.cbarg = cbarg;
	diag->fwping.result = result;
	diag->fwping.data = data;
	diag->fwping.count = cnt;
	diag->fwping.vcidx = 0;

	/* Init test results */
	for (i = 0; i < BFI_IOC_MAX_CQS; i++) {
		diag->fwping.result->data[i] = 0;
		diag->fwping.result->status[i] = BFA_STATUS_OK;
	}

	/* kick off the first ping */
	diag_fwping_send(bfa, 0);

	/*
	 * Start a timer
	 */
	bfa_timer_start(bfa, &diag->fwping.timer, diag_fwping_timeout, bfa,
			    BFA_DIAG_FWPING_TOV);
	diag->fwping.timer_active = 1;
	return (BFA_STATUS_OK);
}

/**
 * @brief
 * Loopback test
 *
 *   @param[in] *bfa		- bfa data struct
 *   @param[in] opmode		- port operation mode
 *   @param[in] speed		- port speed
 *   @param[in] pat			- pattern to build packet
 *   @param[in] lpcnt		- loop count
 *   @param[in] cbfn		- callback function
 *   @param[in] cbarg		- callback functioin arg
 *   @param[in] *result		- pt to bfa_diag_loopback_result_t data struct
 *
 *   @param[out]
 */
bfa_status_t
bfa_diag_loopback(bfa_t *bfa, bfa_pport_opmode_t opmode,
		      bfa_pport_speed_t speed, uint32_t lpcnt, uint32_t pat,
		      bfa_diag_loopback_result_t *result, bfa_cb_diag_t cbfn,
		      void *cbarg)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);
	bfa_diag_loopback_t loopback;
	bfa_pport_attr_t	attr;

	/*
	 * Check to see if IOC is down
	 */
	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	/*
	 * Check to see if port is disable
	 */
	if (bfa_pport_is_disabled(bfa) == BFA_FALSE) {
		bfa_trc(bfa, opmode);
		return (BFA_STATUS_PORT_NOT_DISABLED);
	}

	/*
	 * Check if the speed is supported
	 */
	bfa_pport_get_attr(bfa, &attr);
	bfa_trc(bfa, attr.speed_supported);
	if (speed > attr.speed_supported) 
		return BFA_STATUS_UNSUPP_SPEED;

	/*
	 * check to see if there is another destructive diag cmd running
	 */
	if (diag->block || diag->lb.lock) {
		bfa_trc(bfa, diag->block);
		bfa_trc(bfa, diag->lb.lock);
		return (BFA_STATUS_DEVBUSY);
	}

	diag->lb.lock = 1;
	loopback.lb_mode = opmode;
	loopback.speed = speed;
	loopback.loopcnt = lpcnt;
	loopback.pattern = pat;
	diag->lb.result = result;
	diag->lb.cbfn = cbfn;
	diag->lb.cbarg = cbarg;
	bfa_os_memset(result, 0, sizeof(bfa_diag_loopback_result_t));
	bfa_diag_set_busy_status(diag);

	/* Send msg to fw */
	diag_loopback_send(bfa, &loopback);

	return (BFA_STATUS_OK);
}

/**
 * @brief
 * Read Temperature Sensor
 *
 *   @param[in] *bfa		- bfa data struct
 *   @param[in] *temp		- pt to temperature
 *   @param[in] cbfn		- callback function
 *   @param[in] *cbarg		- callback functioin arg
 *
 *   @param[out]
 */
bfa_status_t
bfa_diag_tsensor_query(bfa_t *bfa, uint16_t *temp,
			   bfa_cb_diag_t cbfn, void *cbarg)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);

	/*
	 * check to see if there is a destructive diag cmd running
	 */
	if (diag->block || diag->tsensor.lock) {
		bfa_trc(bfa, diag->block);
		bfa_trc(bfa, diag->tsensor.lock);
		return (BFA_STATUS_DEVBUSY);
	}

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	/* Init diag mod params */
	diag->tsensor.lock = 1;
	diag->tsensor.temp = temp;
	diag->tsensor.cbfn = cbfn;
	diag->tsensor.cbarg = cbarg;

	/* Send msg to fw */
	diag_tempsensor_send(bfa);

	return (BFA_STATUS_OK);
}

/**
 * @brief
 * SFP Show command
 *
 *   @param[in] *bfa		- bfa data struct
 *   @param[in] sfpmem		- sfp data
 *   @param[in] static_data	- sfp static data flag
 *   @param[in] cbfn 		- callback functioin
 *   @param[in] *cbarg		- callback functioin arg
 *
 */
bfa_status_t
bfa_diag_sfpshow(bfa_t *bfa, sfp_mem_t *sfpmem, uint8_t static_data,
		     bfa_cb_diag_t cbfn, void *cbarg)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);

	bfa_trc(bfa, diag->block);

	/*
	 * check to see if there is another destructive diag cmd running
	 */
	if (diag->block || diag->sfpshow.lock || diag->fwping.lock) {
		bfa_trc(bfa, diag->sfpshow.lock);
		bfa_trc(bfa, diag->fwping.lock);
		return (BFA_STATUS_DEVBUSY);
	}

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	/* Init diag mod params */
	diag->sfpshow.lock = 1;
	diag->sfpshow.sfpmem = sfpmem;
	diag->sfpshow.static_data = static_data;
	diag->sfpshow.cbfn = cbfn;
	diag->sfpshow.cbarg = cbarg;

	/* Send msg to fw */
	diag_sfpshow_send(bfa);

	return (BFA_STATUS_OK);
}

/**
 * @brief
 * LED Test command
 *
 *   @param[in] *bfa		- bfa data struct
 *   @param[in] *ledtest	- pt to ledtest data structure
 *
 *   @param[out]
 */
bfa_status_t
bfa_diag_ledtest(bfa_t *bfa, bfa_diag_ledtest_t *ledtest)
{
	bfa_trc(bfa, ledtest->cmd);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	/* Send msg to fw */
	diag_ledtest_send(bfa, ledtest);

	return (BFA_STATUS_OK);
}

/**
 * @brief
 * Port beaconing command
 *
 *   @param[in] *bfa		- bfa data struct
 *   @param[in] beacon		- port beaconing 1:ON	0:OFF
 *   @param[in] link_e2e_beacon	- link beaconing 1:ON	0:OFF
 *   @param[in] sec		- beaconing duration in seconds
 *
 *   @param[out]
 */
bfa_status_t
bfa_diag_beacon_port(bfa_t *bfa, 
				bfa_boolean_t beacon, 
				bfa_boolean_t link_e2e_beacon, 
				uint32_t sec)
{
	bfa_diag_mod_t *diag = BFA_DIAG_MOD(bfa);

	bfa_trc(bfa, beacon);
	bfa_trc(bfa, link_e2e_beacon);
	bfa_trc(bfa, sec);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	diag->beacon 		= beacon;
	diag->link_e2e_beacon 	= link_e2e_beacon;

	/* set port attributes */
	bfa_pport_beacon(bfa, diag->beacon, diag->link_e2e_beacon);

	/* Send msg to fw */
	diag_port_beacon_send(bfa, beacon, sec);

	return (BFA_STATUS_OK);
}

/**
 * @}
 */
