/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <bfa.h>
#include <bfa_flash.h>
#include <cs/bfa_debug.h>

BFA_TRC_FILE(HAL, FLASH);
BFA_MODULE(flash);

#define BFA_FLASH_DMA_BUF_SZ  	0x010000	/*! 64k dma buffer */

/*
 *****************************************************************************
 *
 * Internal functions
 *
 *****************************************************************************
 */

static void
__bfa_cb_flash(void *cbarg, bfa_boolean_t complete)
{
	bfa_flash_mod_t *mod = cbarg;

	if (complete)
		mod->cbfn(mod->cbarg, mod->status);
	else
		mod->op_busy = 0;
}

/**
 * @brief Set flash request scatter gather element.
 *
 * @param[in] mod - bfa flash module
 * @param[in] sge - scatter gather element
 * @param[in] len - scatter gather len
 */
static void
bfa_flash_sge(bfa_flash_mod_t *mod, bfi_sge_t *sge, uint32_t len)
{
	bfi_addr_t      sga_zero = { {0} };

	sge->sg_len = len;
	sge->flags = BFI_SGE_DATA_LAST;
	bfa_dma_addr_set(sge->sga, mod->dbuf_pa);
	bfa_sge_to_be(sge);

	sge++;
	sge->sg_len = len;
	sge->flags = BFI_SGE_PGDLEN;
	sge->sga = sga_zero;
	bfa_sge_to_be(sge);
}

/**
 * @brief Send flash attribute query request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_flash_query_send(void *cbarg)
{
	bfa_flash_mod_t *mod = cbarg;
	bfi_flash_query_req_t *msg;

	msg = bfa_reqq_next(mod->bfa, BFA_REQQ_FLASH);

	if (!msg) {
		bfa_reqq_wait(mod->bfa, BFA_REQQ_FLASH,
				  &mod->reqq_wait);
		return;
	}

	bfi_h2i_set(msg->mh, BFI_MC_FLASH, BFI_FLASH_H2I_QUERY_REQ,
			mod->bfa->ioc.pci_info.pci_func);
	bfa_flash_sge(mod, msg->req_sge, sizeof(bfa_flash_attr_t));
	bfa_reqq_produce(mod->bfa, BFA_REQQ_FLASH);

	return;
}

/**
 * @brief Send flash write request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_flash_write_send(void *cbarg)
{
	bfa_flash_mod_t *mod = cbarg;
	bfi_flash_write_req_t *msg;
	uint32_t        len;

	msg = bfa_reqq_next(mod->bfa, BFA_REQQ_FLASH);

	if (!msg) {
		bfa_reqq_wait(mod->bfa, BFA_REQQ_FLASH,
							&mod->reqq_wait);
		return;
	}

	msg->type = bfa_os_ntohl(mod->type);
	msg->instance = bfa_os_ntohl(mod->instance);
	msg->offset = bfa_os_ntohl(mod->addr_off + mod->offset);
	len = (mod->residue < BFA_FLASH_DMA_BUF_SZ) ?
				mod->residue : BFA_FLASH_DMA_BUF_SZ;
	msg->length = bfa_os_ntohl(len);

	/*
	 * indicate if it's the last msg of the whole write operation
	 */
msg->last = (len == mod->residue) ? bfa_os_ntohl(1) : 0;

	bfi_h2i_set(msg->mh, BFI_MC_FLASH, BFI_FLASH_H2I_WRITE_REQ,
					mod->bfa->ioc.pci_info.pci_func);
	bfa_flash_sge(mod, msg->req_sge, len);

	memcpy(mod->dbuf_kva, mod->ubuf + mod->offset, len);
	bfa_reqq_produce(mod->bfa, BFA_REQQ_FLASH);

	mod->residue -= len;
	mod->offset += len;

	return;
}

/**
 * @brief Send flash read request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_flash_read_send(void *cbarg)
{
	bfa_flash_mod_t *mod = cbarg;
	bfi_flash_read_req_t *msg;
	uint32_t        len;

	msg = bfa_reqq_next(mod->bfa, BFA_REQQ_FLASH);

	if (!msg) {
		bfa_reqq_wait(mod->bfa, BFA_REQQ_FLASH,
				  &mod->reqq_wait);
		return;
	}

	msg->type = bfa_os_ntohl(mod->type);
	msg->instance = bfa_os_ntohl(mod->instance);
	msg->offset = bfa_os_ntohl(mod->addr_off + mod->offset);
	len = (mod->residue < BFA_FLASH_DMA_BUF_SZ) ?
				mod->residue : BFA_FLASH_DMA_BUF_SZ;
	msg->length = bfa_os_ntohl(len);
	bfi_h2i_set(msg->mh, BFI_MC_FLASH, BFI_FLASH_H2I_READ_REQ,
			mod->bfa->ioc.pci_info.pci_func);
	bfa_flash_sge(mod, msg->req_sge, len);
	bfa_reqq_produce(mod->bfa, BFA_REQQ_FLASH);

	return;
}

/**
 * @brief Send flash erase request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_flash_erase_send(void *cbarg)
{
	bfa_flash_mod_t *mod = cbarg;
	bfi_flash_erase_req_t *msg;

	msg = bfa_reqq_next(mod->bfa, BFA_REQQ_FLASH);

	if (!msg) {
		bfa_reqq_wait(mod->bfa, BFA_REQQ_FLASH,
							&mod->reqq_wait);
		return;
	}

	msg->type = bfa_os_ntohl(mod->type);
	msg->instance = bfa_os_ntohl(mod->instance);
	bfi_h2i_set(msg->mh, BFI_MC_FLASH, BFI_FLASH_H2I_ERASE_REQ,
					mod->bfa->ioc.pci_info.pci_func);
	bfa_reqq_produce(mod->bfa, BFA_REQQ_FLASH);

	return;
}

/**
 * @brief Send flash option rom request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_flash_optrom_send(void *cbarg)
{
	bfa_flash_mod_t *mod = cbarg;
	bfi_flash_optrom_req_t *msg;

	msg = bfa_reqq_next(mod->bfa, BFA_REQQ_FLASH);

	if (!msg) {
		bfa_reqq_wait(mod->bfa, BFA_REQQ_FLASH,
							&mod->reqq_wait);
		return;
	}

	msg->optrom_enable = bfa_os_ntohl(mod->optrom_enable);
	bfi_h2i_set(msg->mh, BFI_MC_FLASH, BFI_FLASH_H2I_OPTROM_REQ,
					mod->bfa->ioc.pci_info.pci_func);
	bfa_reqq_produce(mod->bfa, BFA_REQQ_FLASH);

	return;
}

/**
 * @brief Update virtual and physical memory for flash.
 *
 * @param[in] cfg - ioc config
 * @param[in] ndm_len - non-dma memory length
 * @param[in] dm_len - dma memory length
 */
static void
bfa_flash_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *ndm_len,
		      uint32_t *dm_len)
{
	/* min driver doesn't need flash */
	if (cfg->min_cfg)
		return;

	*dm_len += BFA_FLASH_DMA_BUF_SZ;
}

/**
 * @brief Attach virtual and physical memory for flash.
 *
 * @param[in] bfa - bfa structure
 * @param[in] bfad - driver structure
 * @param[in] cfg - ioc config
 * @param[in] meminfo - bfa memory info
 * @param[in] pcidev - pci device
 */
static void
bfa_flash_attach(bfa_t *bfa, void *bfad, bfa_ioc_cfg_t *cfg,
		     bfa_meminfo_t *meminfo, bfa_pcidev_t *pcidev)
{
	uint8_t        *dm_kva = NULL;
	uint64_t        dm_pa;
	bfa_flash_mod_t *flash_mod = BFA_FLASH_MOD(bfa);

	flash_mod->bfa = bfa;

	/* min driver doesn't need flash */
	if (cfg->min_cfg) {
		flash_mod->dbuf_kva = NULL;
		flash_mod->dbuf_pa = 0;
		return;
	}

	dm_kva = bfa_meminfo_dma_virt(meminfo);
	dm_pa = bfa_meminfo_dma_phys(meminfo);

	flash_mod->dbuf_kva = dm_kva;
	flash_mod->dbuf_pa = dm_pa;
	dm_kva += BFA_FLASH_DMA_BUF_SZ;
	dm_pa += BFA_FLASH_DMA_BUF_SZ;
	bfa_os_memset(flash_mod->dbuf_kva, 0, BFA_FLASH_DMA_BUF_SZ);

	bfa_meminfo_dma_virt(meminfo) = dm_kva;
	bfa_meminfo_dma_phys(meminfo) = dm_pa;
}

static void
bfa_flash_initdone(bfa_t *bfa)
{
}

static void
bfa_flash_detach(bfa_t *bfa)
{
}

static void
bfa_flash_start(bfa_t *bfa)
{
}

static void
bfa_flash_stop(bfa_t *bfa)
{
}

static void
bfa_flash_iocdisable(bfa_t *bfa)
{
	bfa_flash_mod_t *mod = &(bfa->modules.flash_mod); 

	bfa_trc(bfa, mod->op_busy);
	if (!mod->op_busy)
		return;

	mod->status = BFA_STATUS_IOC_FAILURE;
	bfa_cb_queue(bfa, &mod->hcb_qe, __bfa_cb_flash, mod);
}

/*
 *****************************************************************************
 *
 * Externalized functions
 *
 *****************************************************************************
 */

/**
 * @ig hal_flash_api
 * @{
 */

/**
 * @brief Get flash attribute.
 *
 * @param[in] bfa - bfa structure
 * @param[in] attr - flash attribute structure
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_flash_get_attr(bfa_t *bfa, bfa_flash_attr_t *attr,
		       bfa_cb_flash_t cbfn, void *cbarg)
{
	bfa_flash_mod_t *mod = BFA_FLASH_MOD(bfa);

	bfa_trc(bfa, BFI_FLASH_H2I_QUERY_REQ);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	if (mod->op_busy) {
		bfa_trc(bfa, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->ubuf = (uint8_t *) attr;
	bfa_reqq_winit(&mod->reqq_wait, bfa_flash_query_send, mod);
	bfa_flash_query_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Erase flash partition.
 *
 * @param[in] bfa - bfa structure
 * @param[in] type - flash partition type
 * @param[in] instance - flash partition instance
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_flash_erase_part(bfa_t *bfa, bfa_flash_part_type_t type,
			 uint32_t instance, bfa_cb_flash_t cbfn, void *cbarg)
{
	bfa_flash_mod_t *mod = BFA_FLASH_MOD(bfa);

	bfa_trc(bfa, BFI_FLASH_H2I_ERASE_REQ);
	bfa_trc(bfa, type);
	bfa_trc(bfa, instance);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	if (mod->op_busy) {
		bfa_trc(bfa, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->type = type;
	mod->instance = instance;
	bfa_reqq_winit(&mod->reqq_wait, bfa_flash_erase_send, mod);

	bfa_flash_erase_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Update flash partition.
 *
 * @param[in] bfa - bfa structure
 * @param[in] type - flash partition type
 * @param[in] instance - flash partition instance
 * @param[in] buf - update data buffer
 * @param[in] len - data buffer length
 * @param[in] offset - offset relative to the partition starting address
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_flash_update_part(bfa_t *bfa, bfa_flash_part_type_t type,
			  uint32_t instance, void *buf, uint32_t len,
			  uint32_t offset, bfa_cb_flash_t cbfn, void *cbarg)
{
	bfa_flash_mod_t *mod = BFA_FLASH_MOD(bfa);

	bfa_trc(bfa, BFI_FLASH_H2I_WRITE_REQ);
	bfa_trc(bfa, type);
	bfa_trc(bfa, instance);
	bfa_trc(bfa, len);
	bfa_trc(bfa, offset);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	/**
	 * 'len' must be in word (4-byte) boundary
	 * 'offset' must be in sector (16kb) boundary
	 */
	if (!len || (len & 0x03) || (offset & 0x00003FFF))
		return (BFA_STATUS_FLASH_BAD_LEN);

	if (mod->op_busy) {
		bfa_trc(bfa, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->type = type;
	mod->instance = instance;
	mod->residue = len;
	mod->offset = 0;
	mod->addr_off = offset;
	mod->ubuf = buf;
	bfa_reqq_winit(&mod->reqq_wait, bfa_flash_write_send, mod);

	bfa_flash_write_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Read flash partition.
 *
 * @param[in] bfa - bfa structure
 * @param[in] type - flash partition type
 * @param[in] instance - flash partition instance
 * @param[in] buf - read data buffer
 * @param[in] len - data buffer length
 * @param[in] offset - offset relative to the partition starting address
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_flash_read_part(bfa_t *bfa, bfa_flash_part_type_t type,
			uint32_t instance, void *buf, uint32_t len,
			uint32_t offset, bfa_cb_flash_t cbfn, void *cbarg)
{
	bfa_flash_mod_t *mod = BFA_FLASH_MOD(bfa);

	bfa_trc(bfa, BFI_FLASH_H2I_READ_REQ);
	bfa_trc(bfa, type);
	bfa_trc(bfa, instance);
	bfa_trc(bfa, len);
	bfa_trc(bfa, offset);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	/**
	 * 'len' must be in word (4-byte) boundary
	 * 'offset' must be in sector (16kb) boundary
	 */
	if (!len || (len & 0x03) || (offset & 0x00003FFF))
		return (BFA_STATUS_FLASH_BAD_LEN);

	if (mod->op_busy) {
		bfa_trc(bfa, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->type = type;
	mod->instance = instance;
	mod->residue = len;
	mod->offset = 0;
	mod->addr_off = offset;
	mod->ubuf = buf;
	bfa_reqq_winit(&mod->reqq_wait, bfa_flash_read_send, mod);
	bfa_flash_read_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Enable flash option rom.
 *
 * @param[in] bfa - bfa structure
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_flash_enable_optrom(bfa_t *bfa, bfa_cb_flash_t cbfn, void *cbarg)
{
	bfa_flash_mod_t *mod = BFA_FLASH_MOD(bfa);

	bfa_trc(bfa, BFI_FLASH_H2I_OPTROM_REQ);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	if (mod->op_busy) {
		bfa_trc(bfa, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->optrom_enable = 1;
	bfa_reqq_winit(&mod->reqq_wait, bfa_flash_optrom_send, mod);

	bfa_flash_optrom_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Disable flash option rom.
 *
 * @param[in] bfa - bfa structure
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_flash_disable_optrom(bfa_t *bfa, bfa_cb_flash_t cbfn, void *cbarg)
{
	bfa_flash_mod_t *mod = BFA_FLASH_MOD(bfa);

	bfa_trc(bfa, BFI_FLASH_H2I_OPTROM_REQ);

	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_IOC_NON_OP;

	if (mod->op_busy) {
		bfa_trc(bfa, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->optrom_enable = 0;
	bfa_reqq_winit(&mod->reqq_wait, bfa_flash_optrom_send, mod);

	bfa_flash_optrom_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Process flash response messages upon receiving interrupts.
 *
 * @param[in] bfa - bfa structure
 * @param[in] msg - message structure
 */
void
bfa_flash_intr(bfa_t *bfa, bfi_msg_t *msg)
{
	uint32_t        status;
	bfa_flash_mod_t *mod = BFA_FLASH_MOD(bfa);

	union {
		bfi_flash_query_rsp_t *query;
		bfi_flash_erase_rsp_t *erase;
		bfi_flash_write_rsp_t *write;
		bfi_flash_read_rsp_t *read;
		bfi_flash_read_rsp_t *optrom;
		bfi_msg_t      *msg;
	} m;

	m.msg = msg;
	bfa_trc(bfa, msg->mhdr.msg_id);

	if (!mod->op_busy) {
		/*
		 * receiving response after ioc failure 
		 */
		bfa_trc(bfa, 0x9999);
		return;
	}

	switch (msg->mhdr.msg_id) {
	case BFI_FLASH_I2H_QUERY_RSP:
		status = bfa_os_ntohl(m.query->status);
		bfa_trc(bfa, status);

		if (status == BFA_STATUS_OK) {
			uint32_t        i;
			bfa_flash_attr_t *attr, *f;

			attr = (bfa_flash_attr_t *) mod->ubuf;
			f = (bfa_flash_attr_t *) mod->dbuf_kva;
			attr->status = bfa_os_ntohl(f->status);
			attr->npart = bfa_os_ntohl(f->npart);
			bfa_trc(bfa, attr->status);
			bfa_trc(bfa, attr->npart);
			for (i = 0; i < attr->npart; i++) {
				attr->part[i].part_type =
					bfa_os_ntohl(f->part[i].part_type);
				attr->part[i].part_instance =
					bfa_os_ntohl(f->part[i].part_instance);
				attr->part[i].part_off =
					bfa_os_ntohl(f->part[i].part_off);
				attr->part[i].part_size =
					bfa_os_ntohl(f->part[i].part_size);
				attr->part[i].part_len =
					bfa_os_ntohl(f->part[i].part_len);
				attr->part[i].part_status =
					bfa_os_ntohl(f->part[i].part_status);
			}

		}

		mod->status = status;
		bfa_cb_queue(mod->bfa, &mod->hcb_qe, __bfa_cb_flash, mod);
		break;
	case BFI_FLASH_I2H_ERASE_RSP:
		status = bfa_os_ntohl(m.erase->status);
		bfa_trc(bfa, status);

		mod->status = status;
		bfa_cb_queue(mod->bfa, &mod->hcb_qe, __bfa_cb_flash, mod);
		break;
	case BFI_FLASH_I2H_WRITE_RSP:
		status = bfa_os_ntohl(m.write->status);
		bfa_trc(bfa, status);

		if (status != BFA_STATUS_OK || mod->residue == 0) {
			mod->status = status;
			bfa_cb_queue(mod->bfa, &mod->hcb_qe, __bfa_cb_flash,
				      mod);
		} else {
			bfa_trc(bfa, mod->offset);
			bfa_flash_write_send(mod);
		}
		break;
	case BFI_FLASH_I2H_READ_RSP:
		status = bfa_os_ntohl(m.read->status);
		bfa_trc(bfa, status);

		if (status != BFA_STATUS_OK) {
			mod->status = status;
			bfa_cb_queue(mod->bfa, &mod->hcb_qe, __bfa_cb_flash,
				      mod);
		} else {
			uint32_t        len = bfa_os_ntohl(m.read->length);
			bfa_trc(bfa, mod->offset);
			bfa_trc(bfa, len);

			memcpy(mod->ubuf + mod->offset, mod->dbuf_kva, len);
			mod->residue -= len;
			mod->offset += len;

			if (mod->residue == 0) {
				mod->status = status;
				bfa_cb_queue(mod->bfa, &mod->hcb_qe,
					      __bfa_cb_flash, mod);
			} else {
				bfa_flash_read_send(mod);
			}
		}
		break;
	case BFI_FLASH_I2H_OPTROM_RSP:
		status = bfa_os_ntohl(m.optrom->status);
		bfa_trc(bfa, status);

		mod->status = status;
		bfa_cb_queue(mod->bfa, &mod->hcb_qe, __bfa_cb_flash, mod);
		break;
	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */
