/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
#include <bfa.h>
#include <bfi/bfi_cbreg.h>
#include <bfa_port_priv.h>
#include <bfa_intr_priv.h>
#include <cs/bfa_debug.h>

BFA_TRC_FILE(HAL, INTR);

extern bfa_isr_func_t bfa_isrs[BFI_MC_MAX];

/**
 * Resume request queue once it was full and now non-full
 */
static void
bfa_reqq_resume(bfa_t *bfa, int qid)
{
	bfa_q_t 		*waitq, *qe, *qen;
	struct bfa_reqq_wait_s	*wqe;

	waitq = bfa_reqq(bfa, qid);
	bfa_q_iter_safe(waitq, qe, qen) {
		/**
		 * Callback only as long as there is room in request queue
		 */
		if (bfa_reqq_full(bfa, qid))
			break;

		bfa_q_qe_deq(qe);
		wqe = (struct bfa_reqq_wait_s *) qe;
		wqe->qresume(wqe->cbarg);
	}
}

/**
 * @ig hal_intr_api
 * @{
 */
bfa_isr_status_t
bfa_intx(bfa_t *bfa)
{
	uint32_t        intr;
	int             vec;

	intr = bfa_reg_read(bfa->ioc.bfa_regs.intr_status);
	if (!intr)
		return BFA_ISR_NOTCLAIMED;

	/**
	 * mailbox interrupt
	 */
	if (bfa->ioc.pci_info.pci_func == 0) {
		if (intr & __HFN_INT_MBOX_LPU0)
			bfa_msix_lpu(bfa);
	} else {
		if (intr & __HFN_INT_MBOX_LPU1)
			bfa_msix_lpu(bfa);
	}

	/**
	 * RME completion queue interrupt
	 */
	for (vec = HFN_MSIX_RME_Q0; vec <= HFN_MSIX_RME_Q7; vec++)
		if (intr & (__HFN_INT_RME_Q0 << (vec - HFN_MSIX_RME_Q0)))
			bfa_msix_rspq(bfa, vec);

	/**
	 * CPE completion queue interrupt
	 */
	for (vec = HFN_MSIX_CPE_Q0; vec <= HFN_MSIX_CPE_Q7; vec++)
		if (intr & (__HFN_INT_CPE_Q0 << (vec - HFN_MSIX_CPE_Q0)))
			bfa_msix_reqq(bfa, vec);

	if (intr & (__HFN_INT_ERR_EMC | __HFN_INT_ERR_LPU0 |
				__HFN_INT_ERR_LPU1 | __HFN_INT_ERR_PSS))
		bfa_msix_errint(bfa);

	return (BFA_ISR_CLAIMED_COMP);
}

void
bfa_isr_enable(bfa_t *bfa)
{
	uint32_t        intr_unmask;
	int             pci_func = bfa->ioc.pci_info.pci_func;

	bfa_trc(bfa, pci_func);

	intr_unmask = (__HFN_INT_ERR_EMC | __HFN_INT_ERR_LPU0 |
				__HFN_INT_ERR_LPU1 | __HFN_INT_ERR_PSS);

	if (pci_func == 0)
		intr_unmask |= (__HFN_INT_CPE_Q0 | __HFN_INT_CPE_Q1 |
				__HFN_INT_CPE_Q2 | __HFN_INT_CPE_Q3 |
				__HFN_INT_RME_Q0 | __HFN_INT_RME_Q1 |
				__HFN_INT_RME_Q2 | __HFN_INT_RME_Q3 |
				__HFN_INT_MBOX_LPU0);
	else
		intr_unmask |= (__HFN_INT_CPE_Q4 | __HFN_INT_CPE_Q5 |
				__HFN_INT_CPE_Q6 | __HFN_INT_CPE_Q7 |
				__HFN_INT_RME_Q4 | __HFN_INT_RME_Q5 |
				__HFN_INT_RME_Q6 | __HFN_INT_RME_Q7 |
				__HFN_INT_MBOX_LPU1);

	bfa_reg_write(bfa->ioc.bfa_regs.intr_mask, ~intr_unmask);
}

void
bfa_isr_disable(bfa_t *bfa)
{
	bfa_reg_write(bfa->ioc.bfa_regs.intr_mask, -1);
}

void
bfa_msix_getvecs(bfa_t *bfa, uint32_t *msix_vecs_bmap,
		     uint32_t *num_vecs, uint32_t *max_vec_bit)
{
#define __HFN_NUMINTS	13
	if (bfa->ioc.pci_info.pci_func == 0) {
		*msix_vecs_bmap = (__HFN_INT_CPE_Q0 | __HFN_INT_CPE_Q1 |
					__HFN_INT_CPE_Q2 | __HFN_INT_CPE_Q3 |
					__HFN_INT_RME_Q0 | __HFN_INT_RME_Q1 |
					__HFN_INT_RME_Q2 | __HFN_INT_RME_Q3 |
					__HFN_INT_MBOX_LPU0);
		*max_vec_bit = __HFN_INT_MBOX_LPU0;
	} else {
		*msix_vecs_bmap = (__HFN_INT_CPE_Q4 | __HFN_INT_CPE_Q5 |
					__HFN_INT_CPE_Q6 | __HFN_INT_CPE_Q7 |
					__HFN_INT_RME_Q4 | __HFN_INT_RME_Q5 |
					__HFN_INT_RME_Q6 | __HFN_INT_RME_Q7 |
					__HFN_INT_MBOX_LPU1);
		*max_vec_bit = __HFN_INT_MBOX_LPU1;
	}

	*msix_vecs_bmap |= (__HFN_INT_ERR_EMC | __HFN_INT_ERR_LPU0 |
				__HFN_INT_ERR_LPU1 | __HFN_INT_ERR_PSS);
	*num_vecs = __HFN_NUMINTS;
}

bfa_isr_status_t
bfa_msix_reqq(bfa_t *bfa, int qid)
{
	/**
	 * ack the interrupt first
	 */
	bfa_reg_write(bfa->ioc.bfa_regs.intr_status, (1 << (qid)));

	qid &= (BFI_IOC_MAX_CQS - 1);
	bfa_reqq_resume(bfa, qid);
	return (BFA_ISR_CLAIMED_COMP);
}

void
bfa_isr_unhandled(bfa_t *bfa, bfi_msg_t *m)
{
	bfa_trc(bfa, m->mhdr.msg_class);
	bfa_trc(bfa, m->mhdr.msg_id);
	bfa_trc(bfa, m->mhdr.mtag.i2htok);
	bfa_assert(0);
	bfa_trc_stop(bfa->trcmod);
}

bfa_isr_status_t
bfa_msix_rspq(bfa_t *bfa, int rsp_qid)
{
	bfi_msg_t      *m;
	uint32_t        pi, ci;

	bfa_trc_fp(bfa, rsp_qid);

	/**
	 * ack the interrupt first
	 */
	bfa_reg_write(bfa->ioc.bfa_regs.intr_status,
					  (1 << (rsp_qid)));

	rsp_qid &= (BFI_IOC_MAX_CQS - 1);

	ci = bfa_rspq_ci(bfa, rsp_qid);
	pi = bfa_rspq_pi(bfa, rsp_qid);

	bfa_trc_fp(bfa, ci);
	bfa_trc_fp(bfa, pi);

	while (ci != pi) {
		m = bfa_rspq_elem(bfa, rsp_qid, ci);
		bfa_assert_fp(m->mhdr.msg_class < BFI_MC_MAX);

		bfa_isrs[m->mhdr.msg_class] (bfa, m);

		CQ_INCR(ci, bfa->ioc.cfg.num_rspq_elems);
	}

	/**
	 * update CI
	 */
	bfa_rspq_ci(bfa, rsp_qid) = ci;
	bfa_reg_write(bfa->ioc.bfa_regs.rme_q_ci[rsp_qid],
						bfa->ioc.rsp_cq_ci[rsp_qid]);
	bfa_os_mmiowb();

	/**
	 * Resume corresponding request queue.
	 */
	bfa_reqq_resume(bfa, rsp_qid);

	return (BFA_ISR_CLAIMED_COMP);
}

bfa_isr_status_t
bfa_msix_lpu(bfa_t *bfa)
{
	struct bfa_ioc_s	*ioc = &bfa->ioc;
	bfi_ioc_i2h_msg_t	msg;

	bfa_ioc_msgget(ioc, &msg);

	/**
	 * ack the interrupt next
	 */
	if (ioc->pci_info.pci_func == 0)
		bfa_reg_write(ioc->bfa_regs.intr_status, __HFN_INT_MBOX_LPU0);
	else
		bfa_reg_write(ioc->bfa_regs.intr_status, __HFN_INT_MBOX_LPU1);

	bfa_trc(bfa, msg.mh.msg_class);
	bfa_trc(bfa, msg.mh.msg_id);

	switch (msg.mh.msg_class) {
	case BFI_MC_IOC:
		bfa_ioc_isr(ioc, &msg);
		break;

	case BFI_MC_IOCFC:
		bfa_iocfc_isr(ioc, (bfi_iocfc_i2h_msg_t *)&msg);
		break;

	default:
		bfa_assert(0);
	}

	return (BFA_ISR_CLAIMED_COMP);
}

bfa_isr_status_t
bfa_msix_errint(bfa_t *bfa)
{
	uint32_t        intr;

	intr = bfa_reg_read(bfa->ioc.bfa_regs.intr_status);
	bfa_trc(bfa, intr);
	bfa_reg_write(bfa->ioc.bfa_regs.intr_status, intr);
	bfa_isr_disable(bfa);
	bfa_assert(0);

	return (BFA_ISR_CLAIMED_COMP);
}

void
bfa_isr_bind(bfi_mclass_t mc, bfa_isr_func_t isr_func)
{
	bfa_isrs[mc] = isr_func;
}

/**
 * @}
 */
