/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
#include "bfa_ioc_priv.h"
#include "bfa_callback_priv.h"
#include <cs/bfa_debug.h>
#include <bfa_priv.h>
#include <log/bfa_log_hal.h>
#include <bfi/bfi_boot.h>
#include <cs/bfa_plog.h>
#include <cs/bfa_stats.h>
#include <aen/bfal_aen_ioc.h>

BFA_TRC_FILE(HAL, IOC);

/**
 * IOC local definitions
 */
#define BFA_IOC_TOV		5000	/* msecs */
#define BFA_IOC_HB_TOV      2000	/* msecs */
#define BFA_IOC_HB_FAIL_MAX 10
#define BFA_IOC_HWINIT_MAX	2
#define BFA_IOC_FWIMG_MINSZ     (16 * 1024)

#define bfa_ioc_timer_start(__ioc)					\
	bfa_timer_start((__ioc)->bfa, &(__ioc)->ioc_timer,          \
			bfa_ioc_timeout, (__ioc), BFA_IOC_TOV)
#define bfa_ioc_timer_stop(__ioc)   bfa_timer_stop(&(__ioc)->ioc_timer)

bfa_boolean_t bfa_auto_recover = BFA_TRUE;

/*
 * forward declarations
 */
static void	bfa_ioc_aen_post(bfa_t *bfa, bfa_ioc_aen_event_t event);
static void     bfa_ioc_hw_sem_get(struct bfa_ioc_s *ioc);
static void     bfa_ioc_hw_sem_release(struct bfa_ioc_s *ioc);
static void     bfa_ioc_hw_sem_get_cancel(struct bfa_ioc_s *ioc);
static void     bfa_ioc_hwinit(struct bfa_ioc_s *ioc, bfa_boolean_t force);
static void     bfa_ioc_timeout(void *ioc);
static void     bfa_ioc_send_enable(struct bfa_ioc_s *ioc);
static void     bfa_ioc_send_disable(struct bfa_ioc_s *ioc);
static void     bfa_ioc_hb_monitor(struct bfa_ioc_s *ioc);
static void     bfa_ioc_hb_stop(struct bfa_ioc_s *ioc);
static void     bfa_ioc_callback(struct bfa_ioc_s *ioc);
static void     bfa_ioc_reset(struct bfa_ioc_s *ioc, bfa_boolean_t force);

/**
 * @ig hal_ioc_sm
 * @{
 * @img ioc_sm.jpg
 */

/**
 * IOC state machine events
 */
typedef enum {
	IOC_E_INIT	= 1,
	IOC_E_ENABLE	= 2,
	IOC_E_DISABLE	= 3,
	IOC_E_TIMEOUT	= 4,
	IOC_E_FWREADY	= 5,
	IOC_E_FWRSP_CFG	= 6,
	IOC_E_FWRSP_ENABLE = 7,
	IOC_E_FWRSP_DISABLE = 8,
	IOC_E_HBFAIL	= 9,	/* !< heartbeat failure */
	IOC_E_SEMLOCKED	= 10,	/* !< h/w semaphore is locked */
} ioc_event_t;

static void     bfa_ioc_sm_reset(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_semwait(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_hwinit(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_cfginit(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_cfgdone(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_op(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_hwfail(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_hbfail(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_disabling(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_disabled(struct bfa_ioc_s *ioc, ioc_event_t event);
static void     bfa_ioc_sm_recover(struct bfa_ioc_s *ioc, ioc_event_t event);

static bfa_sm_table_t ioc_sm_table[] = {
	{BFA_SM(bfa_ioc_sm_reset), BFA_IOC_RESET},
	{BFA_SM(bfa_ioc_sm_semwait), BFA_IOC_SEMWAIT},
	{BFA_SM(bfa_ioc_sm_hwinit), BFA_IOC_HWINIT},
	{BFA_SM(bfa_ioc_sm_cfginit), BFA_IOC_CFGINIT},
	{BFA_SM(bfa_ioc_sm_cfgdone), BFA_IOC_CFGDONE},
	{BFA_SM(bfa_ioc_sm_op), BFA_IOC_OPERATIONAL},
	{BFA_SM(bfa_ioc_sm_hwfail), BFA_IOC_HWFAIL},
	{BFA_SM(bfa_ioc_sm_hbfail), BFA_IOC_HBFAIL},
	{BFA_SM(bfa_ioc_sm_disabling), BFA_IOC_DISABLING},
	{BFA_SM(bfa_ioc_sm_disabled), BFA_IOC_DISABLED},
	{BFA_SM(bfa_ioc_sm_recover), BFA_IOC_RECOVER},
};

/**
 * @brief
 * Beginning state. IOC is in reset state.
 */
static void
bfa_ioc_sm_reset(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_INIT:
		bfa_sm_set_state(ioc, bfa_ioc_sm_semwait);
		bfa_ioc_hw_sem_get(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * Awaiting release of h/w semaphore.
 */
static void
bfa_ioc_sm_semwait(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_SEMLOCKED:
		bfa_sm_set_state(ioc, bfa_ioc_sm_hwinit);
		bfa_ioc_timer_start(ioc);
		ioc->hwinit_count = 0;
		bfa_ioc_reset(ioc, BFA_FALSE);
		break;

	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabled);
		bfa_isr_disable(ioc->bfa);
		bfa_ioc_hw_sem_get_cancel(ioc);
		bfa_ioc_callback(ioc);
		break;

	case IOC_E_FWREADY:
		bfa_sm_set_state(ioc, bfa_ioc_sm_cfginit);
		bfa_ioc_hw_sem_get_cancel(ioc);
		bfa_ioc_timer_start(ioc);
		bfa_iocfc_send_cfg(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * Hardware is being initialized. Interrupts are enabled.
 * Holding hardware semaphore lock.
 */
static void
bfa_ioc_sm_hwinit(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_FWREADY:
	case IOC_E_FWRSP_ENABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_cfginit);
		bfa_ioc_timer_stop(ioc);
		bfa_ioc_hw_sem_release(ioc);

		bfa_ioc_timer_start(ioc);
		bfa_iocfc_send_cfg(ioc);
		break;

	case IOC_E_TIMEOUT:
		ioc->hwinit_count++;
		if (ioc->hwinit_count < BFA_IOC_HWINIT_MAX) {
			bfa_ioc_timer_start(ioc);
			bfa_ioc_reset(ioc, BFA_TRUE);
			break;
		}

		bfa_ioc_hw_sem_release(ioc);
		bfa_sm_set_state(ioc, bfa_ioc_sm_hwfail);
		bfa_ioc_callback(ioc);
		break;

	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabled);
		bfa_ioc_hw_sem_release(ioc);
		bfa_ioc_timer_stop(ioc);
		bfa_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IOC configuration in progress. Timer is active.
 */
static void
bfa_ioc_sm_cfginit(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_FWRSP_CFG:
		bfa_ioc_timer_stop(ioc);
		if (ioc->ioc_started) {
			bfa_sm_set_state(ioc, bfa_ioc_sm_op);
			bfa_iocfc_initdone_submod(ioc->bfa);
			bfa_iocfc_start_submod(ioc->bfa);
			bfa_ioc_hb_monitor(ioc);
		} else {
			bfa_sm_set_state(ioc, bfa_ioc_sm_cfgdone);
			bfa_iocfc_initdone_submod(ioc->bfa);
			bfa_ioc_callback(ioc);
		}
		break;

	case IOC_E_TIMEOUT:
		bfa_sm_set_state(ioc, bfa_ioc_sm_hwfail);
		bfa_ioc_callback(ioc);
		break;

	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabled);
		bfa_ioc_timer_stop(ioc);
		bfa_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_ioc_sm_cfgdone(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_ENABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_op);
		bfa_ioc_aen_post(ioc->bfa, BFA_IOC_AEN_ENABLE);
		bfa_iocfc_start_submod(ioc->bfa);
		bfa_ioc_hb_monitor(ioc);
		break;

	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabling);
		bfa_ioc_timer_start(ioc);
		bfa_ioc_send_disable(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_ioc_sm_op(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabling);
		bfa_ioc_aen_post(ioc->bfa, BFA_IOC_AEN_DISABLE);
		bfa_ioc_hb_stop(ioc);
		bfa_ioc_timer_start(ioc);
		bfa_ioc_send_disable(ioc);
		break;

	case IOC_E_FWREADY:
		/**
		 * IOC recovery by other function. Treat it same as heartbeat failure.
		 */
		bfa_ioc_hb_stop(ioc);
		/* !!! fall through !!! */

	case IOC_E_HBFAIL:
		if (bfa_auto_recover) {
			bfa_sm_set_state(ioc, bfa_ioc_sm_recover);
			bfa_ioc_timer_start(ioc);
		} else {
			bfa_sm_set_state(ioc, bfa_ioc_sm_hbfail);
			bfa_reg_write(ioc->bfa_regs.ioc_fwstate,
					  BFI_IOC_HBFAIL);
		}
		bfa_iocfc_iocdisable_submod(ioc->bfa);
		bfa_ioc_aen_post(ioc->bfa, BFA_IOC_AEN_HBFAIL);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_ioc_sm_disabling(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_FWRSP_DISABLE:
		bfa_ioc_timer_stop(ioc);
		/*
		 * !!! fall through !!!
		 */

	case IOC_E_TIMEOUT:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabled);
		bfa_isr_disable(ioc->bfa);
		bfa_iocfc_iocdisable_submod(ioc->bfa);
		bfa_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_ioc_sm_disabled(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_ENABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_semwait);
		bfa_ioc_aen_post(ioc->bfa, BFA_IOC_AEN_ENABLE);
		bfa_ioc_hw_sem_get(ioc);
		break;

	case IOC_E_DISABLE:
		bfa_ioc_callback(ioc);
		break;

	case IOC_E_FWREADY:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * Hardware initialization failed.
 */
static void
bfa_ioc_sm_hwfail(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabled);
		bfa_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IOC heartbeat failure.
 */
static void
bfa_ioc_sm_hbfail(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabled);
		bfa_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IOC heartbeat failure.
 */
static void
bfa_ioc_sm_recover(struct bfa_ioc_s *ioc, ioc_event_t event)
{
	bfa_trc(ioc->bfa, event);

	switch (event) {
	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_ioc_sm_disabled);
		bfa_ioc_timer_stop(ioc);
		bfa_ioc_callback(ioc);
		break;

	case IOC_E_FWREADY:
		/**
		 * Recovery is already initiated by other function.
		 */
		break;

	case IOC_E_TIMEOUT:
		bfa_sm_set_state(ioc, bfa_ioc_sm_semwait);
		bfa_ioc_hw_sem_get(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg hal_ioc_pvt HAL IOC private functions
 * @{
 */

/**
 * Send AEN notification
 */
static void
bfa_ioc_aen_post(bfa_t *bfa, bfa_ioc_aen_event_t event)
{
	bfa_aen_data_t  aen_data;
	bfa_log_mod_t	*logmod = bfa->logm;
	int32_t		inst_num = bfa_aen_get_inst(bfa->aen);

	switch (event) {
	case BFA_IOC_AEN_HBGOOD:
		bfa_log(logmod, BFAL_AEN_IOC_HBGOOD, inst_num);
		break;
	case BFA_IOC_AEN_HBFAIL:
		bfa_log(logmod, BFAL_AEN_IOC_HBFAIL, inst_num);
		break;
	case BFA_IOC_AEN_ENABLE:
		bfa_log(logmod, BFAL_AEN_IOC_ENABLE, inst_num);
		break;
	case BFA_IOC_AEN_DISABLE:
		bfa_log(logmod, BFAL_AEN_IOC_DISABLE, inst_num);
		break;
	default:
		break;
	}

	aen_data.ioc.pwwn = bfa_iocfc_get_pwwn(bfa);
	bfa_aen_post(bfa->aen, BFA_AEN_CAT_IOC, event, &aen_data);
}

static void
bfa_ioc_sem_timeout(void *ioc_arg)
{
	struct bfa_ioc_s  *ioc = (struct bfa_ioc_s *) ioc_arg;

	bfa_ioc_hw_sem_get(ioc);
}

static void
bfa_ioc_hw_sem_get(struct bfa_ioc_s *ioc)
{
	uint32_t        r32;

	/**
	 * First read to the semaphore register will return 0, subsequent reads
	 * will return 1. Semaphore is released by writing 0 to the register
	 */
	r32 = bfa_reg_read(ioc->bfa_regs.host_sem0_reg);
	if (r32 == 0) {
		bfa_sm_send_event(ioc, IOC_E_SEMLOCKED);
		return;
	}

	bfa_timer_start(ioc->bfa, &ioc->sem_timer, bfa_ioc_sem_timeout,
			    ioc, BFA_IOC_TOV);
}

static void
bfa_ioc_hw_sem_release(struct bfa_ioc_s *ioc)
{
	bfa_reg_write(ioc->bfa_regs.host_sem0_reg, 1);
}

static void
bfa_ioc_hw_sem_get_cancel(struct bfa_ioc_s *ioc)
{
	bfa_timer_stop(&ioc->sem_timer);
}

/**
 * @brief
 * Initialize LPU local memory (aka secondary memory / SRAM)
 */
static void
bfa_ioc_lmem_init(struct bfa_ioc_s *ioc)
{
	uint32_t        pss_ctl;
	int             i;
#define PSS_LMEM_INIT_TIME  10000

	pss_ctl = bfa_reg_read(ioc->bfa_regs.pss_ctl_reg);
	pss_ctl &= ~__PSS_LMEM_RESET;
	pss_ctl |= __PSS_LMEM_INIT_EN;
	pss_ctl |= __PSS_I2C_CLK_DIV(3UL);	/* i2c workaround 12.5khz clock
						 */
	bfa_reg_write(ioc->bfa_regs.pss_ctl_reg, pss_ctl);

    /**
     * wait for memory initialization to be complete
     */
	i = 0;
	do {
		pss_ctl = bfa_reg_read(ioc->bfa_regs.pss_ctl_reg);
		i++;
	} while (!(pss_ctl & __PSS_LMEM_INIT_DONE) && (i < PSS_LMEM_INIT_TIME));

    /**
     * If memory initialization is not successful, IOC timeout will catch
	 * such failures.
     */
	bfa_assert(pss_ctl & __PSS_LMEM_INIT_DONE);
	bfa_trc(ioc->bfa, pss_ctl);

	pss_ctl &= ~(__PSS_LMEM_INIT_DONE | __PSS_LMEM_INIT_EN);
	bfa_reg_write(ioc->bfa_regs.pss_ctl_reg, pss_ctl);
}

static void
bfa_ioc_lpu_start(struct bfa_ioc_s *ioc)
{
	uint32_t        pss_ctl;

    /**
     * Take processor out of reset.
     */
	pss_ctl = bfa_reg_read(ioc->bfa_regs.pss_ctl_reg);
	pss_ctl &= ~__PSS_LPU0_RESET;

	bfa_reg_write(ioc->bfa_regs.pss_ctl_reg, pss_ctl);
}

/**
 * @brief
 * Return true if current running version is same as driver version.
 */
static bfa_boolean_t
bfa_ioc_fwver_check(struct bfa_ioc_s *ioc)
{
	bfa_t	*bfa = ioc->bfa;
	uint32_t        pgnum, pgoff;
	uint32_t        loff = 0;
	int             i;
	bfi_ioc_image_hdr_t fwhdr, *drv_fwhdr;
	uint32_t	*fwsig = (uint32_t *) &fwhdr;

	pgnum = PSS_SMEM_PGNUM(loff);
	pgoff = PSS_SMEM_PGOFF(loff);
	bfa_reg_write(ioc->bfa_regs.host_page_num_fn, pgnum);

	for (i = 0; i < (sizeof(bfi_ioc_image_hdr_t) / sizeof(uint32_t));
	     i++) {
		fwsig[i] =
			bfa_mem_read(ioc->bfa_regs.smem_page_start, loff);
		loff += sizeof(uint32_t);
	}

	bfa_trc(bfa, fwhdr.signature);
	bfa_trc(bfa, fwhdr.cksum);
	bfa_trc(bfa, fwhdr.exec);
	bfa_trc(bfa, fwhdr.param);

	drv_fwhdr = (bfi_ioc_image_hdr_t *) bfi_image;

	bfa_trc(bfa, drv_fwhdr->signature);
	bfa_trc(bfa, drv_fwhdr->cksum);
	bfa_trc(bfa, drv_fwhdr->exec);
	bfa_trc(bfa, drv_fwhdr->param);

	return (((fwhdr.signature == drv_fwhdr->signature))
		&& (fwhdr.cksum == drv_fwhdr->cksum)
		&& (fwhdr.exec == drv_fwhdr->exec)
		);
}

/**
 * @img ioc_init_logic.jpg
 */
static void
bfa_ioc_hwinit(struct bfa_ioc_s *ioc, bfa_boolean_t force)
{
	bfi_ioc_state_t ioc_fwstate;

	ioc_fwstate = bfa_reg_read(ioc->bfa_regs.ioc_fwstate);

	if (force)
		ioc_fwstate = BFI_IOC_UNINIT;

	bfa_trc(ioc->bfa, ioc_fwstate);

	/**
	 * If hardware initialization is in progress (initialized by other IOC),
	 * just wait for an initialization completion interrupt.
	 */
	if ((ioc_fwstate == BFI_IOC_INITING) &&
	    bfa_ioc_fwver_check(ioc)) {
		bfa_trc(ioc->bfa, ioc_fwstate);
		return;
	}

	/**
	 * If IOC function is disabled and firmware version is same,
	 * just re-enable IOC.
	 */
	if (((ioc_fwstate == BFI_IOC_DISABLED) ||
	     (ioc_fwstate == BFI_IOC_HWINIT))
	    && bfa_ioc_fwver_check(ioc)) {
		bfa_trc(ioc->bfa, ioc_fwstate);
		bfa_ioc_send_enable(ioc);
		return;
	}

	/**
	 * Initialize the h/w for any other states.
	 */
	bfa_ioc_boot(ioc->bfa, BFI_BOOT_TYPE_NORMAL,
			 BFI_BOOT_PARAM_VAL_DEFAULT);
}

static void
bfa_ioc_timeout(void *ioc_arg)
{
	struct bfa_ioc_s  *ioc = (struct bfa_ioc_s *) ioc_arg;

	bfa_trc(ioc->bfa, 0);
	bfa_sm_send_event(ioc, IOC_E_TIMEOUT);
}

void
bfa_ioc_mbox_send(struct bfa_ioc_s *ioc, void *ioc_msg, int len)
{
	uint32_t       *msgp = (uint32_t *) ioc_msg;
	uint32_t        i;

	bfa_trc(ioc->bfa, len);

	bfa_assert(len <= BFI_IOC_MSGLEN_MAX);
	len = BFI_IOC_MSGLEN_MAX;

	/*
	 * first write msg to mailbox registers
	 */
	for (i = 0; i < len / sizeof(uint32_t); i++)
		bfa_reg_write(ioc->bfa_regs.hfn_mbox + i * sizeof(uint32_t),
							bfa_os_wtole(msgp[i]));

	/*
	 * write 1 to mailbox CMD to trigger LPU event
	 */
	bfa_reg_write(ioc->bfa_regs.hfn_mbox_cmd, 1);
}

static void
bfa_ioc_send_enable(struct bfa_ioc_s *ioc)
{
	bfi_ioc_enable_req_t enable_req;

	bfi_h2i_set(enable_req.mh, BFI_MC_IOC, BFI_IOC_H2I_ENABLE_REQ,
			bfa_ioc_lpuid(ioc->bfa));
	bfa_ioc_mbox_send(ioc, &enable_req, sizeof(bfi_ioc_enable_req_t));
}

static void
bfa_ioc_send_disable(struct bfa_ioc_s *ioc)
{
	bfi_ioc_enable_req_t disable_req;

	bfi_h2i_set(disable_req.mh, BFI_MC_IOC, BFI_IOC_H2I_DISABLE_REQ,
			bfa_ioc_lpuid(ioc->bfa));
	bfa_ioc_mbox_send(ioc, &disable_req, sizeof(bfi_ioc_disable_req_t));
}

/**
 * Firmware failure detected. Start recovery actions.
 */
static void
bfa_ioc_recover(struct bfa_ioc_s *ioc)
{
	bfa_stats(ioc, ioc_hbfails);
	bfa_debug_ioc_save(ioc->bfa);
	bfa_sm_send_event(ioc, IOC_E_HBFAIL);
}

static void
bfa_ioc_hb_check(void *cbarg)
{
	struct bfa_ioc_s  *ioc = cbarg;
	uint32_t        hb_count;

	hb_count = bfa_reg_read(ioc->bfa_regs.heartbeat);
	if (ioc->hb_count == hb_count) {
		ioc->hb_fail++;
	} else {
		ioc->hb_count = hb_count;
		ioc->hb_fail = 0;
	}

	if (ioc->hb_fail >= BFA_IOC_HB_FAIL_MAX) {
		bfa_log(ioc->bfa->logm, BFA_LOG_HAL_HEARTBEAT_FAILURE,
			hb_count);
		ioc->hb_fail = 0;
		bfa_ioc_recover(ioc);
		return;
	}

	bfa_timer_start(ioc->bfa, &ioc->ioc_timer, bfa_ioc_hb_check,
			    ioc, BFA_IOC_HB_TOV);
}

static void
bfa_ioc_hb_monitor(struct bfa_ioc_s *ioc)
{
	ioc->hb_fail = 0;
	ioc->hb_count = bfa_reg_read(ioc->bfa_regs.heartbeat);
	bfa_timer_start(ioc->bfa, &ioc->ioc_timer, bfa_ioc_hb_check,
			    ioc, BFA_IOC_HB_TOV);
}

static void
bfa_ioc_hb_stop(struct bfa_ioc_s *ioc)
{
	bfa_timer_stop(&ioc->ioc_timer);
}

void
bfa_ioc_reg_init(struct bfa_ioc_s *ioc, bfa_pcidev_t *pci_info)
{
	bfa_os_addr_t   rb;
	int             fn = pci_info->pci_func;

	rb = pci_info->pci_bar_kva;

	if (fn == 0) {
		ioc->bfa_regs.intr_status = (rb + HOSTFN0_INT_STATUS);
		ioc->bfa_regs.intr_mask = (rb + HOSTFN0_INT_MSK);
		ioc->bfa_regs.hfn_mbox_cmd = (rb + HOSTFN0_LPU0_CMD_STAT);
		ioc->bfa_regs.hfn_mbox = (rb + HOSTFN0_LPU0_MBOX_0);
		ioc->bfa_regs.lpu_mbox_cmd = (rb + LPU0_HOSTFN0_CMD_STAT);
		ioc->bfa_regs.lpu_mbox = (rb + LPU0_HOSTFN0_MBOX_0);
		ioc->bfa_regs.lpu_ioc_status = (rb + LPU0_HOSTFN0_CMD_STAT);
		ioc->bfa_regs.host_page_num_fn = (rb + HOST_PAGE_NUM_FN0);
		ioc->bfa_regs.heartbeat = (rb + BFA_IOC0_HBEAT_REG);
		ioc->bfa_regs.ioc_fwstate = (rb + BFA_IOC0_STATE_REG);
	} else {
		ioc->bfa_regs.intr_status = (rb + HOSTFN1_INT_STATUS);
		ioc->bfa_regs.intr_mask = (rb + HOSTFN1_INT_MSK);
		ioc->bfa_regs.hfn_mbox_cmd = (rb + HOSTFN1_LPU1_CMD_STAT);
		ioc->bfa_regs.hfn_mbox = (rb + HOSTFN1_LPU1_MBOX_0);
		ioc->bfa_regs.lpu_mbox_cmd = (rb + LPU1_HOSTFN1_CMD_STAT);
		ioc->bfa_regs.lpu_mbox = (rb + LPU1_HOSTFN1_MBOX_0);
		ioc->bfa_regs.lpu_ioc_status = (rb + LPU1_HOSTFN1_CMD_STAT);
		ioc->bfa_regs.host_page_num_fn = (rb + HOST_PAGE_NUM_FN1);
		ioc->bfa_regs.heartbeat = (rb + BFA_IOC1_HBEAT_REG);
		ioc->bfa_regs.ioc_fwstate = (rb + BFA_IOC1_STATE_REG);
	}

	/*
	 * PSS control registers
	 */
	ioc->bfa_regs.pss_ctl_reg = (rb + PSS_CTL_REG);
	ioc->bfa_regs.app_pll_400_ctl_reg = (rb + APP_PLL_400_CTL_REG);
	ioc->bfa_regs.app_pll_212_ctl_reg = (rb + APP_PLL_212_CTL_REG);

	/*
	 * host semaphore registers
	 */
	ioc->bfa_regs.host_sem0_reg = (rb + HOST_SEM0_REG);
	ioc->bfa_regs.host_sem1_reg = (rb + HOST_SEM1_REG);
	ioc->bfa_regs.host_sem2_reg = (rb + HOST_SEM2_REG);
	ioc->bfa_regs.host_sem3_reg = (rb + HOST_SEM3_REG);

	/**
	 * sram memory access
	 */
	ioc->bfa_regs.smem_page_start = (rb + PSS_SMEM_PAGE_START);
}

static void
bfa_ioc_callback(struct bfa_ioc_s *ioc)
{
	if (ioc->init_cb) {
		ioc->init_cb = BFA_FALSE;
		bfa_cb_queue(ioc->bfa, &ioc->init_hcb_qe, bfa_iocfc_init_cb,
			     ioc);
	}

	if (ioc->disable_cb) {
		ioc->disable_cb = BFA_FALSE;
		bfa_cb_queue(ioc->bfa, &ioc->dis_hcb_qe, bfa_iocfc_disable_cb,
			     ioc);
	}

	if (ioc->stop_cb) {
		ioc->stop_cb = BFA_FALSE;
		bfa_cb_queue(ioc->bfa, &ioc->stop_hcb_qe, bfa_iocfc_stop_cb,
			     ioc);
	}
}

#define BFA_MFG_NAME "Brocade"
void
bfa_ioc_get_adapter_attr(bfa_t *bfa, bfa_adapter_attr_t *ad_attr)
{
	bfi_iocfc_cfgrsp_t *cfginfo;
	char            model[BFA_ADAPTER_MODEL_NAME_LEN];

	cfginfo = bfa->ioc.cfgrsp;
	bfa_os_memcpy((void *)&ad_attr->serial_num,
		      (void *)cfginfo->brcd_serialnum,
		      BFA_ADAPTER_SERIAL_NUM_LEN);

	bfa_os_memcpy(&ad_attr->fw_ver, cfginfo->fw_version, BFA_VERSION_LEN);
	bfa_os_memcpy(&ad_attr->optrom_ver, cfginfo->optrom_version, BFA_VERSION_LEN);
	bfa_os_memcpy(&ad_attr->manufacturer, BFA_MFG_NAME,
		      BFA_ADAPTER_MFG_NAME_LEN);
	ad_attr->nports = BFI_ADAPTER_GETP(NPORTS, cfginfo->adapter_prop);
	ad_attr->max_speed = BFI_ADAPTER_GETP(SPEED, cfginfo->adapter_prop);

	/**
	 * model name
	 */
	strcpy(model, "Brocade-??5");
	model[8] = '0' + BFI_ADAPTER_GETP(SPEED, cfginfo->adapter_prop);
	model[9] = '0' + ad_attr->nports;

	if (BFI_ADAPTER_IS_SPECIAL(cfginfo->adapter_prop))
		ad_attr->prototype = 1;
	else
		ad_attr->prototype = 0;

	bfa_os_memcpy(&ad_attr->model, model, BFA_ADAPTER_MODEL_NAME_LEN);
	bfa_os_memcpy(&ad_attr->model_descr, &ad_attr->model,
		      BFA_ADAPTER_MODEL_NAME_LEN);

	ad_attr->pwwn = bfa_iocfc_get_pwwn(bfa);

	ad_attr->pcie_gen = cfginfo->pcie_gen;
	ad_attr->pcie_lanes = cfginfo->pcie_lanes;
	ad_attr->pcie_lanes_orig = cfginfo->pcie_lanes_orig;
	ad_attr->asic_rev = cfginfo->asic_rev;

	ad_attr->hw_ver[0] = 'R';
	ad_attr->hw_ver[1] = 'e';
	ad_attr->hw_ver[2] = 'v';
	ad_attr->hw_ver[3] = '-';
	ad_attr->hw_ver[4] = cfginfo->asic_rev;
	ad_attr->hw_ver[5] = '\0';

	/**
	 * @todo
	 * ad_attr->node_symname
	 */
}

/**
 * @brief
 *      Initiate a full firmware download.
 */
static void
bfa_ioc_download_fw(bfa_t *bfa, uint32_t boot_type,
			uint32_t boot_param)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	uint32_t       *fwimg = bfi_image;
	uint32_t        pgnum, pgoff;
	uint32_t        loff = 0;
	uint32_t        i;

	/**
	 * Initialize LMEM first before code download
	 */
	bfa_ioc_lmem_init(ioc);

	/**
	 * Flash based firmware boot
	 */
	bfa_trc(bfa, bfi_image_size);
	if (bfi_image_size < BFA_IOC_FWIMG_MINSZ)
		boot_type = BFI_BOOT_TYPE_FLASH;
	fwimg[BFI_BOOT_TYPE_OFF / sizeof(uint32_t)] =
		bfa_os_swap32(boot_type);
	fwimg[BFI_BOOT_PARAM_OFF / sizeof(uint32_t)] =
		bfa_os_swap32(boot_param);

	pgnum = PSS_SMEM_PGNUM(loff);
	pgoff = PSS_SMEM_PGOFF(loff);

	bfa_reg_write(ioc->bfa_regs.host_page_num_fn, pgnum);

	for (i = 0; i < bfi_image_size; i++) {
	/**
	  * write smem
	  */
		bfa_mem_write(ioc->bfa_regs.smem_page_start, loff,
				  fwimg[i]);
		loff += sizeof(uint32_t);

	/**
	  * handle page offset wrap around
	  */
		loff = PSS_SMEM_PGOFF(loff);
		if (loff == 0) {
			pgnum++;
			bfa_reg_write(ioc->bfa_regs.host_page_num_fn,
					  pgnum);
		}
	}

	bfa_reg_write(ioc->bfa_regs.host_page_num_fn, PSS_SMEM_PGNUM(0));
}

static void
bfa_ioc_reset(struct bfa_ioc_s *ioc, bfa_boolean_t force)
{
	bfa_iocfc_reset_queues(ioc);
	bfa_ioc_hwinit(ioc, force);
	if (ioc->ioc_started)
		bfa_isr_enable(ioc->bfa);
}

/**
 * @}
 */

/**
 * @ig hal_ioc_public
 * @{
 */

bfa_status_t
bfa_ioc_pll_init(bfa_t *bfa)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	bfa_os_addr_t   rb;

	rb = ioc->pci_info.pci_bar_kva;
	bfa_reg_write((rb + BFA_IOC0_STATE_REG), BFI_IOC_UNINIT);
	bfa_reg_write((rb + BFA_IOC1_STATE_REG), BFI_IOC_UNINIT);

	bfa_reg_write((rb + HOSTFN0_INT_MSK), 0xffffffffU);
	bfa_reg_write((rb + HOSTFN1_INT_MSK), 0xffffffffU);

	bfa_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_LOGIC_SOFT_RESET);
	bfa_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_BYPASS |
			  __APP_PLL_212_LOGIC_SOFT_RESET);
	bfa_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_LOGIC_SOFT_RESET);
	bfa_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_BYPASS |
			  __APP_PLL_400_LOGIC_SOFT_RESET);
	bfa_os_udelay(2);
	bfa_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_LOGIC_SOFT_RESET);
	bfa_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_LOGIC_SOFT_RESET);

	bfa_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_ENABLE | __APP_PLL_212_LRESETN |
			  __APP_PLL_212_P0_1(3U) | __APP_PLL_212_JITLMT0_1(3U) |
			  __APP_PLL_212_CNTLMT0_1(3U) |
			  __APP_PLL_212_LOGIC_SOFT_RESET);
	bfa_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_ENABLE | __APP_PLL_400_LRESETN |
			  __APP_PLL_400_RSEL200500 | __APP_PLL_400_P0_1(3U) |
			  __APP_PLL_400_JITLMT0_1(3U) |
			  __APP_PLL_400_CNTLMT0_1(3U) |
			  __APP_PLL_400_LOGIC_SOFT_RESET);

	/**
	 * Wait for PLLs to lock.
	 */
	bfa_os_udelay(2000);
	bfa_reg_write((rb + HOSTFN0_INT_STATUS), 0xffffffffU);
	bfa_reg_write((rb + HOSTFN1_INT_STATUS), 0xffffffffU);

	bfa_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_ENABLE | __APP_PLL_212_LRESETN |
			  __APP_PLL_212_P0_1(3U) | __APP_PLL_212_JITLMT0_1(3U) |
			  __APP_PLL_212_CNTLMT0_1(3U));
	bfa_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_ENABLE | __APP_PLL_400_LRESETN |
			  __APP_PLL_400_RSEL200500 | __APP_PLL_400_P0_1(3U) |
			  __APP_PLL_400_JITLMT0_1(3U) |
			  __APP_PLL_400_CNTLMT0_1(3U));

	return BFA_STATUS_OK;
}

void
bfa_ioc_cfg_complete(struct bfa_ioc_s *ioc)
{
	bfa_sm_send_event(ioc, IOC_E_FWRSP_CFG);
}

/**
 * Interface used by diag module to do firmware boot with memory test
 * as the entry vector.
 */
void
bfa_ioc_boot(bfa_t *bfa, uint32_t boot_type, uint32_t boot_param)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	bfa_os_addr_t   rb;

	bfa_stats(ioc, ioc_boots);

	if (bfa_ioc_pll_init(ioc->bfa) != BFA_STATUS_OK)
		return;

	/**
	 * Initialize IOC state of all functions on a chip reset.
	 */
	rb = ioc->pci_info.pci_bar_kva;
	if (boot_param == BFI_BOOT_TYPE_MEMTEST) {
		bfa_reg_write((rb + BFA_IOC0_STATE_REG), BFI_IOC_MEMTEST);
		bfa_reg_write((rb + BFA_IOC1_STATE_REG), BFI_IOC_MEMTEST);
	} else {
		bfa_reg_write((rb + BFA_IOC0_STATE_REG), BFI_IOC_INITING);
		bfa_reg_write((rb + BFA_IOC1_STATE_REG), BFI_IOC_INITING);
	}

	bfa_ioc_download_fw(ioc->bfa, boot_type, boot_param);

	/**
	 * Enable interrupts just before starting LPU
	 */
	bfa_isr_enable(ioc->bfa);
	bfa_ioc_lpu_start(ioc);
}

void
bfa_ioc_get_attr(bfa_t *bfa, bfa_ioc_attr_t *ioc_attr)
{
	struct bfa_ioc_s      *ioc = &bfa->ioc;

	bfa_os_memset((void *)ioc_attr, 0, sizeof(bfa_ioc_attr_t));

	ioc_attr->state = bfa_sm_to_state(ioc_sm_table, bfa->ioc.sm);
	ioc_attr->fwcfg = bfa->ioc.cfg;
	ioc_attr->intr_attr = ioc->cfginfo->intr_attr;
	bfa_ioc_get_adapter_attr(bfa, &ioc_attr->adapter_attr);

	ioc_attr->chip_attr.chip_rev[0] = 'R';
	ioc_attr->chip_attr.chip_rev[1] = 'e';
	ioc_attr->chip_attr.chip_rev[2] = 'v';
	ioc_attr->chip_attr.chip_rev[3] = '-';
	ioc_attr->chip_attr.chip_rev[4] = ioc_attr->adapter_attr.asic_rev;
	ioc_attr->chip_attr.chip_rev[5] = '\0';
}

void
bfa_ioc_pm_ctl(bfa_t *bfa, bfa_pm_cmd_t cmd)
{
}

/**
 * Enable/disable IOC failure auto recovery.
 */
void
bfa_ioc_auto_recover(bfa_boolean_t auto_recover)
{
	bfa_auto_recover = auto_recover;
}

/**
 * @brief
 */
bfa_boolean_t
bfa_ioc_is_operational(bfa_t *bfa)
{
	return bfa_sm_cmp_state(&bfa->ioc, bfa_ioc_sm_op);
}

#define bfa_ioc_state_disabled(__sm)			\
	(((__sm) == BFI_IOC_UNINIT) ||		\
	 ((__sm) == BFI_IOC_INITING) ||		\
	 ((__sm) == BFI_IOC_HWINIT) ||		\
	 ((__sm) == BFI_IOC_DISABLED) ||		\
	 ((__sm) == BFI_IOC_HBFAIL) ||		\
	 ((__sm) == BFI_IOC_CFG_DISABLED))

/**
 * Check if adapter is disabled -- both IOCs should be in a disabled
 * state.
 */
bfa_boolean_t
bfa_ioc_adapter_is_disabled(bfa_t *bfa)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;
	uint32_t        ioc_state;
	bfa_os_addr_t   rb = ioc->pci_info.pci_bar_kva;

	if (!bfa_sm_cmp_state(ioc, bfa_ioc_sm_disabled))
		return BFA_FALSE;

	ioc_state = bfa_reg_read(rb + BFA_IOC0_STATE_REG);
	if (!bfa_ioc_state_disabled(ioc_state))
		return BFA_FALSE;

	ioc_state = bfa_reg_read(rb + BFA_IOC1_STATE_REG);
	if (!bfa_ioc_state_disabled(ioc_state))
		return BFA_FALSE;

	return BFA_TRUE;
}

void
bfa_ioc_msgget(struct bfa_ioc_s *ioc, bfi_ioc_i2h_msg_t *msg)
{
	uint32_t	*msgp = msg->mboxmsg;
	uint32_t	r32;
	int		i;

	/**
	 * read the MBOX msg
	 */
	for (i = 0; i < (sizeof(bfi_ioc_i2h_msg_t) / sizeof(uint32_t)); i++) {
		r32 = bfa_reg_read(ioc->bfa_regs.lpu_mbox +
						       i * sizeof(uint32_t));
		msgp[i] = bfa_os_htonl(r32);
	}

	/**
	 * turn off mailbox interrupt by clearing mailbox status
	 */
	bfa_reg_write(ioc->bfa_regs.lpu_mbox_cmd, 1);
	r32 = bfa_reg_read(ioc->bfa_regs.lpu_mbox_cmd);
}

void
bfa_ioc_isr(struct bfa_ioc_s *ioc, bfi_ioc_i2h_msg_t *msg)
{
	bfa_stats(ioc, ioc_isrs);
	bfa_trc(ioc->bfa, msg->mh.msg_id);

	switch (msg->mh.msg_id) {
	case BFI_IOC_I2H_READY_EVENT:
		bfa_sm_send_event(ioc, IOC_E_FWREADY);
		break;

	case BFI_IOC_I2H_ENABLE_REPLY:
		bfa_sm_send_event(ioc, IOC_E_FWRSP_ENABLE);
		break;

	case BFI_IOC_I2H_DISABLE_REPLY:
		bfa_sm_send_event(ioc, IOC_E_FWRSP_DISABLE);
		break;

	default:
		bfa_assert(0);
	}
}

void
bfa_ioc_attach(struct bfa_ioc_s *ioc, bfa_pcidev_t *pcidev)
{
	bfa_os_memset((void *)ioc, 0, sizeof(struct bfa_ioc_s));
	bfa_ioc_reg_init(ioc, pcidev);
	bfa_sm_set_state(ioc, bfa_ioc_sm_reset);
}

void
bfa_ioc_init(struct bfa_ioc_s *ioc)
{
	ioc->init_cb = BFA_TRUE;
	bfa_sm_send_event(ioc, IOC_E_INIT);
}

void
bfa_ioc_start(struct bfa_ioc_s *ioc)
{
	bfa_stats(ioc, ioc_starts);

	bfa_assert(ioc->ioc_started == BFA_FALSE);
	ioc->ioc_started = BFA_TRUE;
	bfa_sm_send_event(ioc, IOC_E_ENABLE);
}

void
bfa_ioc_enable(struct bfa_ioc_s *ioc)
{
	bfa_stats(ioc, ioc_enables);
	ioc->ioc_started = BFA_TRUE;
	bfa_sm_send_event(ioc, IOC_E_ENABLE);
}

void
bfa_ioc_stop(struct bfa_ioc_s *ioc)
{
	bfa_stats(ioc, ioc_stops);

	ioc->ioc_started = BFA_FALSE;
	ioc->stop_cb = BFA_TRUE;
	bfa_sm_send_event(ioc, IOC_E_DISABLE);
}

void
bfa_ioc_disable(struct bfa_ioc_s *ioc)
{
	bfa_stats(ioc, ioc_disables);
	ioc->ioc_started = BFA_FALSE;
	ioc->disable_cb = BFA_TRUE;
	bfa_sm_send_event(ioc, IOC_E_DISABLE);
}

/**
 * @}
 */
