/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
#include "bfa_ioc_priv.h"
#include "bfa_callback_priv.h"
#include <cs/bfa_debug.h>
#include <bfa_priv.h>
#include <log/bfa_log_hal.h>
#include <bfi/bfi_boot.h>
#include <cs/bfa_plog.h>
#include <cs/bfa_stats.h>
#include <aen/bfal_aen_ioc.h>

BFA_TRC_FILE(HAL, IOCFC);

/**
 * IOC local definitions
 */
#define BFA_IOC_TOV		5000	/* msecs */
#define BFA_IOC_HB_TOV      2000	/* msecs */
#define BFA_IOC_HB_FAIL_MAX 10
#define BFA_IOC_HWINIT_MAX	2
#define BFA_IOC_FWIMG_MINSZ     (16 * 1024)

#define bfa_ioc_timer_start(__ioc)					\
	bfa_timer_start((__ioc)->bfa, &(__ioc)->ioc_timer,          \
			bfa_ioc_timeout, (__ioc), BFA_IOC_TOV)
#define bfa_ioc_timer_stop(__ioc)   bfa_timer_stop(&(__ioc)->ioc_timer)

extern bfa_boolean_t bfa_auto_recover;

/*
 * forward declarations
 */
static void     bfa_iocfc_stats_clear(void *cbarg);
static void     bfa_iocfc_stats_swap(bfa_fw_stats_t *d,
				       bfa_fw_stats_t *s);
static void     bfa_iocfc_stats_clr_cb(void *cbarg, bfa_boolean_t complete);
static void     bfa_iocfc_stats_clr_timeout(void *cbarg);
static void     bfa_iocfc_stats_cb(void *cbarg, bfa_boolean_t complete);
static void     bfa_iocfc_stats_timeout(void *cbarg);

/**
 * @dg hal_ioc_pvt HAL IOC private functions
 * @{
 */

static void
bfa_iocfc_cqs_sz(bfa_ioc_cfg_t *cfg, uint32_t *dm_len)
{
	int             i, per_reqq_sz, per_rspq_sz;

	per_reqq_sz = BFA_ROUNDUP((cfg->num_reqq_elems * BFI_LMSG_SZ),
							BFA_DMA_ALIGN_SZ);
	per_rspq_sz = BFA_ROUNDUP((cfg->num_rspq_elems * BFI_LMSG_SZ),
							BFA_DMA_ALIGN_SZ);

	/*
	 * Calculate CQ size
	 */
	for (i = 0; i < cfg->fwcfg.num_cqs; i++) {
		*dm_len = *dm_len + per_reqq_sz;
		*dm_len = *dm_len + per_rspq_sz;
	}

	/*
	 * Calculate Shadow CI/PI size
	 */
	for (i = 0; i < cfg->fwcfg.num_cqs; i++)
		*dm_len += (2 * BFA_CACHELINE_SZ);
}

static void
bfa_iocfc_fw_cfg_sz(bfa_ioc_cfg_t *cfg, uint32_t *dm_len)
{
	*dm_len +=
		BFA_ROUNDUP(sizeof(bfi_iocfc_cfg_t), BFA_CACHELINE_SZ);
	*dm_len +=
		BFA_ROUNDUP(sizeof(bfi_iocfc_cfgrsp_t),
			    BFA_CACHELINE_SZ);
	*dm_len += BFA_ROUNDUP(sizeof(bfa_fw_stats_t), BFA_CACHELINE_SZ);
}

/**
 * Use the Mailbox interface to send BFI_IOCFC_H2I_CFG_REQ
 */
void
bfa_iocfc_send_cfg(struct bfa_ioc_s *ioc)
{
	int             i;
	bfi_iocfc_cfg_req_t cfg_req;
	bfi_iocfc_cfg_t *cfg_info = ioc->cfginfo;
	bfa_ioc_cfg_t  *hal_cfg = &ioc->cfg;

	bfa_assert(hal_cfg->fwcfg.num_cqs <= BFI_IOC_MAX_CQS);
	bfa_trc(ioc->bfa, hal_cfg->fwcfg.num_cqs);

	ioc->cfgdone = BFA_FALSE;

	/**
	 * initialize IOC configuration info
	 */
	cfg_info->endian_sig = BFI_IOC_ENDIAN_SIG;
	cfg_info->num_cqs = hal_cfg->fwcfg.num_cqs;

	bfa_dma_be_addr_set(cfg_info->cfgrsp_addr, ioc->cfgrsp_dma.pa);
	bfa_dma_be_addr_set(cfg_info->stats_addr, ioc->stats_pa);

	/**
	 * dma map REQ and RSP circular queues and shadow pointers
	 */
	for (i = 0; i < hal_cfg->fwcfg.num_cqs; i++) {
		bfa_dma_be_addr_set(cfg_info->req_cq_ba[i],
				       ioc->req_cq_ba[i].pa);
		bfa_dma_be_addr_set(cfg_info->req_shadow_ci[i],
				       ioc->req_cq_shadow_ci[i].pa);
		cfg_info->req_cq_elems[i] =
			bfa_os_htons(hal_cfg->num_reqq_elems);

		bfa_dma_be_addr_set(cfg_info->rsp_cq_ba[i],
				       ioc->rsp_cq_ba[i].pa);
		bfa_dma_be_addr_set(cfg_info->rsp_shadow_pi[i],
				       ioc->rsp_cq_shadow_pi[i].pa);
		cfg_info->rsp_cq_elems[i] =
			bfa_os_htons(hal_cfg->num_rspq_elems);
	}

	/**
	 * dma map IOC configuration itself
	 */
	bfi_h2i_set(cfg_req.mh, BFI_MC_IOCFC, BFI_IOCFC_H2I_CFG_REQ,
			bfa_ioc_lpuid(ioc->bfa));
	bfa_dma_be_addr_set(cfg_req.ioc_cfg_dma_addr, ioc->cfg_info.pa);

	bfa_ioc_mbox_send(ioc, &cfg_req, sizeof(bfi_iocfc_cfg_req_t));
}

static void
bfa_iocfc_reg_init(struct bfa_ioc_s *ioc)
{
	bfa_os_addr_t   rb;
	int             i, q;
	int             fn = ioc->pci_info.pci_func;

	rb = ioc->pci_info.pci_bar_kva;

	for (i = 0; i < BFI_IOC_MAX_CQS; i++) {
		/*
		 * CPE registers
		 */
		q = CPE_Q_NUM(fn, i);
		ioc->bfa_regs.cpe_q_pi[i] = (rb + CPE_Q_PI(q));
		ioc->bfa_regs.cpe_q_ci[i] = (rb + CPE_Q_CI(q));
		ioc->bfa_regs.cpe_q_depth[i] = (rb + CPE_Q_DEPTH(q));

		/*
		 * RME registers
		 */
		q = CPE_Q_NUM(fn, i);
		ioc->bfa_regs.rme_q_pi[i] = (rb + RME_Q_PI(q));
		ioc->bfa_regs.rme_q_ci[i] = (rb + RME_Q_CI(q));
		ioc->bfa_regs.rme_q_depth[i] = (rb + RME_Q_DEPTH(q));
	}
}

static void
bfa_iocfc_init_mem(bfa_t *bfa, void *bfad, bfa_ioc_cfg_t *cfg,
		    bfa_pcidev_t *pcidev)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;

	bfa->bfad = bfad;
	ioc->bfa = bfa;

	ioc->cfg = *cfg;
	ioc->pci_info = *pcidev;
	bfa_iocfc_reg_init(ioc);
}

static void
bfa_iocfc_mem_claim(bfa_t *bfa, bfa_ioc_cfg_t *cfg,
		      bfa_meminfo_t *meminfo)
{
	uint8_t        *dm_kva;
	uint64_t        dm_pa;
	int             i, per_reqq_sz, per_rspq_sz;
	struct bfa_ioc_s  *ioc = &bfa->ioc;

	dm_kva = bfa_meminfo_dma_virt(meminfo);
	dm_pa = bfa_meminfo_dma_phys(meminfo);

	/*
	 * Claim DMA-able memory for the request/response queues and for shadow
	 * ci/pi registers
	 */
	per_reqq_sz = BFA_ROUNDUP((cfg->num_reqq_elems * BFI_LMSG_SZ),
							BFA_DMA_ALIGN_SZ);
	per_rspq_sz = BFA_ROUNDUP((cfg->num_rspq_elems * BFI_LMSG_SZ),
							BFA_DMA_ALIGN_SZ);

	for (i = 0; i < cfg->fwcfg.num_cqs; i++) {
		ioc->req_cq_ba[i].kva = dm_kva;
		ioc->req_cq_ba[i].pa = dm_pa;
		bfa_os_memset(dm_kva, 0, per_reqq_sz);
		dm_kva += per_reqq_sz;
		dm_pa += per_reqq_sz;

		ioc->rsp_cq_ba[i].kva = dm_kva;
		ioc->rsp_cq_ba[i].pa = dm_pa;
		bfa_os_memset(dm_kva, 0, per_rspq_sz);
		dm_kva += per_rspq_sz;
		dm_pa += per_rspq_sz;
	}

	for (i = 0; i < cfg->fwcfg.num_cqs; i++) {
		ioc->req_cq_shadow_ci[i].kva = dm_kva;
		ioc->req_cq_shadow_ci[i].pa = dm_pa;
		dm_kva += BFA_CACHELINE_SZ;
		dm_pa += BFA_CACHELINE_SZ;

		ioc->rsp_cq_shadow_pi[i].kva = dm_kva;
		ioc->rsp_cq_shadow_pi[i].pa = dm_pa;
		dm_kva += BFA_CACHELINE_SZ;
		dm_pa += BFA_CACHELINE_SZ;
	}

	/*
	 * Claim DMA-able memory for the config info page
	 */
	bfa->ioc.cfg_info.kva = dm_kva;
	bfa->ioc.cfg_info.pa = dm_pa;
	bfa->ioc.cfginfo = (bfi_iocfc_cfg_t *) dm_kva;
	dm_kva += BFA_ROUNDUP(sizeof(bfi_iocfc_cfg_t), BFA_CACHELINE_SZ);
	dm_pa += BFA_ROUNDUP(sizeof(bfi_iocfc_cfg_t), BFA_CACHELINE_SZ);

	/*
	 * Claim DMA-able memory for the config response
	 */
	bfa->ioc.cfgrsp_dma.kva = dm_kva;
	bfa->ioc.cfgrsp_dma.pa = dm_pa;
	bfa->ioc.cfgrsp = (bfi_iocfc_cfgrsp_t *) dm_kva;

	dm_kva +=
		BFA_ROUNDUP(sizeof(bfi_iocfc_cfgrsp_t),
			    BFA_CACHELINE_SZ);
	dm_pa += BFA_ROUNDUP(sizeof(bfi_iocfc_cfgrsp_t),
			     BFA_CACHELINE_SZ);

	/*
	 * Claim DMA-able memory for ioc stats
	 */
	bfa->ioc.stats_kva = dm_kva;
	bfa->ioc.stats_pa = dm_pa;
	bfa->ioc.fw_stats = (bfa_fw_stats_t *) dm_kva;
	dm_kva += BFA_ROUNDUP(sizeof(bfa_fw_stats_t), BFA_CACHELINE_SZ);
	dm_pa += BFA_ROUNDUP(sizeof(bfa_fw_stats_t), BFA_CACHELINE_SZ);

	bfa_meminfo_dma_virt(meminfo) = dm_kva;
	bfa_meminfo_dma_phys(meminfo) = dm_pa;

	bfa_debug_mem_claim(bfa, cfg, meminfo, bfa_auto_recover);
}

/**
 * @brief
 * HAL submodules initialization completion notification.
 */
void
bfa_iocfc_initdone_submod(bfa_t *bfa)
{
	int             i;

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->initdone(bfa);
}

/**
 * @brief
 * Start HAL submodules.
 */
void
bfa_iocfc_start_submod(bfa_t *bfa)
{
	int             i;

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->start(bfa);
}

/**
 * @brief
 * Notify sub-modules of hardware failure.
 */
void
bfa_iocfc_iocdisable_submod(bfa_t *bfa)
{
	int             i;

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->iocdisable(bfa);
}

void
bfa_iocfc_init_cb(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioc_s  *ioc = cbarg;

	if (complete) {
		if (ioc->cfgdone)
			bfa_cb_init(ioc->bfa->bfad, BFA_STATUS_OK);
		else
			bfa_cb_init(ioc->bfa->bfad, BFA_STATUS_FAILED);
	}
}

void
bfa_iocfc_stop_cb(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioc_s  *ioc = cbarg;

	if (complete)
		bfa_cb_stop(ioc->bfa->bfad, BFA_STATUS_OK);
}

void
bfa_iocfc_disable_cb(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioc_s  *ioc = cbarg;

	if (complete)
		bfa_cb_ioc_disable(ioc->bfa->bfad);
}

/**
 * @brief
 * Update HAL configuration from firmware configuration.
 */
static void
bfa_iocfc_cfgrsp(struct bfa_ioc_s *ioc)
{
	bfi_iocfc_cfgrsp_t	*cfgrsp  = ioc->cfgrsp;
	bfa_ioc_fwcfg_t		*fwcfg   = &cfgrsp->fwcfg;
	bfi_iocfc_cfg_t		*cfginfo = ioc->cfginfo;

	cfgrsp->adapter_prop  = bfa_os_ntohl(cfgrsp->adapter_prop);
	cfgrsp->maxfrsize     = bfa_os_ntohs(cfgrsp->maxfrsize);

	fwcfg->num_cqs        = fwcfg->num_cqs;
	fwcfg->num_itnims     = bfa_os_ntohs(fwcfg->num_itnims);
	fwcfg->num_ioim_reqs  = bfa_os_ntohs(fwcfg->num_ioim_reqs);
	fwcfg->num_tskim_reqs = bfa_os_ntohs(fwcfg->num_tskim_reqs);
	fwcfg->num_fcxp_reqs  = bfa_os_ntohs(fwcfg->num_fcxp_reqs);
	fwcfg->num_uf_bufs    = bfa_os_ntohs(fwcfg->num_uf_bufs);
	fwcfg->num_rports     = bfa_os_ntohs(fwcfg->num_rports);
	fwcfg->num_itntms     = bfa_os_ntohs(fwcfg->num_itntms);

	cfginfo->intr_attr.coalesce = cfgrsp->intr_attr.coalesce;
	cfginfo->intr_attr.delay    = bfa_os_ntohs(cfgrsp->intr_attr.delay);
	cfginfo->intr_attr.latency  = bfa_os_ntohs(cfgrsp->intr_attr.latency);

	ioc->cfgdone = BFA_TRUE;
	bfa_ioc_cfg_complete(ioc);
}

static void
bfa_iocfc_stats_clear(void *cbarg)
{
	struct bfa_ioc_s  *ioc = (struct bfa_ioc_s *) cbarg;
	bfi_iocfc_stats_clr_req_t stats_req;

	bfa_timer_start(ioc->bfa, &ioc->stats_timer,
			    bfa_iocfc_stats_clr_timeout, ioc,
			    BFA_IOC_TOV);

	bfi_h2i_set(stats_req.mh, BFI_MC_IOCFC, BFI_IOCFC_H2I_CLEAR_STATS_REQ,
			bfa_ioc_lpuid(ioc->bfa));
	bfa_ioc_mbox_send(ioc, &stats_req, sizeof(bfi_iocfc_stats_clr_req_t));
}

static void
bfa_iocfc_stats_swap(bfa_fw_stats_t *d, bfa_fw_stats_t *s)
{
	uint32_t       *dip = (uint32_t *) d;
	uint32_t       *sip = (uint32_t *) s;
	int             i;

	for (i = 0; i < (sizeof(bfa_fw_stats_t) / sizeof(uint32_t)); i++)
		dip[i] = bfa_os_ntohl(sip[i]);
}

static void
bfa_iocfc_stats_clr_cb(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioc_s  *ioc = cbarg;

	if (complete) {
		bfa_os_memset(&ioc->stats, 0, sizeof(ioc->stats));
		ioc->stats_cbfn(ioc->stats_cbarg, ioc->stats_status);
	} else {
		ioc->stats_busy = BFA_FALSE;
		ioc->stats_status = BFA_STATUS_OK;
	}
}

static void
bfa_iocfc_stats_clr_timeout(void *cbarg)
{
	struct bfa_ioc_s  *ioc = (struct bfa_ioc_s *) cbarg;

	bfa_trc(ioc->bfa, 0);

	ioc->stats_status = BFA_STATUS_ETIMER;
	bfa_cb_queue(ioc->bfa, &ioc->stats_hcb_qe, bfa_iocfc_stats_clr_cb,
		      ioc);
}

static void
bfa_iocfc_stats_cb(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioc_s  *ioc = cbarg;

	if (complete) {
		if (ioc->stats_status == BFA_STATUS_OK) {
			bfa_os_memset(ioc->stats_ret, 0,
				sizeof(bfa_ioc_stats_t));
			bfa_iocfc_stats_swap(&ioc->stats_ret->fw_stats,
				ioc->fw_stats);
			ioc->stats_ret->drv_stats = ioc->stats;
		}
		ioc->stats_cbfn(ioc->stats_cbarg, ioc->stats_status);
	} else {
		ioc->stats_busy = BFA_FALSE;
		ioc->stats_status = BFA_STATUS_OK;
	}
}

static void
bfa_iocfc_stats_timeout(void *cbarg)
{
	struct bfa_ioc_s  *ioc = (struct bfa_ioc_s *) cbarg;

	bfa_stats(ioc, stats_tmos);
	bfa_trc(ioc->bfa, 0);

	ioc->stats_status = BFA_STATUS_ETIMER;
	bfa_cb_queue(ioc->bfa, &ioc->stats_hcb_qe, bfa_iocfc_stats_cb, ioc);
}

static void
bfa_iocfc_stats_query(struct bfa_ioc_s *ioc)
{
	bfi_iocfc_stats_get_req_t stats_req;

	bfa_timer_start(ioc->bfa, &ioc->stats_timer,
			    bfa_iocfc_stats_timeout, ioc, BFA_IOC_TOV);

	bfi_h2i_set(stats_req.mh, BFI_MC_IOCFC, BFI_IOCFC_H2I_GET_STATS_REQ,
			bfa_ioc_lpuid(ioc->bfa));
	bfa_ioc_mbox_send(ioc, &stats_req, sizeof(bfi_iocfc_stats_get_req_t));
}

void
bfa_iocfc_reset_queues(struct bfa_ioc_s *ioc)
{
	int             q;

	for (q = 0; q < BFI_IOC_MAX_CQS; q++) {
		bfa_rspq_ci(ioc->bfa, q) = 0;
		bfa_reqq_pi(ioc->bfa, q) = 0;
	}
}

/**
 * @}
 */

/**
 * @ig hal_ioc_public
 * @{
 */

void
bfa_iocfc_set_snsbase(bfa_t *bfa, uint64_t snsbase_pa)
{
	struct bfa_ioc_s      *ioc = &bfa->ioc;

	ioc->cfginfo->sense_buf_len = (BFI_IOIM_SNSLEN - 1);
	bfa_dma_be_addr_set(ioc->cfginfo->ioim_snsbase, snsbase_pa);
}

wwn_t
bfa_iocfc_get_pwwn(bfa_t *bfa)
{
	return bfa_wwn_get_naa1(bfa);
}

bfa_status_t
bfa_iocfc_get_stats(bfa_t *bfa, bfa_ioc_stats_t *stats,
		      bfa_cb_ioc_t cbfn, void *cbarg)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;

	if (ioc->stats_busy) {
		bfa_trc(bfa, ioc->stats_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	ioc->stats_busy = BFA_TRUE;
	ioc->stats_ret = stats;
	ioc->stats_cbfn = cbfn;
	ioc->stats_cbarg = cbarg;

	bfa_iocfc_stats_query(ioc);

	return (BFA_STATUS_OK);
}

bfa_status_t
bfa_iocfc_clear_stats(bfa_t *bfa, bfa_cb_ioc_t cbfn, void *cbarg)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;

	if (ioc->stats_busy) {
		bfa_trc(bfa, ioc->stats_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	ioc->stats_busy = BFA_TRUE;
	ioc->stats_cbfn = cbfn;
	ioc->stats_cbarg = cbarg;

	bfa_iocfc_stats_clear(ioc);
	return (BFA_STATUS_OK);
}

/**
 * @brief
 * Query IOC memory requirement information.
 */
void
bfa_iocfc_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *km_len, uint32_t *dm_len)
{
	bfa_iocfc_fw_cfg_sz(cfg, dm_len);
	bfa_iocfc_cqs_sz(cfg, dm_len);
	bfa_debug_fwsave_sz(cfg, km_len, bfa_auto_recover);
}

/**
 * @brief
 * Query IOC memory requirement information.
 */
void
bfa_iocfc_attach(bfa_t *bfa, void *bfad, bfa_ioc_cfg_t *cfg,
		   bfa_meminfo_t *meminfo, bfa_pcidev_t *pcidev)
{
	int             i;

	bfa_ioc_attach(&bfa->ioc, pcidev);
	bfa_iocfc_init_mem(bfa, bfad, cfg, pcidev);
	bfa_iocfc_mem_claim(bfa, cfg, meminfo);
	bfa_timer_init(bfa);

	bfa_q_init(&bfa->comp_q);
	for (i = 0; i < BFI_IOC_MAX_CQS; i++)
		bfa_q_init(&bfa->reqq_waitq[i]);

	bfa->ioc.hwinit_count = 0;
}

/**
 * @brief
 * Query IOC memory requirement information.
 */
void
bfa_iocfc_detach(bfa_t *bfa)
{
	/*
	 * no-op
	 */
}

/**
 * @brief
 * Query IOC memory requirement information.
 */
void
bfa_iocfc_init(bfa_t *bfa)
{
	bfa_ioc_init(&bfa->ioc);
}

/**
 * @brief
 * IOC start called from bfa_start(). Called to start IOC operations
 * at driver instantiation for this instance.
 */
void
bfa_iocfc_start(bfa_t *bfa)
{
	bfa_ioc_start(&bfa->ioc);
}

/**
 * @brief
 * IOC stop called from bfa_stop(). Called only when driver is unloaded
 * for this instance.
 */
void
bfa_iocfc_stop(bfa_t *bfa)
{
	bfa_ioc_stop(&bfa->ioc);
}

/**
 * Enable IOC after it is disabled.
 */
bfa_status_t
bfa_iocfc_enable(bfa_t *bfa)
{
	struct bfa_ioc_s  *ioc = &bfa->ioc;

	bfa_trc(bfa, ioc->ioc_started);

	if (ioc->ioc_started)
		return BFA_STATUS_IOC_ENABLED;

	bfa_plog_str(bfa->plog, BFA_PL_MID_HAL, BFA_PL_EID_MISC, 0,
		     "IOC Enable");

	bfa_ioc_enable(&bfa->ioc);
	return BFA_STATUS_OK;
}

void
bfa_iocfc_disable(bfa_t *bfa)
{
	bfa_plog_str(bfa->plog, BFA_PL_MID_HAL, BFA_PL_EID_MISC, 0,
		     "IOC Disable");
	bfa_ioc_disable(&bfa->ioc);

	bfa_trc(bfa, bfa->ioc.ioc_started);
}

bfa_status_t
bfa_iocfc_israttr_set(bfa_t *bfa, bfa_ioc_intr_attr_t *attr)
{
	struct bfa_ioc_s  	   *ioc = &bfa->ioc;
	bfi_iocfc_set_intr_req_t *m;

	ioc->cfginfo->intr_attr = *attr;
	if (!bfa_ioc_is_operational(bfa))
		return BFA_STATUS_OK;

	m = bfa_reqq_next(bfa, BFA_REQQ_IOC);
	if (!m)
		return BFA_STATUS_DEVBUSY;

	bfi_h2i_set(m->mh, BFI_MC_IOCFC, BFI_IOCFC_H2I_SET_INTR_REQ,
			bfa_ioc_lpuid(bfa));
	m->coalesce = attr->coalesce;
	m->delay    = bfa_os_htons(attr->delay);
	m->latency  = bfa_os_htons(attr->latency);

	bfa_trc(bfa, attr->delay);
	bfa_trc(bfa, attr->latency);

	bfa_reqq_produce(bfa, BFA_REQQ_IOC);
	return BFA_STATUS_OK;
}

void
bfa_iocfc_isr(struct bfa_ioc_s *ioc, bfi_iocfc_i2h_msg_t *msg)
{
	bfa_stats(ioc, ioc_isrs);
	bfa_trc(ioc->bfa, msg->mh.msg_id);

	switch (msg->mh.msg_id) {
	case BFI_IOCFC_I2H_CFG_REPLY:
		ioc->event_args.cfg_reply = &msg->cfg_reply;
		bfa_iocfc_cfgrsp(ioc);
		break;

	case BFI_IOCFC_I2H_GET_STATS_RSP:
		if (ioc->stats_busy == BFA_FALSE
		    || ioc->stats_status == BFA_STATUS_ETIMER)
			break;

		bfa_timer_stop(&ioc->stats_timer);
		ioc->stats_status = BFA_STATUS_OK;
		bfa_cb_queue(ioc->bfa, &ioc->stats_hcb_qe, bfa_iocfc_stats_cb,
			      ioc);
		break;
	case BFI_IOCFC_I2H_CLEAR_STATS_RSP:
		/*
		 * check for timer pop before processing the rsp
		 */
		if (ioc->stats_busy == BFA_FALSE
		    || ioc->stats_status == BFA_STATUS_ETIMER)
			break;

		bfa_timer_stop(&ioc->stats_timer);
		ioc->stats_status = BFA_STATUS_OK;
		bfa_cb_queue(ioc->bfa, &ioc->stats_hcb_qe,
			      bfa_iocfc_stats_clr_cb, ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */
