/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_uf.c HAL unsolicited frame receive implementation
 */

#include <bfa.h>
#include <bfa_svc.h>
#include <bfi/bfi_uf.h>
#include <cs/bfa_debug.h>

BFA_TRC_FILE(HAL, UF);
BFA_MODULE(uf);

/*
 *****************************************************************************
 * Internal functions
 *****************************************************************************
 */
static void
__bfa_cb_uf_recv(void *cbarg, bfa_boolean_t complete)
{
	bfa_uf_t   *uf = cbarg;
	bfa_uf_mod_t *ufm = BFA_UF_MOD(uf->bfa);

	if (complete)
		ufm->ufrecv(ufm->cbarg, uf);
}

static void
claim_uf_pbs(bfa_uf_mod_t *ufm, bfa_meminfo_t *mi)
{
	uint32_t        uf_pb_tot_sz;

	ufm->uf_pbs_kva = (bfa_uf_buf_t *) bfa_meminfo_dma_virt(mi);
	ufm->uf_pbs_pa = bfa_meminfo_dma_phys(mi);
	uf_pb_tot_sz = BFA_ROUNDUP((sizeof(bfa_uf_buf_t) * ufm->num_ufs),
							BFA_DMA_ALIGN_SZ);

	bfa_meminfo_dma_virt(mi) += uf_pb_tot_sz;
	bfa_meminfo_dma_phys(mi) += uf_pb_tot_sz;

	bfa_os_memset((void *)ufm->uf_pbs_kva, 0, uf_pb_tot_sz);
}

static void
claim_uf_post_msgs(bfa_uf_mod_t *ufm, bfa_meminfo_t *mi)
{
	bfi_uf_buf_post_t *uf_bp_msg;
	bfi_sge_t      *sge;
	bfi_addr_t      sga_zero = { {0} };
	uint16_t        i;
	uint16_t        buf_len;

	ufm->uf_buf_posts = (bfi_uf_buf_post_t *) bfa_meminfo_kva(mi);
	uf_bp_msg = ufm->uf_buf_posts;

	for (i = 0, uf_bp_msg = ufm->uf_buf_posts; i < ufm->num_ufs;
	     i++, uf_bp_msg++) {
		bfa_os_memset(uf_bp_msg, 0, sizeof(bfi_uf_buf_post_t));

		uf_bp_msg->buf_tag = i;
		buf_len = sizeof(bfa_uf_buf_t);
		uf_bp_msg->buf_len = bfa_os_htons(buf_len);
		bfi_h2i_set(uf_bp_msg->mh, BFI_MC_UF,
				BFI_UF_H2I_BUF_POST,
				ufm->bfa->ioc.pci_info.pci_func);

		sge = uf_bp_msg->sge;
		sge[0].sg_len = buf_len;
		sge[0].flags = BFI_SGE_DATA_LAST;
		bfa_dma_addr_set(sge[0].sga, ufm_pbs_pa(ufm, i));
		bfa_sge_to_be(sge);

		sge[1].sg_len = buf_len;
		sge[1].flags = BFI_SGE_PGDLEN;
		sge[1].sga = sga_zero;
		bfa_sge_to_be(&sge[1]);
	}

	/**
	 * advance pointer beyond consumed memory
	 */
	bfa_meminfo_kva(mi) = (uint8_t *) uf_bp_msg;
}

static void
claim_ufs(bfa_uf_mod_t *ufm, bfa_meminfo_t *mi)
{
	uint16_t        i;
	bfa_uf_t   *uf;

	/*
	 * Claim block of memory for UF list
	 */
	ufm->uf_list = (bfa_uf_t *) bfa_meminfo_kva(mi);

	/*
	 * Initialize UFs and queue it in UF free queue
	 */
	for (i = 0, uf = ufm->uf_list; i < ufm->num_ufs; i++, uf++) {
		bfa_os_memset(uf, 0, sizeof(bfa_uf_t));
		uf->bfa = ufm->bfa;
		uf->uf_tag = i;
		uf->pb_len = sizeof(bfa_uf_buf_t);
		uf->buf_kva = (void *)&ufm->uf_pbs_kva[i];
		uf->buf_pa = ufm_pbs_pa(ufm, i);
		bfa_q_enq(&ufm->uf_free_q, uf);
	}

	/**
	 * advance memory pointer
	 */
	bfa_meminfo_kva(mi) = (uint8_t *) uf;
}

static void
uf_mem_claim(bfa_uf_mod_t *ufm, bfa_meminfo_t *mi)
{
	claim_uf_pbs(ufm, mi);
	claim_ufs(ufm, mi);
	claim_uf_post_msgs(ufm, mi);
}

static void
bfa_uf_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *ndm_len, uint32_t *dm_len)
{
	uint32_t        num_ufs = cfg->fwcfg.num_uf_bufs;

	/*
	 * dma-able memory for UF posted bufs
	 */
	*dm_len += BFA_ROUNDUP((sizeof(bfa_uf_buf_t) * num_ufs),
							BFA_DMA_ALIGN_SZ);

	/*
	 * kernel Virtual memory for UFs and UF buf post msg copies
	 */
	*ndm_len += sizeof(bfa_uf_t) * num_ufs;
	*ndm_len += sizeof(bfi_uf_buf_post_t) * num_ufs;
}

static void
bfa_uf_attach(bfa_t *bfa, void *bfad, bfa_ioc_cfg_t *cfg,
		  bfa_meminfo_t *meminfo, bfa_pcidev_t *pcidev)
{
	bfa_uf_mod_t *ufm = BFA_UF_MOD(bfa);

	bfa_os_memset(ufm, 0, sizeof(bfa_uf_mod_t));
	ufm->bfa = bfa;
	ufm->num_ufs = cfg->fwcfg.num_uf_bufs;
	bfa_q_init(&ufm->uf_free_q);
	bfa_q_init(&ufm->uf_posted_q);

	uf_mem_claim(ufm, meminfo);
}

static void
bfa_uf_initdone(bfa_t *bfa)
{
}

static void
bfa_uf_detach(bfa_t *bfa)
{
}

static bfa_uf_t *
bfa_uf_get(bfa_uf_mod_t *uf_mod)
{
	bfa_uf_t   *uf;

	bfa_q_deq(&uf_mod->uf_free_q, &uf);
	return (uf);
}

static void
bfa_uf_put(bfa_uf_mod_t *uf_mod, bfa_uf_t *uf)
{
	bfa_q_enq(&uf_mod->uf_free_q, uf);
}

static void
bfa_uf_post(bfa_uf_mod_t *ufm, bfa_uf_t *uf)
{
	bfi_uf_buf_post_t *uf_post_msg;

	uf_post_msg = bfa_reqq_next(ufm->bfa, BFA_REQQ_FCXP);

	bfa_os_memcpy(uf_post_msg, &ufm->uf_buf_posts[uf->uf_tag],
		      sizeof(bfi_uf_buf_post_t));
	bfa_reqq_produce(ufm->bfa, BFA_REQQ_FCXP);

	bfa_trc(ufm->bfa, uf->uf_tag);

	bfa_q_enq(&ufm->uf_posted_q, uf);
}

static void
bfa_uf_post_all(bfa_uf_mod_t *uf_mod)
{
	bfa_uf_t   *uf;

	while ((uf = bfa_uf_get(uf_mod)) != NULL)
		bfa_uf_post(uf_mod, uf);
}

static void
uf_recv(bfa_t *bfa, bfi_uf_frm_rcvd_t *m)
{
	bfa_uf_mod_t *ufm = BFA_UF_MOD(bfa);
	uint16_t        uf_tag = m->buf_tag;
	bfa_uf_buf_t *uf_buf = &ufm->uf_pbs_kva[uf_tag];
	bfa_uf_t   *uf = &ufm->uf_list[uf_tag];
	uint8_t        *buf = &uf_buf->d[0];
	fchs_t         *fchs;

	m->frm_len = bfa_os_ntohs(m->frm_len);
	m->xfr_len = bfa_os_ntohs(m->xfr_len);

	fchs = (fchs_t *) uf_buf;

	bfa_q_qe_deq(uf);	/* dequeue from posted queue */

	uf->data_ptr = buf;
	uf->data_len = m->xfr_len;

	bfa_assert(uf->data_len >= sizeof(fchs_t));

	if (uf->data_len == sizeof(fchs_t)) {
		bfa_plog_fchdr(bfa->plog, BFA_PL_MID_HAL_UF, BFA_PL_EID_RX,
			       uf->data_len, (fchs_t *) buf);
	} else {
		uint32_t        pld_w0 = *((uint32_t *) (buf + sizeof(fchs_t)));
		bfa_plog_fchdr_and_pl(bfa->plog, BFA_PL_MID_HAL_UF,
				      BFA_PL_EID_RX, uf->data_len,
				      (fchs_t *) buf, pld_w0);
	}

	bfa_cb_queue(bfa, &uf->hcb_qe, __bfa_cb_uf_recv, uf);
}

static void
bfa_uf_stop(bfa_t *bfa)
{
}

static void
bfa_uf_iocdisable(bfa_t *bfa)
{
	bfa_uf_mod_t *ufm = BFA_UF_MOD(bfa);
	bfa_uf_t   *uf;
	bfa_q_t        *qe, *qen;

	bfa_q_iter_safe(&ufm->uf_posted_q, qe, qen) {
		uf = (bfa_uf_t *) qe;
		bfa_q_qe_deq(uf);
		bfa_uf_put(ufm, uf);
	}
}

static void
bfa_uf_start(bfa_t *bfa)
{
	bfa_uf_post_all(BFA_UF_MOD(bfa));
}

/**
 * @}
 */

/**
 * @ig hal_uf_api
 * @{
 */

/**
 * @brief
 * 		Register handler for all unsolicted recieve frames.
 *
 * @param[in]	bfa		HAL instance
 * @param[in]	ufrecv	receive handler function
 * @param[in]	cbarg	receive handler arg
 */
void
bfa_uf_recv_register(bfa_t *bfa, bfa_cb_uf_recv_t ufrecv, void *cbarg)
{
	bfa_uf_mod_t *ufm = BFA_UF_MOD(bfa);

	ufm->ufrecv = ufrecv;
	ufm->cbarg = cbarg;
}

/**
 * @brief
 * 		Free an unsolicited frame back to HAL.
 *
 * @param[in]		uf		unsolicited frame to be freed
 *
 * @return None
 */
void
bfa_uf_free(bfa_uf_t *uf)
{
	bfa_uf_post(BFA_UF_MOD(uf->bfa), uf);
}

/**
 * @}
 */

/**
 * @dg uf_pub HAL uf module public functions
 * @{
 */

void
bfa_uf_isr(bfa_t *bfa, bfi_msg_t *msg)
{
	bfa_trc(bfa, msg->mhdr.msg_id);

	switch (msg->mhdr.msg_id) {
	case BFI_UF_I2H_FRM_RCVD:
		uf_recv(bfa, (bfi_uf_frm_rcvd_t *) msg);
		break;

	default:
		bfa_trc(bfa, msg->mhdr.msg_id);
		bfa_assert(0);
	}
}

/**
 * @}
 */
