/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#ifndef __BFI_H__
#define __BFI_H__

#include <defs/bfa_defs_status.h>

#pragma pack(1)

/**
 * Msg header common to all msgs
 */
typedef struct bfa_mhdr {
	uint8_t         msg_class;	/*!< @ref bfi_mclass_t	    */
	uint8_t         msg_id;		/*!< msg opcode with in the class   */
	union {
		struct {
			uint8_t         rsvd;
			uint8_t         lpu_id;	/*!< msg destination	    */
		} h2i;
		uint16_t        i2htok;	/*!< token in msgs to host	    */
	} mtag;
} bfi_mhdr_t;

#define bfi_h2i_set(_mh, _mc, _op, _lpuid) do {		\
	(_mh).msg_class 		= (_mc);		\
	(_mh).msg_id			= (_op);		\
	(_mh).mtag.h2i.lpu_id	= (_lpuid);			\
} while (0)

#define bfi_i2h_set(_mh, _mc, _op, _i2htok) do {		\
	(_mh).msg_class 		= (_mc);		\
	(_mh).msg_id			= (_op);		\
	(_mh).mtag.i2htok		= (_i2htok);		\
} while (0)

/*
 * Message opcodes: 0-127 to firmware, 128-255 to host
 */
#define BFI_I2H_OPCODE_BASE	128
#define BFA_I2HM(_x) 			((_x) + BFI_I2H_OPCODE_BASE)

/**
 ****************************************************************************
 *
 * Scatter Gather Element and Page definition
 *
 ****************************************************************************
 */

#define BFI_SGE_INLINE	1
#define BFI_SGE_INLINE_MAX	(BFI_SGE_INLINE + 1)

/**
 * SG Flags
 */
enum {
	BFI_SGE_DATA	= 0,	/*!< data address, not last	     */
	BFI_SGE_DATA_CPL	= 1,	/*!< data addr, last in current page */
	BFI_SGE_DATA_LAST	= 3,	/*!< data address, last		     */
	BFI_SGE_LINK	= 2,	/*!< link address		     */
	BFI_SGE_PGDLEN	= 2,	/*!< cumulative data length for page */
};

/**
 * DMA addresses
 */
typedef union {
	struct {
		uint32_t        addr_lo;
		uint32_t        addr_hi;
	} a32;
} bfi_addr_t;

/**
 * Scatter Gather Element
 */
typedef struct bfi_sge {
#ifdef __BIGENDIAN
	uint32_t        flags	: 2,
			rsvd	: 2,
			sg_len	: 28;
#else
	uint32_t        sg_len	: 28,
			rsvd	: 2,
			flags	: 2;
#endif
	bfi_addr_t	sga;
} bfi_sge_t;

/**
 * Scatter Gather Page
 */
#define BFI_SGPG_DATA_SGES		7
#define BFI_SGPG_SGES_MAX		(BFI_SGPG_DATA_SGES + 1)
#define BFI_SGPG_RSVD_WD_LEN	8
typedef struct bfi_sgpg {
	bfi_sge_t	sges[BFI_SGPG_SGES_MAX];
	uint32_t	rsvd[BFI_SGPG_RSVD_WD_LEN];
} bfi_sgpg_t;

/*
 * Large Message structure - 128 Bytes size Msgs
 */
#define BFI_LMSG_SZ		128
#define BFI_LMSG_PL_WSZ	\
			((BFI_LMSG_SZ - sizeof (bfi_mhdr_t)) / 4)

typedef struct bfa_msg {
	bfi_mhdr_t	mhdr;
	uint32_t	pl[BFI_LMSG_PL_WSZ];
} bfi_msg_t;

/**
 * Message Classes
 */
typedef enum {
	BFI_MC_IOC		= 1,	/*!< IO Controller (IOC)	    */
	BFI_MC_PORT		= 2,	/*!< physical port		    */
	BFI_MC_FLASH		= 3,	/*!< Flash message class	    */
	BFI_MC_UF		= 4,	/*!< Unsolicited frame receive	    */
	BFI_MC_FCXP		= 5,	/*!< FC Transport		    */
	BFI_MC_RPORT		= 6,	/*!< Remote port		    */
	BFI_MC_ITNIM		= 7,	/*!< I-T nexus (Initiator mode)	    */
	BFI_MC_IOIM_READ	= 8,	/*!< read IO (Initiator mode)	    */
	BFI_MC_IOIM_WRITE	= 9,	/*!< write IO (Initiator mode)	    */
	BFI_MC_IOIM_IO		= 10,	/*!< IO (Initiator mode)	    */
	BFI_MC_IOIM		= 11,	/*!< IO (Initiator mode)	    */
	BFI_MC_IOIM_IOCOM	= 12,	/*!< good IO completion		    */
	BFI_MC_TSKIM		= 16,	/*!< Initiator Task management	    */
	BFI_MC_DIAG		= 17,	/*!< Diagnostic Msgs		    */
	BFI_MC_IPFC		= 18,	/*!< IP over FC Msgs		    */
	BFI_MC_SBOOT		= 19,	/*!< SAN boot services		    */
	BFI_MC_IOCFC		= 29,	/*!< FC - IO Controller (IOC)	    */
	BFI_MC_IOC_LL		= 30,	/*!< LL - IO Controller (IOC)	    */
	BFI_MC_SB		= 31,	/*!< Single Byte Command Code Set   */
	BFI_MC_MAX
} bfi_mclass_t;

#define BFI_IOC_MAX_CQS		4
#define BFI_IOC_MSGLEN_MAX		32	/* 32 bytes */

#pragma pack()

#endif /* __BFI_H__ */

