/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#ifndef __BFI_FCPIM_H__
#define __BFI_FCPIM_H__

#include "bfi.h"
#include <protocol/fcp.h>

#pragma pack(1)

/*
 * Initiator mode I-T nexus interface defines.
 */

typedef enum {
	BFI_ITNIM_H2I_CREATE_REQ = 1,	/*!< i-t nexus creation */
	BFI_ITNIM_H2I_DELETE_REQ = 2,	/*!< i-t nexus deletion */
} bfi_itnim_h2i_t;

typedef enum {
	BFI_ITNIM_I2H_CREATE_RSP = BFA_I2HM(1),
	BFI_ITNIM_I2H_DELETE_RSP = BFA_I2HM(2),
	BFI_ITNIM_I2H_SLER_EVENT = BFA_I2HM(3),
} bfi_itnim_i2h_t;

typedef struct {
	bfi_mhdr_t  mh;		/*!< common msg header		 */
	uint16_t        fw_handle;	/*!< f/w handle for itnim	 */
	uint8_t         class;		/*!< FC class for IO		 */
	uint8_t         seq_rec;	/*!< sequence recovery support	 */
	uint8_t         msg_no;		/*!< seq id of the msg		 */
} bfi_itnim_create_req_t;

typedef struct {
	bfi_mhdr_t  mh;		/*!< common msg header		 */
	uint16_t        bfa_handle;	/*!< bfa handle for itnim	 */
	uint8_t         status;		/*!< fcp request status		 */
	uint8_t         seq_id;		/*!< seq id of the msg		 */
} bfi_itnim_create_rsp_t;

typedef struct {
	bfi_mhdr_t  mh;		/*!< common msg header		 */
	uint16_t        fw_handle;	/*!< f/w itnim handle		 */
	uint8_t         seq_id;		/*!< seq id of the msg		 */
	uint8_t         rsvd;
} bfi_itnim_delete_req_t;

typedef struct {
	bfi_mhdr_t  mh;		/*!< common msg header		 */
	uint16_t        bfa_handle;	/*!< bfa handle for itnim	 */
	uint8_t         status;		/*!< fcp request status		 */
	uint8_t         seq_id;		/*!< seq id of the msg		 */
} bfi_itnim_delete_rsp_t;

typedef struct {
	bfi_mhdr_t  mh;		/*!< common msg header		 */
	uint16_t        bfa_handle;	/*!< bfa handle for itnim	 */
	uint16_t        rsvd;
} bfi_itnim_sler_event_t;

typedef union {
	bfi_itnim_create_req_t *create_req;
	bfi_itnim_delete_req_t *delete_req;
	bfi_msg_t      *msg;
} bfi_itnim_h2i_msg_t;

typedef union {
	bfi_itnim_create_rsp_t *create_rsp;
	bfi_itnim_delete_rsp_t *delete_rsp;
	bfi_itnim_sler_event_t *sler_event;
	bfi_msg_t      *msg;
} bfi_itnim_i2h_msg_t;

/*
 * Initiator mode IO interface defines.
 */

typedef enum {
	BFI_IOIM_H2I_IOABORT_REQ = 1,	/*!< IO abort request	 */
	BFI_IOIM_H2I_IOCLEANUP_REQ = 2,	/*!< IO cleanup request	 */
} bfi_ioim_h2i_t;

typedef enum {
	BFI_IOIM_I2H_IO_RSP = BFA_I2HM(1),	/*!< non-fp IO response	 */
	BFI_IOIM_I2H_IOABORT_RSP = BFA_I2HM(2),/*!< ABORT rsp	 */
} bfi_ioim_i2h_t;

/**
 * IO command DIF info
 */
typedef struct {
	uint32_t	dif_info[4];
} bfi_ioim_dif_t;

/**
 * FCP IO messages overview
 *
 * @note
 * - Max CDB length supported is 64 bytes.
 * - SCSI Linked commands and SCSI bi-directional Commands not
 * 	 supported.
 *
 */
typedef struct {
	bfi_mhdr_t  mh;		/*!< Common msg header		 */
	uint16_t        io_tag;		/*!< I/O tag			 */
	uint16_t        rport_hdl;	/*!< itnim/rport firmware handle */
	fcp_cmnd_t      cmnd;		/*!< IO request info		 */

	/**
	 * SG elements array within the IO request must be double word
	 * aligned. This aligment is required to optimize SGM setup for the IO.
	 */
	bfi_sge_t   sges[BFI_SGE_INLINE_MAX];
	uint8_t         io_timeout;
	uint8_t         dif_en;
	uint8_t         rsvd_a[2];
	bfi_ioim_dif_t  dif;
} bfi_ioim_req_t;

/**
 *  @brief
 * 	This table shows various IO status codes from firmware and their
 *	meaning. Host driver can use these status codes to further process
 *	IO completions.
 *
 * 	BFI_IOIM_STS_OK		: IO completed with error free SCSI &
 *					   transport status.
 * 					   - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_SCSI_ERR		: IO completed with scsi error.
 * 					  - io-tag can be reused.
 *
 * 	BFI_IOIM_STS_HOST_ABORTED	: IO was aborted successfully due to
 *					   host request.
 * 					   - io-tag cannot be reused yet.
 *
 * 	BFI_IOIM_STS_ABORTED		: IO was aborted successfully
 *					   internally by f/w.
 * 					   - io-tag cannot be reused yet.
 *
 * 	BFI_IOIM_STS_TIMEDOUT	: IO timedout and ABTS/RRQ is happening
*					   in the firmware and
 * 					   - io-tag cannot be reused yet.
 *
 * 	BFI_IOIM_STS_SQER_NEEDED	: Firmware could not recover the IO
 *					  with sequence level error
 * 					  logic and hence host needs to retry
 *					  this IO with a different IO tag
 * 					  - io-tag cannot be used yet.
 *
 * 	BFI_IOIM_STS_NEXUS_ABORT	: Second Level Error Recovery from host
 *					  is required because 2 consecutive ABTS
 *					  timedout and host needs logout and
 *					  re-login with the target
 * 					  - io-tag cannot be used yet.
 *
 * 	BFI_IOIM_STS_UNDERRUN	: IO completed with SCSI status good,
 *					  but the data tranferred is less than
 *					  the fcp data length in the command.
 *					  ex. SCSI INQUIRY where transferred
 *					  data length and residue count in FCP
 *					  response accounts for total fcp-dl
 * 					  - io-tag can be reused.
 *
 * 	BFI_IOIM_STS_OVERRUN	: IO completed with SCSI status good,
 *					  but the data transerred is more than
 *					  fcp data length in the command. ex.
 *					  TAPE IOs where blocks can of unequal
 *					  lengths.
 * 					  - io-tag can be reused.
 *
 * 	BFI_IOIM_STS_RES_FREE	: Firmware has completed using io-tag
 *					  during abort process
 * 					  - io-tag can be reused.
 *
 * 	BFI_IOIM_STS_PROTO_ERR	: Firmware detected a protocol error.
 *					  ex target sent more data than
 *					  requested, or there was data frame
 *					  loss and other reasons
 * 					  - io-tag cannot be used yet.
 *
 * 	BFI_IOIM_STS_DIF_ERR	: Firwmare detected DIF error. ex: DIF
 * 					  CRC err or Ref Tag err or App tag err.
 * 					  - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_TSK_MGT_ABORT	: IO was aborted because of Task
 *					  Management command from the host
 * 					  - io-tag can be reused.
 *
 * 	BFI_IOIM_STS_UTAG		: Firmware does not know about this
 *					  io_tag.
 * 					  - io-tag can be reused.
 */
typedef enum {
	BFI_IOIM_STS_OK = 0,
	BFI_IOIM_STS_HOST_ABORTED = 1,
	BFI_IOIM_STS_ABORTED = 2,
	BFI_IOIM_STS_TIMEDOUT = 3,
	BFI_IOIM_STS_RES_FREE = 4,
	BFI_IOIM_STS_SQER_NEEDED = 5,
	BFI_IOIM_STS_PROTO_ERR = 6,
	BFI_IOIM_STS_UTAG = 7,
	BFI_IOIM_STS_PATHTOV = 8,
} bfi_ioim_status_t;

typedef struct {
	uint16_t        io_tag;		/*!< completed IO tag		 */
	uint16_t        bfa_rport_hndl;	/*!< releated rport handle	 */
	uint8_t         io_status;	/*!< IO completion status	 */
	uint8_t         reuse_io_tag;	/*!< IO tag can be reused        */
	uint16_t	abort_tag;	/*!< host abort request tag      */
} bfi_ioim_rsp_info_t;

#define BFI_IOIM_RSP_PAD					\
	((sizeof(fchs_t) - sizeof(bfi_mhdr_t) -		\
	  sizeof(bfi_ioim_rsp_info_t)) / sizeof(uint32_t))
#define BFI_IOIM_SNSLEN	(256)

/**
 * I/O response message
 */
typedef struct {
	bfi_mhdr_t  mh;			/*!< common msg header	 */
	bfi_ioim_rsp_info_t rsp_info;	/*!< io completion info	 */
	uint32_t        pad[BFI_IOIM_RSP_PAD];
	fcp_resp_t      fcp_rsp;		/*!< FCP_RSP contents	 */
} bfi_ioim_rsp_t;

typedef struct {
	bfi_mhdr_t  mh;             /*!< Common msg header  */
	uint16_t        io_tag;         /*!< I/O tag            */
	uint16_t        abort_tag;      /*!< unique request tag */
} bfi_ioim_abort_req_t;

/*
 * Initiator mode task management command interface defines.
 */

typedef enum {
	BFI_TSKIM_H2I_TM_REQ    = 1, /*!< task-mgmt command         */
	BFI_TSKIM_H2I_ABORT_REQ = 2, /*!< task-mgmt command         */
} bfi_tskim_h2i_t;

typedef enum {
	BFI_TSKIM_I2H_TM_RSP = BFA_I2HM(1),
} bfi_tskim_i2h_t;

typedef struct {
	bfi_mhdr_t  mh;             /*!< Common msg header          */
	uint16_t        tsk_tag;        /*!< task management tag        */
	uint16_t        itn_fhdl;       /*!< itn firmware handle        */
	lun_t           lun;            /*!< LU number                  */
	uint8_t         tm_flags;       /*!< see fcp_tm_cmnd_t          */
	uint8_t         t_secs;         /*!< Timeout value in seconds   */
	uint8_t         rsvd[2];
} bfi_tskim_req_t;

typedef struct {
	bfi_mhdr_t  mh;             /*!< Common msg header          */
	uint16_t        tsk_tag;        /*!< task management tag        */
	uint16_t        rsvd;
} bfi_tskim_abortreq_t;

typedef enum {
	/*
	 * Following are FCP-4 spec defined status codes,
	 * **DO NOT CHANGE THEM **
	 */
	BFI_TSKIM_STS_OK       = 0,
	BFI_TSKIM_STS_NOT_SUPP = 4,
	BFI_TSKIM_STS_FAILED   = 5,

	/**
	 * Defined by BFA
	 */
	BFI_TSKIM_STS_TIMEOUT  = 10,    /*!< TM request timedout     */
	BFI_TSKIM_STS_ABORTED  = 11,    /*!< Aborted on host request */
} bfi_tskim_status_t;

typedef struct {
	bfi_mhdr_t  mh;		/*!< Common msg header		 */
	uint16_t        tsk_tag;	/*!< task mgmt cmnd tag		 */
	uint8_t         tsk_status;	/*!< @ref bfi_tskim_status_t */
	uint8_t         rsvd;
} bfi_tskim_rsp_t;

#pragma pack()

#endif /* __BFI_FCPIM_H__ */

