/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#ifndef __BFA_DEFS_IOC_H__
#define __BFA_DEFS_IOC_H__

#include <protocol/types.h>
#include <defs/bfa_defs_types.h>
#include <defs/bfa_defs_version.h>
#include <defs/bfa_defs_adapter.h>
#include <defs/bfa_defs_pm.h>

/**
 * IOC driver attributes
 */
enum {
	BFA_IOC_DRIVER_LEN = 16,
};

/**
 * Driver and firmware versions.
 */
typedef struct {
	char            driver[BFA_IOC_DRIVER_LEN];	/*!< driver name */
	char            driver_ver[BFA_VERSION_LEN];	/*!< driver version */
	char            fw_ver[BFA_VERSION_LEN];	/*!< firmware version*/
	char            bios_ver[BFA_VERSION_LEN];	/*!< bios version */
	char            efi_ver[BFA_VERSION_LEN];	/*!< EFI version */
	char            ob_ver[BFA_VERSION_LEN];	/*!< openboot version*/
} bfa_ioc_driver_attr_t;

/**
 * IOC chip attributes
 */
enum {
	BFA_IOC_CHIP_NAME_LEN 	= 16,
	BFA_IOC_CHIP_REV_LEN 	= 8,
};

#define BFA_CHIP_NAME		"Crossbow"

typedef struct {
	uint8_t         chip_name[BFA_IOC_CHIP_NAME_LEN];/*!< chip name */
	uint8_t         chip_rev[BFA_IOC_CHIP_REV_LEN];	 /*!< chip revision */
} bfa_ioc_chip_attr_t;

/**
 * IOC PCI device attributes
 */
typedef struct {
	uint16_t        vendor_id;	/*!< PCI vendor ID */
	uint16_t        device_id;	/*!< PCI device ID */
	uint16_t        ssid;		/*!< subsystem ID */
	uint16_t        ssvid;		/*!< subsystem vendor ID */
	uint32_t        pcifn;		/*!< PCI device function */
	uint32_t        rsvd;		/* padding */
} bfa_ioc_pci_attr_t;

#define BFA_IOC_INTR_DELAY	1125
#define BFA_IOC_INTR_LATENCY	225
/**
 * Interrupt coalescing configuration.
 */
typedef struct {
	bfa_boolean_t	coalesce;	/*!< enable/disable coalescing */
	uint16_t	latency;	/*!< latency in microseconds   */
	uint16_t	delay;		/*!< delay in microseconds     */
} bfa_ioc_intr_attr_t;

/**
 * IOC firmware configuraton
 */
typedef struct {
	uint8_t         num_cqs;	/*!< number of req/rsp queues */
	uint8_t         rsvd;		/*!< padding */
	uint16_t        num_itnims;	/*!< number of i-t-nexus initiator
					 *    mode */
	uint16_t        num_ioim_reqs;	/*!< number of outstanding IO reqs */
	uint16_t        num_tskim_reqs;	/*!< active task management requests */
	uint16_t        num_fcxp_reqs;	/*!< number of active FC transport
					 *    reqs */
	uint16_t        num_uf_bufs;	/*!< number of unsolicited bufs posted
					 */
	uint16_t        num_rports;	/*!< number of unsolicited bufs posted
					 */
	uint16_t        num_itntms;	/*!< number of target mode I-T nexuses
					 */
} bfa_ioc_fwcfg_t;

/**
 * IOC configuration
 */
typedef struct {
	bfa_ioc_fwcfg_t fwcfg;		/*!< firmware side configuration */
	uint16_t        num_reqq_elems;	/*!< number of req queue elements */
	uint16_t        num_rspq_elems;	/*!< number of rsp queue elements */
	uint16_t        num_sgpgs;	/*!< number of total SG pages */
	uint16_t        num_sboot_tgts;	/*!< number of SAN boot targets */
	uint16_t        num_sboot_luns;	/*!< number of SAN boot luns */
	uint16_t        min_cfg;        /*!< min cfg flag */
	uint16_t        rsvd;
	uint16_t        path_tov;	/*!< path timeout in non-mpio mode */
	bfa_boolean_t   mpio_mode;	/*!< Enable fast failover */
	bfa_boolean_t 	delay_comp;	/*!< Delay completion of failed inflight IOs */
	bfa_boolean_t   ioc_recover;/*!< IOC recovery mode */
	bfa_boolean_t   fcxp_sbuf;	/*!< FCXP small buffers */
} bfa_ioc_cfg_t;

/**
 * IOC states
 */
typedef enum {
	BFA_IOC_RESET       = 1,  /*!< IOC is in reset state */
	BFA_IOC_SEMWAIT     = 2,  /*!< Waiting for IOC hardware semaphore */
	BFA_IOC_HWINIT 	    = 3,  /*!< IOC hardware is being initialized */
	BFA_IOC_CFGINIT     = 4,  /*!< IOC is being configured */
	BFA_IOC_CFGDONE     = 5,  /*!< IOC is configured */
	BFA_IOC_OPERATIONAL = 6,  /*!< IOC is operational */
	BFA_IOC_HWFAIL      = 7,  /*!< IOC hardware failure */
	BFA_IOC_HBFAIL      = 8,  /*!< IOC heart-beat failure */
	BFA_IOC_RECOVER     = 9,  /*!< IOC recovery in progress */
	BFA_IOC_DISABLING   = 10, /*!< IOC is being disabled */
	BFA_IOC_DISABLED    = 11, /*!< IOC is disabled */
} bfa_ioc_state_t;

/**
 * IOC firmware stats
 */
typedef struct {
	uint32_t        hb_count;
	uint32_t        cfg_reqs;
	uint32_t        enable_reqs;
	uint32_t        disable_reqs;
	uint32_t        stats_reqs;
	uint32_t        clrstats_reqs;
	uint32_t        unknown_reqs;
	uint32_t        ic_reqs;		/*!< interrupt coalesce reqs */
} bfa_fw_ioc_stats_t;

/**
 * IOC firmware IO stats
 */
typedef struct {
	uint32_t	host_abort;		/*!< IO aborted by host driver*/
	uint32_t	host_cleanup;		/*!< IO clean up by host driver */

	uint32_t	fw_io_timeout;		/*!< IOs timedout */
	uint32_t	fw_frm_parse;		/*!< frame parsed by f/w */
	uint32_t	fw_frm_data;		/*!< fcp_data frame parsed by f/w */
	uint32_t	fw_frm_rsp;		/*!< fcp_rsp frame parsed by f/w */
	uint32_t	fw_frm_xfer_rdy;	/*!< xfer_rdy frame parsed by f/w */
	uint32_t	fw_frm_bls_acc;		/*!< BLS ACC  frame parsed by f/w */
	uint32_t	fw_frm_tgt_abort;	/*!< target ABTS parsed by f/w */
	uint32_t	fw_frm_unknown;		/*!< unknown parsed by f/w */
	uint32_t	fw_data_dma;		/*!< f/w DMA'ed the data frame */
	uint32_t	fw_frm_drop;		/*!< f/w drop the frame */

	uint32_t	rec_timeout;		/*!< FW rec timed out */
	uint32_t	error_rec;		/*!< FW sending rec on 
						  *  an error condition*/
	uint32_t	wait_for_si;		/*!< FW wait for SI */
	uint32_t	rec_rsp_inval;		/*!< REC rsp invalid */
	uint32_t	seqr_io_abort;		/*!< target does not know cmd so abort */
	uint32_t	seqr_io_retry;		/*!< SEQR failed so retry IO */
	
	uint32_t	itn_cisc_upd_rsp;	/*!< ITN cisc updated on fcp_rsp */
	uint32_t	itn_cisc_upd_data;	/*!< ITN cisc updated on fcp_data */
	uint32_t	itn_cisc_upd_xfer_rdy;	/*!< ITN cisc updated on fcp_data */

	uint32_t	fcp_data_lost;		/*!< fcp data lost */

	uint32_t	ro_set_in_xfer_rdy;	/*!< Target set RO in Xfer_rdy frame */
	uint32_t	xfer_rdy_ooo_err;	/*!< Out of order Xfer_rdy received */
	uint32_t	xfer_rdy_unknown_err;	/*!< unknown error in xfer_rdy frame */
			
	uint32_t	io_abort_timeout;	/*!< ABTS timedout  */
	uint32_t	sler_initiated;		/*!< SLER initiated */

	uint32_t	unexp_fcp_rsp;		/*!< fcp response in wrong state */

	uint32_t	fcp_rsp_under_run;	/*!< fcp rsp IO underrun */
	uint32_t	fcp_rsp_under_run_err;	/*!< fcp rsp IO underrun error */
	uint32_t	fcp_rsp_over_run;	/*!< fcp rsp IO overrun */
	uint32_t	fcp_rsp_over_run_err;	/*!< fcp rsp IO overrun error */
	uint32_t	fcp_rsp_proto_err;	/*!< protocol error in fcp rsp */
	uint32_t	fcp_rsp_sense_err;	/*!< error in sense info in fcp rsp */
	uint32_t	fcp_conf_req;		/*!< FCP conf requested */

	uint32_t	tgt_aborted_io;		/*!< target initiated abort */

	uint32_t	ioh_edtov_timeout_event;/*!< IOH edtov timer popped */
	uint32_t	ioh_fcp_rsp_excp_event;	/*!< IOH FCP_RSP exception */
	uint32_t	ioh_fcp_conf_event;	/*!< IOH FCP_CONF */
	uint32_t	ioh_mult_frm_rsp_event;	/*!< IOH multi_frame FCP_RSP */
	uint32_t	ioh_hit_class2_event;	/*!< IOH hit class2 */
	uint32_t	ioh_miss_other_event;	/*!< IOH miss other */
	uint32_t	ioh_seq_cnt_err_event;	/*!< IOH seq cnt error */
	uint32_t	ioh_len_err_event;	/*!< IOH len error- fcp_dl != bytes xfered */
	uint32_t	ioh_seq_len_err_event;	/*!< IOH seq len error */
	uint32_t	ioh_data_oor_event;	/*!< Data out of range */
	uint32_t	ioh_ro_ooo_event;	/*!< Relative offset out of range */
	uint32_t	ioh_cpu_owned_event;	/*!< IOH hit -iost owned by f/w */
	uint32_t	ioh_unexp_frame_event;	/*!< unexpected frame recieved
						 *   count */
	uint32_t	ioh_err_int;		/*!< IOH error int during data-phase
						 *   for scsi write
						 */
} bfa_fw_io_stats_t;

/**
 * IOC port firmware stats
 */

typedef struct {
    uint32_t    intr_evt;
    uint32_t    intr;
    uint32_t    intr_excess;
    uint32_t    intr_cause0;
    uint32_t    intr_other;
    uint32_t    intr_other_ign;
    uint32_t    sig_lost;
    uint32_t    sig_regained;
    uint32_t    sync_lost;
    uint32_t    sync_to;
    uint32_t    sync_regained;
    uint32_t    div2_overflow;
    uint32_t    div2_underflow;
    uint32_t    efifo_overflow;
    uint32_t    efifo_underflow;
    uint32_t    idle_rx;
    uint32_t    lrr_rx;
    uint32_t    lr_rx;
    uint32_t    ols_rx;
    uint32_t    nos_rx;
    uint32_t    lip_rx;
    uint32_t    arbf0_rx;
    uint32_t    mrk_rx;
    uint32_t    const_mrk_rx;
    uint32_t    prim_unknown;
    uint32_t    rsvd;
} bfa_fw_port_fpg_stats_t;

/**
 * @brief Link state machine statistics
 */
typedef struct {
    uint32_t    hwsm_success;       /*!< hwsm state machine success          */
    uint32_t    hwsm_fails;         /*!< hwsm fails                          */
    uint32_t    hwsm_wdtov;         /*!< hwsm timed out                      */
    uint32_t    swsm_success;       /*!< swsm success                        */
    uint32_t    swsm_fails;         /*!< swsm fails                          */
    uint32_t    swsm_wdtov;         /*!< swsm timed out                      */
    uint32_t    busybufs;           /*!< link init failed due to busybuf     */
    uint32_t    buf_waits;          /*!< bufwait state entries               */
    uint32_t    link_fails;         /*!< link failures                       */
    uint32_t    psp_errors;         /*!< primitive sequence protocol errors  */
    uint32_t    lr_unexp;           /*!< No. of times LR rx-ed unexpectedly  */
    uint32_t    lrr_unexp;          /*!< No. of times LRR rx-ed unexpectedly */
    uint32_t    lr_tx;              /*!< No. of times LR tx started          */
    uint32_t    lrr_tx;             /*!< No. of times LRR tx started         */
    uint32_t    ols_tx;             /*!< No. of times OLS tx started         */
    uint32_t    nos_tx;             /*!< No. of times NOS tx started         */
} bfa_fw_port_lksm_stats_t;

/**
 * @brief SNSM statistics
 */
typedef struct {
    uint32_t    hwsm_success;       /*!< Successful hwsm terminations        */
    uint32_t    hwsm_fails;         /*!< hwsm fail count                     */
    uint32_t    hwsm_wdtov;         /*!< hwsm timed out                      */
    uint32_t    swsm_success;       /*!< swsm success                        */
    uint32_t    swsm_wdtov;         /*!< swsm timed out                      */
    uint32_t    error_resets;       /*!< error resets initiated by upsm      */
} bfa_fw_port_snsm_stats_t;

/**
 * @brief PHYSM statistics
 */
typedef struct {
    uint32_t    module_inserts;     /*!< Module insert count                 */
    uint32_t    module_xtracts;     /*!< Module extracts count               */
    uint32_t    module_invalids;    /*!< Invalid module inserted count       */
    uint32_t    module_read_ign;    /*!< Module validation status ignored    */
    uint32_t    laser_faults;       /*!< Laser fault count                   */
    uint32_t    sig_lost;           /*!< Signal loss count                   */
    uint32_t    sync_lost;          /*!< Sync loss count                     */
    uint32_t    rsvd;
} bfa_fw_port_physm_stats_t;

/**
 * IOC firmware port stats
 */
typedef struct {
	bfa_fw_port_fpg_stats_t		fpg_stats;
	bfa_fw_port_physm_stats_t	physm_stats;
	bfa_fw_port_snsm_stats_t	snsm_stats;
	bfa_fw_port_lksm_stats_t	lksm_stats;
} bfa_fw_port_stats_t;

/**
 * IOC firmware stats
 */
typedef struct {
	bfa_fw_ioc_stats_t	ioc_stats;
	bfa_fw_io_stats_t	io_stats;
	bfa_fw_port_stats_t	port_stats;
} bfa_fw_stats_t;

/**
 * IOC driver stats
 */
typedef struct {
	uint32_t	ioc_isrs;
	uint32_t	ioc_starts;
	uint32_t	ioc_stops;
	uint32_t	ioc_enables;
	uint32_t	ioc_disables;
	uint32_t	ioc_hbfails;
	uint32_t	ioc_boots;
	uint32_t	stats_tmos;
} bfa_ioc_drv_stats_t;

/**
 * IOC statistics
 */
typedef struct {
	uint32_t        hb_count;
	uint32_t        disable_reqs;
	uint32_t        enable_reqs;
	uint32_t        disable_replies;
	uint32_t        enable_replies;
	uint32_t        nsuspends;		/*!< number of IOC suspends  */
	bfa_ioc_drv_stats_t	drv_stats;	/*!< driver IOC stats        */
	bfa_fw_stats_t		fw_stats;	/*!< firmware IOC stats      */
} bfa_ioc_stats_t;

/**
 * IOC attributes returned in queries
 */
typedef struct {
	bfa_ioc_cfg_t   	fwcfg;		/*!< IOC firmware config */
	bfa_ioc_driver_attr_t 	driver_attr; 	/*!< driver attributes */
	bfa_ioc_state_t 	state;		/*!< IOC current state */
	bfa_pm_ds_t     	power_attr;	/*!< IOC power attribute */
	bfa_adapter_attr_t 	adapter_attr;	/*!< HBA attributes */
	bfa_ioc_chip_attr_t 	chip_attr;	/*!< chip level attributes */
	bfa_ioc_pci_attr_t 	pci_attr;	/*!< PCI device attributes */
	bfa_ioc_intr_attr_t 	intr_attr;	/*!< IOC interrupt attributes*/
} bfa_ioc_attr_t;

/**
 * BFA IOC level events
 */
typedef enum {
	BFA_IOC_AEN_HBGOOD  = 1,	/*!< Heart Beat restore event */
	BFA_IOC_AEN_HBFAIL  = 2,	/*!< Heart Beat failure event */
	BFA_IOC_AEN_ENABLE  = 3,	/*!< IOC enabled event */
	BFA_IOC_AEN_DISABLE = 4,	/*!< IOC disabled event */
} bfa_ioc_aen_event_t;

/**
 * BFA IOC level event data, now just a place holder
 */
typedef struct {
	wwn_t           pwwn;
} bfa_ioc_aen_data_t;

#define BFA_IOC_MAX_PATH_FAILOVER_TOV	60
#define BFA_IOC_MAX_Q_DEPTH		2000

#endif /* __BFA_DEFS_IOC_H__ */
