/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
#ifndef __BFA_H__
#define __BFA_H__

#include <bfa_os_inc.h>
#include <cs/bfa_debug.h>
#include <cs/bfa_q.h>
#include <cs/bfa_trc.h>
#include <cs/bfa_log.h>
#include <cs/bfa_plog.h>
#include <defs/bfa_defs_status.h>
#include <defs/bfa_defs_ioc.h>
#include <aen/bfa_aen.h>
#include <bfi/bfi.h>

typedef struct bfa_s bfa_t;
#include <bfa_intr_priv.h>

/**
 * HAL interrupt return enums
 */
typedef enum {
	BFA_ISR_NOTCLAIMED = 1,	/*!< interrupt not ours */
	BFA_ISR_CLAIMED_COMP = 2,	/*!< claimed; completions pending */
	BFA_ISR_CLAIMED_NO_COMP = 3,	/*!< claimed; no completions */
} bfa_isr_status_t;

/**
 * PCI device information required by HAL
 */
typedef struct {
	int             pci_slot;
	uint8_t         pci_func;
	bfa_os_addr_t   pci_bar_kva;
} bfa_pcidev_t;

/**
 * PCI devices supported by the current HAL
 */
typedef struct {
	uint16_t        device_id;
	uint16_t        vendor_id;
} bfa_pciid_t;

extern char     bfa_version[];

/**
 * HAL Power Mgmt Commands
 */
typedef enum {
	BFA_PM_CTL_D0 = 0,
	BFA_PM_CTL_D1 = 1,
	BFA_PM_CTL_D2 = 2,
	BFA_PM_CTL_D3 = 3,
} bfa_pm_cmd_t;

/**
 * BFA memory resources
 */
typedef enum {
	BFA_MEM_TYPE_KVA = 1,	/*! Kernel Virtual Memory *(non-dma-able) */
	BFA_MEM_TYPE_DMA = 2,	/*! DMA-able memory */
	BFA_MEM_TYPE_MAX = BFA_MEM_TYPE_DMA,
} bfa_mem_type_t;

typedef struct {
	bfa_mem_type_t	mem_type;	/*!< see bfa_mem_type_t		*/
	uint32_t        mem_len;	/*!< Total Length in Bytes	*/
	uint8_t       	*kva;		/*!< kernel virtual address	*/
	uint64_t        dma;		/*!< dma address if DMA memory	*/
	uint8_t       	*kva_curp;	/*!< kva allocation cursor	*/
	uint64_t        dma_curp;	/*!< dma allocation cursor	*/
} bfa_mem_elem_t;

typedef struct {
	bfa_mem_elem_t meminfo[BFA_MEM_TYPE_MAX];
} bfa_meminfo_t;
#define bfa_meminfo_kva(_m)	\
	(_m)->meminfo[BFA_MEM_TYPE_KVA - 1].kva_curp
#define bfa_meminfo_dma_virt(_m)	\
	(_m)->meminfo[BFA_MEM_TYPE_DMA - 1].kva_curp
#define bfa_meminfo_dma_phys(_m)	\
	(_m)->meminfo[BFA_MEM_TYPE_DMA - 1].dma_curp

/**
 * Generic Scatter Gather Element used by driver
 */
typedef struct {
	uint32_t        sg_len;
	void           *sg_addr;
} bfa_sge_t;

#define bfa_sge_to_be(__sge) do {                                          \
	((uint32_t *)(__sge))[0] = bfa_os_htonl(((uint32_t *)(__sge))[0]); \
	((uint32_t *)(__sge))[1] = bfa_os_htonl(((uint32_t *)(__sge))[1]); \
	((uint32_t *)(__sge))[2] = bfa_os_htonl(((uint32_t *)(__sge))[2]); \
} while (0)


/**
 * BFA timer data structure
 */
typedef struct {
	bfa_q_t         qe;
	void            (*timercb) (void *);
	void            *arg;
	int             timeout;	/**< in millisecs. */
} bfa_timer_t;

/*
 * bfa API functions
 */
void            bfa_get_pciids(bfa_pciid_t **pciids, int *npciids);
void            bfa_cfg_get_default(bfa_ioc_cfg_t *cfg);
void            bfa_cfg_get_min(bfa_ioc_cfg_t *cfg);
void            bfa_cfg_get_meminfo(bfa_ioc_cfg_t *cfg,
					bfa_meminfo_t *meminfo);
void            bfa_attach(bfa_t *bfa, void *bfad, bfa_ioc_cfg_t *cfg,
			       bfa_meminfo_t *meminfo,
			       bfa_pcidev_t *pcidev);
void            bfa_init_trc(bfa_t *bfa, bfa_trc_mod_t *trcmod);
void            bfa_init_log(bfa_t *bfa, bfa_log_mod_t *logmod);
void            bfa_init_aen(bfa_t *bfa, bfa_aen_t *aen);
void            bfa_init_plog(bfa_t *bfa, bfa_plog_t *plog);
void            bfa_detach(bfa_t *bfa);
void            bfa_init(bfa_t *bfa);
void            bfa_start(bfa_t *bfa);
void            bfa_stop(bfa_t *bfa);
void            bfa_attach_fcs(bfa_t *bfa);
void            bfa_cb_init(void *bfad, bfa_status_t status);
void            bfa_cb_stop(void *bfad, bfa_status_t status);

#define BFA_TIMER_FREQ 500 /**< specified in millisecs */

void            bfa_timer_tick(bfa_t *bfa);
void            bfa_msix_getvecs(bfa_t *bfa, uint32_t *msix_vecs_bmap,
				     uint32_t *num_vecs,
				     uint32_t *max_vec_bit);
bfa_isr_status_t bfa_intx(bfa_t *bfa);
bfa_isr_status_t bfa_msix_reqq(bfa_t *bfa, int qid);
bfa_isr_status_t bfa_msix_rspq(bfa_t *bfa, int qid);
bfa_isr_status_t bfa_msix_lpu(bfa_t *bfa);
bfa_isr_status_t bfa_msix_errint(bfa_t *bfa);
void            bfa_isr_enable(bfa_t *bfa);
void            bfa_isr_disable(bfa_t *bfa);

void            bfa_comp_deq(bfa_t *bfa, bfa_q_t *comp_q);
void            bfa_comp_process(bfa_t *bfa, bfa_q_t *comp_q);
void            bfa_comp_free(bfa_t *bfa, bfa_q_t *comp_q);

typedef void    (*bfa_cb_ioc_t) (void *cbarg, bfa_status_t status);
bfa_status_t    bfa_iocfc_get_stats(bfa_t *bfa, bfa_ioc_stats_t *stats,
				      bfa_cb_ioc_t cbfn, void *cbarg);
bfa_status_t    bfa_iocfc_clear_stats(bfa_t *bfa, bfa_cb_ioc_t cbfn,
					void *cbarg);
void            bfa_ioc_get_attr(bfa_t *bfa, bfa_ioc_attr_t *ioc_attr);

void bfa_ioc_get_adapter_attr(bfa_t *bfa, bfa_adapter_attr_t *ad_attr);

bfa_status_t    bfa_iocfc_israttr_set(bfa_t *bfa, bfa_ioc_intr_attr_t *attr);

bfa_status_t    bfa_iocfc_enable(bfa_t *bfa);
void            bfa_iocfc_disable(bfa_t *bfa);
void		bfa_ioc_auto_recover(bfa_boolean_t auto_recover);
void            bfa_cb_ioc_disable(void *bfad);

/*
 * bfa PM API functions
 */
void            bfa_ioc_pm_ctl(bfa_t *bfa, bfa_pm_cmd_t cmd);

/*
 * bfa mfg wwn API functions
 */
wwn_t           bfa_wwn_get_naa1(bfa_t *bfa);
wwn_t           bfa_wwn_get_naa2(bfa_t *bfa);
wwn_t           bfa_wwn_get_naa5(bfa_t *bfa, uint16_t inst);

/*
 * bfa mfg adapter ID API functions
 */
uint64_t        bfa_get_adapter_id(bfa_t *bfa);

/*
 * bfa HAL debug API functions
 */
bfa_status_t bfa_debug_fwtrc(bfa_t *bfa, void *trcdata, int *trclen);
bfa_status_t bfa_debug_fwsave(bfa_t *bfa, void *trcdata, int *trclen);

/*
 * bfa timer API functions
 */
void            bfa_timer_init(struct bfa_s *bfa);
void            bfa_timer_start(struct bfa_s *bfa,
				    bfa_timer_t *timer,
				    void (*timercb) (void *), void *arg,
				    unsigned int timeout);
void            bfa_timer_stop(bfa_timer_t *timer);

#include "bfa_priv.h"

#endif /* __BFA_H__ */
